#if __cplusplus >= 201703L // >= C++17

// $Id: op_delete_a.cc 2742 2022-09-17 21:59:20Z teshields $

// This file is part of the Libnew package that provides replacement
// implementations for core standard C library memory management
// functions, and for the standard C++ 'new()' and 'delete()' operators.
//
// Copyright (C) 2020-2022  Thomas E. Shields
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#if defined (TRACE)
#include <iostream>
#endif

#include <new>

#if defined (USE_BDWGC)
#include <gc/gc.h>
#elif defined (USE_DMALLOC)
#include <dmalloc.h>
#endif

void operator delete (void* p, std::align_val_t al) noexcept
{
#if defined (TRACE)
  std::cerr << "operator delete (void*, align_val_t) noexcept"
               ": deallocating @ " << p << std::endl;
#endif

#if defined (USE_BDWGC)
  GC_FREE (p);
#elif defined (USE_DMALLOC)
  dmalloc_free (__FILE__, __LINE__, p, DMALLOC_FUNC_FREE);
#else
  free (p);
#endif
}

#endif // __cplusplus >= 201703L // >= C++17
