/* $Id: strndup.c 2741 2022-09-17 21:58:12Z teshields $ */

/* This file is part of the Libnew package that provides replacement
   implementations for core standard C library memory management
   functions, and for the standard C++ 'new()' and 'delete()' operators.

   Copyright (C) 2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#if defined (TRACE)
#include <stdio.h>
#endif

#if defined (__sun) && defined (__SVR4)
#include <strings.h>
#else
#include <string.h>
#endif

#if defined (USE_BDWGC)
#include <gc/gc.h>
#warning "The BDWGC package must be configured with --enable-redirect-malloc \
in order to support replacement of the 'strndup()' function.  When so \
configured, it provides an implementation of 'strndup()'."
#elif defined (USE_DMALLOC)
#define DMALLOC_DISABLE
#include <dmalloc.h>
#warning "The DMALLOC package provides an implementation of 'strndup()'."
#else
#error "Neither USE_BDWGC nor USE_DMALLOC is defined"
#endif

char* strndup (const char* s1, size_t n)
{
#if defined (TRACE)
#if defined (USE_BDWGC)
  char* result = GC_STRNDUP (s1, n);
#elif defined (USE_DMALLOC)
  char* result = dmalloc_strndup (NULL, 0,
				  s1,
				  n, /* max length */
				  0 /* no xalloc messages */
				 );
#endif
  if (result != NULL)
     fprintf (stdout, "strndup (\"%s\" [@%#lx], %lu)=\"%s\" [@%#lx]\n",
		      s1, s1, n, result, result);
  else
     fprintf (stdout, "strndup (\"%s\" [@%#lx], %lu)=ENOMEM\n",
		      s1, s1, n);
  return result;
#else
#if defined (USE_BDWGC)
  return GC_STRNDUP (s1, n);
#elif defined (USE_DMALLOC)
  return dmalloc_strndup (NULL, 0,
			  s1,
			  n, /* max length */
			  0 /* no xalloc messages */
			 );
#endif
#endif
}
