# $Id: clean-am.awk 2436 2022-02-09 02:53:04Z teshields $

# This file is part of Ox, an Attribute Grammar Compiling System
# that augments the specification functionality of Lex and Yacc, and
# input specification compatible lexer and parser generators.
#
# Copyright (C) 2019-2022  Thomas E. Shields
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

/^Makefile: \$\(srcdir\)\/Makefile\.in \$\(top_builddir\)\/config\.status$/ {
   getline # this is a build rule, skip it and all of the script
   while (index($0,"\t") == 1) {
      getline # line is part of the script, skip it and get the next line
   }
}
/^\$\(top_builddir\)\/config\.status: \$\(top_srcdir\)\/configure \$\(CONFIG_STATUS_DEPENDENCIES\)$/ {
   getline # this is a build rule, skip it and all of the script
   while (index($0,"\t") == 1) {
      getline # line is part of the script, skip it and get the next line
   }
}
/^\$\(top_srcdir\)\/configure: @MAINTAINER_MODE_TRUE@ \$\(am__configure_deps\)$/ {
   getline # this is a "maintainer-mode" build rule, skip it & all of the script
   while (index($0,"\t") == 1) {
      getline # line is part of the script, skip it and get the next line
   }
}
/^\$\(ACLOCAL_M4\): @MAINTAINER_MODE_TRUE@ \$\(am__aclocal_m4_deps\)$/ {
   getline # this is a "maintainer-mode" build rule, skip it & all of the script
   while (index($0,"\t") == 1) {
      getline # line is part of the script, skip it and get the next line
   }
}
/^\$\(am__aclocal_m4_deps\):$/ {
   getline # this is a build rule, skip it and all of the script
   while (index($0,"\t") == 1) {
      getline # line is part of the script, skip it and get the next line
   }
}
/^\$\(srcdir\)\/config_h\.in: @MAINTAINER_MODE_TRUE@ \$\(am__configure_deps\) $/ {
   getline # this is a "maintainer-mode" build rule, skip it & all of the script
   while (index($0,"\t") == 1) {
      getline # line is part of the script, skip it and get the next line
   }
}
/@MAINTAINER_MODE_TRUE@/ {
   # strip "maintainer-mode" only contents from pattern to end of line
   sub(/@MAINTAINER_MODE_TRUE@.*$/,"",$0)
   print $0
   getline # get the next line
   while (index($0,"\t") == 1) {
      getline # line is part of the script, skip it and get the next line
   }
}
/^am__aclocal_m4_deps = \$\(top_srcdir\)\/m4\/stdbool\.m4 .$/ {
   while (index($0,"\\") == length($0)) {
      getline # line is continued, skip it and get the next line
   }
   next # skip the (last) line (of the continuation)
}
/^am__configure_deps = \$\(am__aclocal_m4_deps\) \$\(CONFIGURE_DEPENDENCIES\) .$/ {
   while (index($0,"\\") == length($0)) {
      getline # line is continued, skip it and get the next line
   }
   next # skip the (last) line (of the continuation)
}
/^ACLOCAL = @ACLOCAL@/ {
   next # skip the line
}
/^AUTOCONF = @AUTOCONF@/ {
   next # skip the line
}
/^ACLOCAL_M4 = \$\(top_srcdir\)\/aclocal\.m4$/ {
   next # skip the line
}
/^\$\(top_srcdir\).*\/include\.am.*\$\(am__empty\):$/ {
   next # skip the line
}
/^@MAINTAINER_MODE_TRUE@/ {
   next # this is a "maintainer-mode" only line, skip it
}
/^@MAINTAINER_MODE_FALSE@\t\$\(AM_V_at\)\(test -e \$\(distdir\)\/configure\.ac.*\|\| true$/ {
   next # this is a "remove maintainer-mode only files" line, skip it
}
/^@MAINTAINER_MODE_FALSE@\t\$\(AM_V_at\)\(test -e \$\(distdir\)\/.*\.am.*\|\| true$/ {
   next # this is a "remove maintainer-mode only files" line, skip it
}
/^@MAINTAINER_MODE_FALSE@\t\$\(AM_V_at\)\(test -e \$\(distdir\)\/.*\.awk.*\|\| true$/ {
   next # this is a "remove maintainer-mode only files" line, skip it
}
/^@MAINTAINER_MODE_FALSE@\t\$\(AM_V_at\)\(test -e \$\(distdir\)\/.*\.in.*\|\| true$/ {
   next # this is a "remove maintainer-mode only files" line, skip it
}
/^@MAINTAINER_MODE_FALSE@\t\$\(AM_V_at\)\(test -e \$\(distdir\)\/.*\.m4.*\|\| true$/ {
   next # this is a "remove maintainer-mode only files" line, skip it
}
/\$\(srcdir\)\/Makefile\.am/ {
   # strip pattern from line
   sub(/\$\(srcdir\)\/Makefile\.am */,"",$0)
}
/\$\(am__configure_deps\)/ {
   # strip pattern from line
   sub(/\$\(am__configure_deps\) */,"",$0)
}
/\$\(top_srcdir\)\/dist-hook\.am/ {
   # strip pattern from line
   sub(/\$\(top_srcdir\)\/dist-hook\.am */,"",$0)
}
/\$\(top_srcdir\)\/([^\/]+\/)*include\.am/ {
   # strip all pattern occurrences from line
   gsub(/\$\(top_srcdir\)\/([^\/]+\/)*include\.am */,"",$0)
}
/ clean-am\.awk/ {
   # strip pattern from line
   sub(/ clean-am\.awk/,"",$0)
}
{
   print # pass the line through
}
