#!/bin/bash
shopt -s extglob nullglob

# $Id: run-GPPL-test.sh 2379 2022-01-07 22:41:51Z teshields $

# This file is part of Ox, an Attribute Grammar Compiling System
# that augments the specification functionality of Lex and Yacc, and
# input specification compatible lexer and parser generators.
#
# Copyright (C) 2017-2022  Thomas E. Shields
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

DIFF_CMD="${DIFF}"
case "$build_os" in
    cygwin*)
	host_os="cygwin" ;;
    darwin*)
	host_os="macOS" ;;
    linux-gnu*)
	host_os="linux" ;;
    mingw*)
	DIFF_CMD="${DIFF} --strip-trailing-cr"
	host_os="mingw" ;;
    msys*)
	host_os="msys" ;;
    solaris*)
	host_os="sunOS" ;;
    *)
	host_os="unknown" ;;
esac

if [ -x ${GPPL} ]
then # run the `${GPPL}' executable on this test case

     subdir=${1##*/}
     if [ -d ${srcdir}/$subdir ]
     then # compile & test ${GPPL} generated code

	  # clean out everything first
	  ${MAKE} --directory=$subdir ${AM_MAKEFLAGS} clean > /dev/null 2>&1

	  # compile using ${GPPL} compiler
	  if ${MAKE} --directory=$subdir ${AM_MAKEFLAGS} gc${EXEEXT} > $subdir/makerr 2>&1
	  then if [ -s $subdir/makerr ]
	       then echo "Messages making in $PWD/$subdir using ${GPPL}:"
		    echo ""
		    cat $subdir/makerr
		    echo ""
	       fi
	       rm -f $subdir/makerr
	  else echo "Error(s) making in $PWD/$subdir using ${GPPL}:"
	       echo ""
	       cat $subdir/makerr
	       echo ""
	       rm -f $subdir/makerr
	       exit 1
	  fi

	  # check for changes in the ${GPPL} generated code (should be none)
	  if [ -f ${srcdir}/$subdir/g.c.bak ]
	  then failed=0
	       # compare the baseline file with the file just generated 
	       echo -n "Checking $PWD/$subdir/g.c: "
	       ${DIFF_CMD} ${srcdir}/$subdir/g.c.bak $subdir/g.c > $subdir/g.c.diff
	       [ $? -eq 2 ] && exit 2
	       if [ -s $subdir/g.c.diff ]
	       then echo "changes:"
		    cat $subdir/g.c.diff
		    echo ""
		    failed=1
	       else echo "no change"
	       fi
	       rm -f $subdir/g.c.diff
	       [ "$failed" = "1" ] && exit 1
	  else echo "Missing master copy ${srcdir}/$subdir/g.c.bak"
	       exit 1
	  fi

	  if [ -x $subdir/gc${EXEEXT} ]
	  then # run the `gc' executable on any test cases present
	       if [ -d ${srcdir}/$subdir/gct ]
	       then mkdir -p $subdir/gct
		    failed=0
		    for j in `echo ${srcdir}/$subdir/gct/t+([0-9])` ; do
		       k=${j##*/}
		       $subdir/gc${EXEEXT} < $j > $subdir/gct/$k.o 2>&1
		       if [ -f $j.ob ]
		       then # compare the expected output with the file just written
			    echo -n "Checking $PWD/$subdir/gct/$k.o: "
			    ${DIFF_CMD} $j.ob $subdir/gct/$k.o > $subdir/gct/$k.diff
			    res=$?
			    if [ $res -ne 0 ]
			    then
				if [ -f $j.$host_os.ob ]
				then
				    ${DIFF_CMD} $j.$host_os.ob $subdir/gct/$k.o > $subdir/gct/$k.diff
				else
				    [ $res -eq 2 ] && exit 2
				fi
			    fi
			    [ $? -eq 2 ] && exit 2
			    if [ -s $subdir/gct/$k.diff ]
			    then echo "changes:"
				 cat $subdir/gct/$k.diff
				 echo ""
				 failed=1
			    else echo "no change"
			    fi
			    rm -f $subdir/gct/$k.diff
		       else echo "Missing master copy $j.ob"
			    failed=1
		       fi
		    done
		    exit $failed
	       fi
	  fi
     else
	  echo "Error invoking $0"
	  exit 99 # hard error!
     fi
else
     echo "${GPPL} compiler missing"
     exit 99 # hard error!
fi
