%language "C++"
%require "3.8.2"
%skeleton "glr2.cc"

%glr-parser

%define lr.type ielr
%define lr.default-reduction consistent

%define api.token.raw

//%define parse.assert
%define parse.trace
%define parse.error custom

%locations

%initial-action
{
  @$.initialize (&theFiles.inName);

  yoxOutL.setCurPos (@$.begin);

  returnSTMT.setCurPos (@$.begin);
  returnSTMTtrailer.setCurPos (@$.begin);

  separatorMARK.setCurPos (@$.begin);
}

%header
%define api.namespace {AGC}
%define api.parser.class {lexer}
%define api.location.file "location.hh"
%define api.location.include {"location.hh"}

%parse-param {AGC::LFILE::Lexer& theLexer}
%parse-param {yoxControl& ctrl}
%parse-param {specIO& theFiles}


%verbose

%{
/* $Id: agc-Lfile.yy 2842 2022-12-13 01:19:44Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 1992-1994  Kurt M. Bischoff
   Copyright (C) 2004-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */
%}

%code top
{
// hack to avoid a 'yydebug' global in every parser
#define yydebug agc_Lfile_debug
}

%code requires
{
#include "lextypes.hh"
// attrbTriple:     lexical representation of an attribute occurrence
// numberRep:       lexical representation of a number

#include "semtypes.hh"
// attrbPair:       semantic representation of an attribute occurence
// attrbRefSeqElem: attribute occurrence w/location
// attrbRefSeq:     sequence of occurrences in an attribute definition

#include "scan.hh"

namespace AGC
{
namespace LFILE
{
class Lexer;
}
}
#include "agc-Lfile.hh"
}

%code provides
{
#if defined(WITH_REFLEX)
#include <reflex/input.h>
#endif
// called from 'main ()'
extern "C" void
agc_lexer_parse (
#if defined(WITH_REFLEX)
                 const reflex::Input&,
#elif defined(WITH_FLEX)
                 std::ifstream&,
#else
#error "missing -DWITH_FLEX or -DWITH_REFLEX compile option"
// assume -DWITH_FLEX to avoid cascading errors
                 std::ifstream&,
#endif
                 Lfile_specIO&,
                 int);
}

%define api.value.type variant



// Ox tokens: common to lexer & parser specs

%token                OX_LBRACE      "@{"
%token                OX_RBRACE      "@}"
                                   // @<grammar symbol>(.<n>)?.<attribute name>@
%token <attrbTriple>  OX_ATTRB_REF   "attribute reference"
                                   // grammar symbol name, C/C++ identifier,
                                   // Ox macro name, Ox attribute name
%token <name_literal> ID             "identifier"

%token <char_literal> LITERAL_CHAR   "character literal"
%token <char_literal> LITERAL_STRING "string literal"



// non-alpha character tokens common to lexer & parser specs

%token                LBRACE         "{"
%token                RBRACE         "}"
%token                SEMICOLON      ";"



// Lex/Yacc & Lex/Yacc-extension reserved tokens

%token                MARK           "%%"
%token                LCURL          "%{"
%token                RCURL          "%}"



// Ox tokens: lexer specs

%token <name_literal> TERM_ID        "terminal symbol"
%token <char_literal> TERM_LITERAL   "terminal literal"



// lexer: C/C++ keyword tokens

%token                RETURN         "return"



// lexer: miscellaneous tokens

%token                NEWLINE        "\\n"
%token                SOSF           "start of generic scanner source file"
%token                NO_ATTRBREFSEC "no attribute reference section"



// Ox tokens: parser specs
                                    // <grammar symbol>(.<n>)?.<attribute name>
%token <attrbTriple>  OX_ATTRB_OCCUR "attribute occurrence"
%token <keyword>      OX_ATTRIBUTES  "@attributes"
%token <keyword>      OX_MACRO       "@macro"
%token                OX_END         "@end"
%token <keyword>      OX_TRAV        "@traversal"
%token                OX_PREORDER    "@preorder"
%token                OX_POSTORDER   "@postorder"
%token                OX_LEFTTORIGHT "@lefttoright"
%token                OX_RIGHTTOLEFT "@righttoleft"
%token                OX_REVORDER    "@revorder"
%token                OX_REVDIR      "@revdirection"
%token                OX_DISABLE     "@disable"
%token <keyword>      OX_AUTOSYN     "@autosyn"
%token <keyword>      OX_AUTOINH     "@autoinh"
%token <keyword>      OX_WARN        "@warn"
%token                OX_I           "@i"
%token                OX_M           "@m"
%token                OX_E           "@e"
%token                OX_DECORATE    "@decorate"
%token                OX_VOID        "@void"
                                   // <traversal name>
%token <name_literal> OX_ID          "traversal mode annunciator"



// parser: non-alpha character tokens

%token                COMMA          ","
%token                LPAREN         "("
%token                RPAREN         ")"
%token                LBRACKET       "["
%token                RBRACKET       "]"
                                   // MSTA named reference
%token                AT             "@"
%token                ELLIPSIS       "..."
                                   // MSTA charset specifications
%token                LTARROW        "<-"
%token                DBLARROW       "<->"
                                   // Bison predicate action
%token                PREDICATE      "%?"
                                   // grammar symbol, (dotted) name
%token <name_literal> GS_ID          "nonterminal symbol"
%token <name_literal> NAMESPACE_ID   "C++ qualified name"
%token <char_literal> LITERAL_PATH   "literal path name"
                                   // $ID or $NUM (BtYacc, BYacc)
%token <name_literal> INHATTR_ID     "inherited attribute"
%token <name_literal> API_VARIABLE   "bison API variable"
%token <name_literal> API_VALUE      "bison API value"
%token <numberRep>    NUM            "number" // [0-9]+



// parser: Yacc & Yacc-extension reserved words

%token <keyword>      LEFT           "%left"
%token <keyword>      RIGHT          "%right"
%token <keyword>      NONASSOC       "%nonassoc"
%token <keyword>      TOKEN          "%token"
%token                PREC           "%prec"
%token <keyword>      TYPE           "%type"
%token <keyword>      NTERM          "%nterm" // Bison
%token <keyword>      START          "%start"
%token <keyword>      UNION          "%union"
%token                EMPTY          "%empty" // Bison



// parser: Yacc-extension directives

%token <keyword>      EXPECT         "%expect"      // Bison, BYacc, MSTA
%token <keyword>      EXPECT_RR      "%expect-rr"   // Bison, BYacc
%token <keyword>      PURE_PARSER    "%pure-parser" // Bison, BYacc
%token                DPREC          "%dprec"       // Bison GLR parsing
%token                MERGE          "%merge"       // Bison GLR parsing
%token <keyword>      PRECEDENCE     "%precedence"  // Bison
%token                LOOKAHEAD      "%la"          // MSTA
%token <keyword>      SCANNER        "%scanner"     // MSTA (not supported)



// parser: Yacc-extension composite directives

%token <keyword>      CODE         // %code, %initial-action   // Bison, BYacc
%token <keyword>      LOCATION     // %location, %position, %@ // BtYacc
%token <keyword>      DESTRUCTOR     "%destructor"             // Bison, BYacc
%token <keyword>      LEX_PARAM      "%lex-param"              // Bison, BYacc
%token <keyword>      PARSE_PARAM    "%parse-param"            // Bison, BYacc
%token <keyword>      DEFINE         "%define"                 // Bison
%token <keyword>      LANGUAGE       "%language"               // Bison
%token <keyword>      NAME_PREFIX    "%name-prefix"            // Bison
%token <keyword>      PARAM          "%param"                  // Bison
%token <keyword>      PRINTER        "%printer"                // Bison
%token <keyword>      SKELETON       "%skeleton"               // Bison
%token <keyword>      LOCAL          "%local"                  // MSTA
%token <keyword>      EXPORT         "%export"                 // MSTA
%token <keyword>      IMPORT         "%import"                 // MSTA



// parser: C/C++ keyword tokens

%token <keyword>      CONST          "const"
%token <keyword>      VOLATILE       "volatile"
%token <keyword>      RESTRICT       "restrict" // C99
%token <keyword>      REGISTER       "register"
%token <keyword>      STATIC         "static"
%token <keyword>      TYPE_KEYWORD            // C/C++ type keywords
%token <keyword>      ELABORATED_TYPE_KEYWORD // C/C++ composite type keywords



// parser: C/C++ operator tokens

%token                PERIOD         "."
%token                RTARROW        "->" // & MSTA charset specifications
%token                PLUSPLUS       "++"
%token                MINUSMINUS     "--"
%token                TILDE          "~"
%token                EXCLAMATION    "!"
%token                MINUS          "-"
%token                PLUS           "+"
%token                AMPERSAND      "&" // & C/C++ type specification
%token                ASTERISK       "*" // & C/C++ type specification
%token                SLASH          "/"
%token                PERCENT        "%"
%token                LTSHIFT        "<<"
%token                RTSHIFT        ">>"
%token                LESSTHAN       "<" // & Yacc type tag, Bison '%merge',
                                         //   C++ template types
%token                GTRTHAN        ">" // & Yacc type tag, Bison '%merge',
                                         //   C++ template types
%token                LTOREQ         "<="
%token                GTOREQ         ">="
%token                EQUAL          "=="
%token                NOTEQUAL       "!="
%token                UPARROW        "^"
%token                VBAR           "|"
%token                LOGAND         "&&"
%token                LOGOR          "||"
%token                QUESTION       "?"
%token                COLON          ":" // & C/C++ type specification
%token                ASSIGN         "="
%token                PLUSEQ         "+="
%token                MINUSEQ        "-="
%token                TIMESEQ        "*="
%token                DIVEQ          "/="
%token                REMEQ          "%="
%token                LTSHIFTEQ      "<<="
%token                RTSHIFTEQ      ">>="
%token                BITANDEQ       "&="
%token                BITXOREQ       "^="
%token                BITOREQ        "|="



// parser: C++ operator tokens

%token                PTRTOMEMBER    "->*"
%token                MEMBERSELECT   ".*"
%token                SCOPE          "::"



// parser: miscellaneous tokens

%token                END 0          "end of file"



%printer
{
  yyo << "\"" + $$ + "\"";
} <name_literal>
%printer
{
  yyo << "\"" + $$ + "\"";
} <char_literal>
%printer
{
  yyo << "\"" + $$ + "\"";
} <keyword>

%start                AG_spec

%type <name_literal>  terminal_symbol

%code
{
#include "config.h"

#include <iomanip>
#include <ostream>
#include <sstream>

#include "codegen.hh"
#include "ctc.hh"
#include "depGs.hh"
#include "fileMgr.hh"
#include "joinTabs.hh"
#include "messages.hh"
#include "opts.hh"
#include "outMgr.hh"
#include "scan.hh"
#include "symTrans.hh"
#include "treeviz.hh"


static int nested = 0;

#include "agc-Lfile-lexer.hh"

// override Bison default
#define yylex theLexer.yylex
}

%%
AG_spec
  :   lexer_spec
  ;

expr_value
  :   "identifier"
  |   LITERAL_CHAR
  |   LITERAL_STRING
  |   "number"
  ;

expr_operator // C/C++ operators
  :   "."
  |   "->"
  |   "++"
  |   "--"
  |   "~"
  |   "!"
  |   "-"
  |   "+"
  |   "&"
  |   "*"
  |   "/"
  |   "%"
  |   "<<"
  |   ">>"
  |   "<"
  |   ">"
  |   "<="
  |   ">="
  |   "=="
  |   "!="
  |   "^"
  |   "|"
  |   "&&"
  |   "||"
  |   "?"
  |   ":"
  |   "="
  |   "+="
  |   "-="
  |   "*="
  |   "/="
  |   "%="
  |   "<<="
  |   ">>="
  |   "&="
  |   "^="
  |   "|="
  |   ","
  |   "->*"
  |   ".*"
  |   "::"
  ;

lexer_spec
  :   // Lex specification lexer
      "%%"
      {
        yoxOutL.gen ().put ("\n%{\n");

        if (!ctrl.buildHeaderFile &&
            ctrl.targetParserGen != bisonParserGen)
          {
            // append to LexSpec if no OxHeader
            yoxOutL.append (yoxOutH);
            yoxOutL.append (yoxOutHrequires);
            yoxOutL.append (yoxOutHprovides);
          }

        yoxOutL.skeletonText4 ();

        yoxOutL.gen ().put ("%}\n");

        yoxOutL.append (separatorMARK); // holds the 1st "%%" line
        separatorMARK.clear ();

        yoxOutL.gen ()
            .put ("\n%{\n"
                  "\t" +
                  yoxPrefix ("init") + "(); "
                  "/* initialize the Ox-generated evaluator */\n%}\n");
      }
      lexer_rule_seq MARK_opt
  |   // Generic lexer
      SOSF
      {
        nested = 0;

        if (!ctrl.buildHeaderFile &&
            ctrl.targetParserGen != bisonParserGen)
          {
            // append to LexSpec if no OxHeader
            yoxOutL.append (yoxOutH);
            yoxOutL.append (yoxOutHrequires);
            yoxOutL.append (yoxOutHprovides);
          }

        yoxOutL.skeletonText4 ();
      }
      generic_lexer_rule_seq
  ;

MARK_opt
  :   %empty
  |   "%%"
  ;

lexer_rule_seq[left]
  :   %empty
      {
        @left = yyla.location;
      }
  |   lexer_rule_seq[right]
      {
        nested = 0;
      }
      lexer_action_opt "\\n"[NEWLINE]
  ;

lexer_action_opt
  :   %empty
  |   lexer_action
  |   "%{" generic_lexer_rule_seq "%}" // 'flex'
  |   "%{" error "%}"
  |   "%{" /* empty */ "%}"
  ;

lexer_action
  :   return_action attrb_ref_section
      {
        if (!retObjectSeen && atLeastOneSymHasAttrbs)
          ambiguousReturn (@return_action);

        if (prodBufferPointer2 == 0)
          {
            // no return expression or no terminal symbols in the
            // return expression
            putSymbolInProdBuffer2 (omittedRetSym);
            yoxOutL.putGenLeafCall (omittedRetSym);
            initLDG ();
          }

        finishLfileAttrbRefSection (@lexer_action);

        // end the generated 'return' semantic action code
        yoxOutL.gen ().put ("}");

        yoxOutL.append (returnSTMT);
        returnSTMT.clear ();

        // end the generated 'return' lexer action
        if (theFiles.scanner == LEX_SPEC_LEXER && nested == 0)
          {
            // hack: force a "#line NN <inName>" before the '}', pretending that
            //       the '}' is from the lexer source, to avoid a bare
            //       "#line NN <inName>" outside of the Lex action block
            yoxOutL.lex ().put ("}");
          }
        else
          yoxOutL.gen ().put ("}");

        yoxOutL.append (returnSTMTtrailer);
        returnSTMTtrailer.clear ();

        if (prodBuffer2[prodBufferPointer2 - 1].lexeme == omittedRetSym &&
            atLeastOneSymHasAttrbs)
          unknownNodeType (@return_action);
      }
  |   "{"
      {
        nested++;
      }
      generic_lexer_rule_seq
      {
        nested--;
      }
      "}"
  |   "{" error "}"
  |   "{" /* empty */ "}"
  ;

generic_lexer_rule_seq
  :   lexer_action
  |   generic_lexer_rule_seq lexer_action
  ;

return_action
  :   return_symbol ";"
  |   return_symbol return_expr ";"
  ;

return_symbol
  :   "return"
      {
        // start the generated 'return' semantic action
        const std::string ifhdr = "{if(" + yoxPrefix ("Yok") + "){";

        if (theFiles.scanner == LEX_SPEC_LEXER && nested == 0)
          yoxOutL.lex ().put (ifhdr);
        else
          yoxOutL.gen ().put (ifhdr);

        retObjectSeen = false;
        multipleReturnsIssued = false;

        prodBufferPointer2 = 0;
        yaccSemIndex = 0; // 0 for LHS; ignored
      }
  ;

return_expr[left] // visible components in a return expression (possibly with
                  // multiple parser "tokens" being returned)

      // process a sequence of tokens legal in an expression: identifiers,
      // literals, operators and nested '()' pairs

      // NOTE: better syntax errors requires C/C++ expression syntax

  :   return_component
  |   return_expr[right] return_component
  ;

return_component
  :   terminal_symbol
      {
        const std::string& terminal_symbol = $terminal_symbol;
        putSymbolInProdBuffer2 (terminal_symbol);
        yoxOutL.putGenLeafCall (terminal_symbol);
        initLDG ();
      }
  |   expr_operator
  |   expr_value
  |   "(" return_expr ")"
  |   error
      {
        yyclearin;
      }
  ;

terminal_symbol
  :   "terminal symbol"[TERM_ID]
      {
        $terminal_symbol = $TERM_ID;
      }
  |   "terminal literal"[TERM_LITERAL]
      {
        $terminal_symbol = $TERM_LITERAL;
      }
  ;

attrb_ref_section
  :   %empty
      {
        @attrb_ref_section = AGC::location (@0.end); // @return_action.end
      }
  |   NO_ATTRBREFSEC
      {
        @attrb_ref_section = AGC::location (@0.end); // @return_action.end
      }
  |   "@{" attrb_stmt_seq "@}"
  ;

attrb_stmt_seq[left] // visible tokens in a statement sequence

      // process a sequence of tokens legal in a statement sequence
      // w/ expressions: identifier (includes C/C++ keyword), literal,
      // attribute reference, operator (includes ','), nested '()' pair,
      // nested '{}' pair and ';'

      // NOTE: better syntax errors requires C/C++ statement/expression syntax

  :   %empty
      {
        @left = yyla.location;
      }
  |   attrb_stmt_seq_component attrb_stmt_seq[right]
  |   error
      {
        yyclearin;
      }
  ;

attrb_stmt_seq_component
  :   attrb_stmt_component
//|   attrb_compound_stmt
  |   ";"
  ;

// Ox restriction: each attribute occurrence in a L-file attribute reference
// section can only be referred to once as a C/C++ l-value, so nested compound
// statement syntax is not supported

//attrb_compound_stmt
//  :   "{" attrb_stmt_seq "}"
//  ;

attrb_stmt_component
  :   OX_ATTRB_REF
      {
        attrbPair ref =
            transAttrbTripleToAttrbPair (@OX_ATTRB_REF, $OX_ATTRB_REF);

        if (ref.num == NIOsymAttrbNum)
          {
            // NIO pseudo attribute reference
            if (ctrl.genTreeViz)
              yoxOutL.putImmediateAttrbRef (ref, false);
          }
        else
          {
            const attrbTriple& attref = $OX_ATTRB_REF;

            if (memberDepend (ref, true, true))
              yox_error (@OX_ATTRB_REF,
                         "multiple definitions for "
                         "\"" +
                             attref.gsName + "." +
                             std::to_string (attref.occurrence) + "." +
                             attref.atName +
                         "\".\n");
            yoxOutL.putImmediateAttrbRef (ref, false);
            processAdepSpec ();
          }
      }
  |   attrb_expr_component
  ;

attrb_expr[left] // visible tokens in an expression

      // process a sequence of tokens legal in an expression: identifier,
      // literal, operator (includes ',') and nested '()' pair

      // NOTE: better syntax errors requires C/C++ expression syntax

  :   %empty
  |   attrb_expr[right] attrb_expr_component
  ;

attrb_expr_component
  :   expr_operator
  |   expr_value
  |   "(" attrb_expr ")" // sub-expression & parameter list
  ;
%%

// called from 'main ()'
extern "C" void
agc_lexer_parse (
#if defined(WITH_REFLEX)
                 const reflex::Input& yoxIn,
#else // defined(WITH_FLEX)
                 std::ifstream& yoxIn,
#endif
                 Lfile_specIO& Lfile,
                 int debug)
{
  AGC::LFILE::Lexer agc_Lfile_lexer (Lfile, yoxIn);
  agc_Lfile_lexer.set_debug (debug >> 1);
  AGC::lexer parser (agc_Lfile_lexer, outMgr::ctrl, Lfile);
  parser.set_debug_level (debug & 0x1);
  (void)parser ();
}

void
AGC::lexer::error (const AGC::location& locp, const std::string& msg)
{
  agc_syntax_error (locp, msg);
}

void
AGC::lexer::report_syntax_error (const context& yyctx) const
{
  // # of expected tokens
  enum
  {
    YYARGS_MAX = 4
  };

  symbol_kind_type yyarg[YYARGS_MAX] = { symbol_kind::S_YYEMPTY };

  // Actual size of yyarg[]
  int yycount = yyctx.expected_tokens (yyarg, YYARGS_MAX);

  if (yycount < 0)
    return;
  if ((yycount == 0) && (yyarg[0] != (symbol_kind_type)symbol_kind::S_YYEMPTY))
    yycount = YYARGS_MAX;

  const std::string token_name = std::string (symbol_name (yyctx.token ()));

  std::ostringstream yymsg;

  yymsg << "syntax error";

  if (yycount > 0)
    {
      yymsg << ", unexpected ";

      switch (yyctx.token ())
        {
        case symbol_kind::S_ID:             // identifier
        case symbol_kind::S_TERM_ID:        // Yacc terminal symbol
          yymsg << "\'" + token_name + "\' "
                       "[" +
                       yyctx.lookahead ().value.as<name_literal> () +
                       "]";
          break;

        case symbol_kind::S_LITERAL_CHAR:
        case symbol_kind::S_LITERAL_STRING:
        case symbol_kind::S_TERM_LITERAL:   // literal terminal grammar token
          yymsg << "\'" + token_name + "\' "
                       "[" +
                       yyctx.lookahead ().value.as<char_literal> () +
                       "]";
          break;

        case symbol_kind::S_OX_ATTRB_REF:   // attribute reference
          yymsg << "\'" + token_name + "\' "
                       "[" +
                       yyctx.lookahead ().value.as<attrbTriple> ().gsName +
                       "." +
                       std::to_string (yyctx.lookahead ()
                                           .value.as<attrbTriple> ()
                                           .occurrence) +
                       "." +
                       yyctx.lookahead ().value.as<attrbTriple> ().atName +
                       "]";
          break;

        default:
          yymsg << "\'" + token_name + "\'";
          break;
        }

      std::string trailing_list;
      for (int i = yycount; --i > 0;)
        {
          std::string temp;

          if (i + 1 == yycount)
            temp = " or";
          else
            temp = ",";

          trailing_list =
              temp + " \'" + symbol_name (yyarg[i]) + "\'" + trailing_list;
        }

      yymsg << ", expecting \'" << symbol_name (yyarg[0]) << "\'"
            << trailing_list;
    }
  yymsg << ".";

  AGC::location loc = yyctx.location ();
  agc_syntax_error (loc, yymsg);
}
