%start YACC_DEFS START_SYM YACC_UNION_HD YACC_UNION
%start YACC_CCODE CCODE_BLOCK CCODE_IMPORT
%start INT_ARG STRING_ARG BISONDEF TYPESPEC USER_ARGS
%start YACC_RULES INHERIT_ATTR YACC_ACT
%start YACC_USER YACC_TRAILER
%start BTYACC_ACT
%start ACTIONCLOSER_DIRECTIVE YOXOUTG_DIRECTIVE
%start COMMENT SKIP_COMMENT
%start STRING
%start CPP_DIRECTIVE

%start MACRO_HEAD MACRO_BODY
%start ATTR_DECL OXVOID ATTR_DECL_SYMS
%start TRAV_DECL
%start EXPAND_MACRO MACRO_ARG
%start YATTR_REF ATTR_DEPS

%option c++

%option never-interactive stack

%option noyywrap
%option noyy_top_state
%option nounput

%{
/* $Id: agc-Yfile-lexer.ll 2834 2022-12-10 21:33:22Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 1992-1994  Kurt M. Bischoff
   Copyright (C) 2004-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include "config.h"

#include <string>

#include "depGs.hh"
#include "joinTabs.hh"
#include "lextypes.hh"
#include "symTrans.hh"
#include "trav.hh"
#include "type-keywords.hh"
#include "yox-directives.hh"

#include "YY_LOCATIONS.hh"
%}

%top{
#include <fstream>

#include "agc-Yfile.hh"
#include "mac.hh"
#include "macutil.hh"
#include "messages.hh"
#include "opts.hh"
#include "outMgr.hh"
#include "scan.hh"

#define YY_DECL \
  int AGC::YFILE::Lexer::yylex (AGC::parser::value_type& yylval, \
                                AGC::location& yylloc)

// inhibit the Flex builtin definition of 'struct yy_buffer_state'
#define YY_STRUCT_YY_BUFFER_STATE
}

%{
#include "agc-Yfile-lexer.hh"
%}

WS		[ \t\f\v]+
OPTWS		[ \t\f\v]*
NON_WS		[^ \t\f\v\n]

ESCSEQ		\\([^\n]|[0-7]{1,3}|x[[:xdigit:]]{1,2})
CHAR_LIT	"'"([^\\'\n]|{ESCSEQ})"'"
STRING_LIT	"\""([^\\"\n]|{ESCSEQ})*"\""

CHAR_CONST	"L"?{CHAR_LIT}
STRING_CONST	"L"?{STRING_LIT}

ID		[a-zA-Z_][a-zA-Z_0-9]*
DOT_NAME	[_\.]*[a-zA-Z][-_\.a-zA-Z0-9]*

INTEGER		[0-9]+
HEX		"0"[xX][0-9a-fA-F]+
OCTAL		"0"[0-7]+
INT_SUFFIX	([uU][lL]{0,2})|([lL]{1,2}[uU]?)
NUMBER		({INTEGER}|{HEX}|{OCTAL}){INT_SUFFIX}?

SYMBOL		{DOT_NAME}|{CHAR_LIT}|{STRING_LIT}
ATTRB_OCCUR	{SYMBOL}{OPTWS}"."({OPTWS}{INTEGER}(OPTWS)".")?{OPTWS}{ID}
ATTRB_REF	"@"{OPTWS}{ATTRB_OCCUR}({OPTWS}"@")?

NAMESPACE_ID	({ID})?({OPTWS}"::"{OPTWS}{ID})+

DIR_KEYWORD	"%"[a-z][a-z_\-]*
LEGACY_KEYWORD	"%"[<>02@]
OX_KEYWORD	"@"[a-z]+
YOX_DIRECTIVE	{DIR_KEYWORD}|{LEGACY_KEYWORD}|{OX_KEYWORD}

FILE_NAME	[^/\n]
FILE_PATH	"<"((\.|\.\.|{FILE_NAME})"/"+)*{FILE_NAME}">"

/* Bison: '%define' <api-name> */
API_NAME	[a-zA-Z_\-][a-zA-Z_\-0-9\.]*

PVAR_NAME	("$")|("-"?[0-9]+)|([a-zA-Z_][a-zA-Z_0-9]*)|\[{ID}\]
PVAR_REF	"$"(<{ID}>)?{PVAR_NAME}
PLOC_REF	"@"{PVAR_NAME}

CPP_DIRECTIVE	^{OPTWS}"#"{OPTWS}("define"|"if"|"ifdef"|"ifndef"|"else"|"elif"|"endif").*$

%%

<YACC_DEFS>{

  /* Y-file Definitions Section */

"/*" {
  yoxOutGdefs.lex ().put (text ());
  yox_push_state (COMMENT, yoxOutGdefs);
}

{CHAR_LIT} {
  yoxOutGdefs.lex ().put (text ());

  const std::string& lexeme (text ());
  yylval.emplace<char_literal> (stringLookup (lexeme, true));
  return (parser::token::LITERAL_CHAR);
}

{STRING_LIT} {
  yoxOutGdefs.lex ().put (text ());

  const std::string& lexeme (text ());
  yylval.emplace<char_literal> (stringLookup (lexeme, true));
  return (parser::token::LITERAL_STRING);
}

"L"?"\"" {
  yoxOutGdefs.lex ().put (text ());
  yox_push_state (STRING, yoxOutGdefs);
}

{NUMBER} {
  yoxOutGdefs.lex ().put (text ());

  const std::string& lexeme (text ());
  yylval.emplace<numberRep> (
      numberRep (lexeme, std::stoll (lexeme, nullptr, 0)));
  return (parser::token::NUM);
}

{ID} {
  if (!ctrl.process_autoSynOrInh)
    yoxOutGdefs.lex ().put (text ());

  const std::string& lexeme (text ());
  yylval.emplace<name_literal> (stringLookup (lexeme, true));
  return (parser::token::ID);
}

{DOT_NAME} {
  yoxOutGdefs.lex ().put (text ());

  const std::string& lexeme (text ());
  yylval.emplace<name_literal> (stringLookup (lexeme, true));
  return (parser::token::GS_ID);
}

{YOX_DIRECTIVE} {
  const struct yox_directive* directive =
      in_yox_directive_set (text (), (size_t)size ());
  if (directive == nullptr)
    {
      if ((char)chr () == '%') // found "%<directive>" not important to Ox
        yoxOutGdefs.lex ().put (text ()); // just pass it on through
      else if ((char)chr () == '@')
        {
          // found "@{ID}" -> will be a syntax error
          const std::string& lexeme (text ());
          yylval.emplace<name_literal> (stringLookup (lexeme.substr (1), true));
          return (parser::token::OX_ID);
        }
    }
  else
    {
      yylval.emplace<keyword> (directive->name);

      // special case directive processing

      // copy to output?
      switch (directive->token)
        {
        case parser::token::UNION:
        case parser::token::OX_ATTRIBUTES:
        case parser::token::OX_AUTOSYN:
        case parser::token::OX_AUTOINH:
        case parser::token::OX_WARN:
        case parser::token::OX_MACRO:
        case parser::token::OX_TRAV:
          break;
        case parser::token::IMPORT:
          yoxOutGimport.lex ().put (text ());
          break;
        default:
          yoxOutGdefs.lex ().put (text ());
          break;
        }

      // prep to count '{'?
      switch (directive->token)
        {
        case parser::token::DESTRUCTOR:
        case parser::token::PRINTER:
        case parser::token::CODE:
        case parser::token::EXPORT:
        case parser::token::LOCAL:
        case parser::token::LOCATION:
        case parser::token::IMPORT:
        case parser::token::UNION:
        case parser::token::OX_ATTRIBUTES:
        case parser::token::LEX_PARAM:
        case parser::token::PARSE_PARAM:
        case parser::token::PARAM:
          nLbraces = 0;
        default:
          break;
        }

      // change start condition?
      switch (directive->token)
        {
        case parser::token::DESTRUCTOR:
        case parser::token::PRINTER:
        case parser::token::CODE:
        case parser::token::EXPORT:
        case parser::token::LOCAL:
        case parser::token::LOCATION:
          push_state (yylloc, CCODE_BLOCK);
          break;
        case parser::token::IMPORT:
          // MSTA extension hack for '%import'
          push_state (yylloc, CCODE_IMPORT);
          break;
        case parser::token::START:
          push_state (yylloc, START_SYM);
          break;
        case parser::token::UNION:
          push_state (yylloc, YACC_UNION_HD);
          break;
        case parser::token::LEX_PARAM:
        case parser::token::PARSE_PARAM:
        case parser::token::PARAM:
          push_state (yylloc, USER_ARGS);
          break;
        case parser::token::OX_AUTOSYN:
        case parser::token::OX_AUTOINH:
        case parser::token::OX_WARN:
          break;
        case parser::token::OX_ATTRIBUTES:
          push_state (yylloc, ATTR_DECL);
          break;
        case parser::token::OX_MACRO:
          push_state (yylloc, MACRO_HEAD);
          break;
        case parser::token::OX_TRAV:
          push_state (yylloc, TRAV_DECL);
          break;
        case parser::token::DEFINE:
          push_state (yylloc, BISONDEF);
          break;
        case parser::token::YYEMPTY:
          switch (directive->param_type)
            {
            case C_int:
              push_state (yylloc, INT_ARG); // there should be an int argument
              pass2parser = false; // do not pass it to the parser
              break;
            case C_string:
              push_state (yylloc, STRING_ARG); // there may be a string argument
              pass2parser = false; // do not pass it to the parser
              break;
            case other:
              break;
            }
          break;
        default:
          switch (directive->param_type)
            {
            case C_int:
              push_state (yylloc, INT_ARG); // there should be an int argument
              pass2parser = true; // pass it to the parser
              break;
            case C_string:
              push_state (yylloc, STRING_ARG); // there may be a string argument
              pass2parser = true; // pass it to the parser
              break;
            case other:
              // no special handling for these
              break;
            }
          break;
        }

      // pass token to parser?
      switch (directive->token)
        {
        case parser::token::YYEMPTY:
          break;
        default:
          return (directive->token);
        }
    }
}

^{OPTWS}"%%".*$ {
  yylloc.begin += std::strcspn (text (), "%");
  yylloc.end = yylloc.begin + 2;
  separatorMARK.put (text ());
  start (YACC_RULES);
  return (parser::token::MARK);
}

"%{" {
  if (yaccUnionDefined != 0)
    yaccCODE.lex ().put (text ());
  else
    yoxOutGdefs.lex ().put (text ());
  push_state (yylloc, YACC_CCODE);
  return (parser::token::LCURL);
}

"<" {
  yoxOutGdefs.lex ().put ((char)chr ());
  push_state (yylloc, TYPESPEC);
  nLbraces = 0;
  nLTbraces = 1;
  // normal return from TYPESPEC when nLTbraces == 0, rescan ">"
  return (parser::token::LESSTHAN);
}

">" {
  yoxOutGdefs.lex ().put ((char)chr ());
  return (parser::token::GTRTHAN);
}

"(" {
  yoxOutGdefs.lex ().put ((char)chr ());
  return (parser::token::LPAREN);
}

")" {
  yoxOutGdefs.lex ().put ((char)chr ());
  return (parser::token::RPAREN);
}

"," {
  yoxOutGdefs.lex ().put ((char)chr ());
  return (parser::token::COMMA);
}

"{" {
  yoxOutGdefs.lex ().put ((char)chr ());
  return (parser::token::LBRACE);
}

"}" {
  yoxOutGdefs.lex ().put ((char)chr ());
  return (parser::token::RBRACE);
}

"//".*$ {
  yoxOutGdefs.lex ().put (text ());
}

\n {
  yoxOutGdefs.lex ().put ((char)chr ());
}

{WS} {
  yoxOutGdefs.put (text ());
}

. {
  yoxOutGdefs.put ((char)chr ());
}

<<EOF>> {
  // EOF w/o finding the first "%%": the Y-file is corrupt, so just quit
  unexpectedEOF (yylloc);
}

}



<CCODE_BLOCK>{

  /* YACC_DEFS 'subroutine': echo "{ ... }" C code blocks */

"/*" {
  yoxOutGdefs.lex ().put (text ());
  yox_push_state (COMMENT, yoxOutGdefs);
}

"L"?"\"" {
  yoxOutGdefs.lex ().put (text ());
  yox_push_state (STRING, yoxOutGdefs);
}

"{" {
  yoxOutGdefs.lex ().put ((char)chr ());
  nLbraces++;
  return (parser::token::LBRACE);
}

"}" {
  yoxOutGdefs.lex ().put ((char)chr ());
  if (--nLbraces == 0)
    pop_state ();
  return (parser::token::RBRACE);
}

"//".*$|{WS}|{NUMBER}|{ID}|{CHAR_CONST} {
  yoxOutGdefs.put (text ());
}

\n|. {
  yoxOutGdefs.put ((char)chr ());
}

<<EOF>> {
  unterminatedCode (yylloc, startPos.top ());
}

}



<CCODE_IMPORT>{

  /* YACC_DEFS 'subroutine': echo "{ ... }" C code blocks for MSTA '%import'
     extension hack */

"/*" {
  yoxOutGimport.lex ().put (text ());
  yox_push_state (COMMENT, yoxOutGimport);
}

"L"?"\"" {
  yoxOutGimport.lex ().put (text ());
  yox_push_state (STRING, yoxOutGimport);
}

"{" {
  yoxOutGimport.lex ().put ((char)chr ());
  nLbraces++;
  return (parser::token::LBRACE);
}

"}"{OPTWS}\n? {
  yylloc.end = yylloc.begin + 1;
  yoxOutGimport.lex ().put (text ());
  if (--nLbraces == 0)
    pop_state ();
  return (parser::token::RBRACE);
}

"//".*$|{WS}|{NUMBER}|{ID}|{CHAR_CONST} {
  yoxOutGimport.put (text ());
}

\n|. {
  yoxOutGimport.put ((char)chr ());
}

<<EOF>> {
  unterminatedCode (yylloc, startPos.top ());
}

}



<START_SYM>{

  /* YACC_DEFS 'subroutine': intercept symbol following '%start'; Ox adds a new
     start symbol w/ an action. */

"/*" {
  yoxOutGdefs.lex ().put (text ());
  yox_push_state (COMMENT, yoxOutGdefs);
}

{DOT_NAME} {
  yylval.emplace<name_literal> (stringLookup (str (), true));
  pop_state ();
  return (parser::token::GS_ID);
}

"//".*$ {
  yoxOutGdefs.lex ().put (text ());
}

\n|. {
  yoxOutGdefs.put ((char)chr ());
}

}



<YACC_UNION_HD>{

  /* YACC_DEFS 'subroutine': process text between '%union' & '{' */

"/*" {
  yox_push_state (SKIP_COMMENT, yoxOutGdefs);
}

"//".*$|\n|{WS} {
  // skip comments & whitespace between "%union" & "{"
  ; // no-op
}

"{"{OPTWS}\n? {
  yylloc.end = yylloc.begin + 1;
  nLbraces++; // omit the first "{" in the LA buffer

  // starting position if EOF in '%union'
  push_state (startPos.top (), YACC_UNION);

  if (yaccUnionDefined > 1)
    unionBODY.gen ().put ("/* concatenated from `%union' #" +
                          std::to_string (yaccUnionDefined) +
                          " */\n");

  return (parser::token::LBRACE);
}

"}"{OPTWS}\n? {
  yylloc.end = yylloc.begin + 1;
  pop_state ();
  // omit the last "}" in the unionBODY lookaheadBuf
  return (parser::token::RBRACE);
}

{ID} {
  // Bison extension: "%union TAG { ... }"
  if (ctrl.targetParserGen == bisonParserGen ||
      ctrl.targetParserGen == bisonParserGenOld)
    {
      const std::string& lexeme (text ());
      yylval.emplace<name_literal> (stringLookup (lexeme, true));
      return (parser::token::ID);
    }
  else
    yox_error (yylloc, "unexpected %union tag: \"" + str () + "\": "
                       "target parser is not bison.\n");
}

. {
  unexpectedChar (yylloc, (char)chr (), "in `%union' declaration header");
}

}



<YACC_UNION>{

  /* YACC_UNION_HD 'subroutine': accumulate the '%union' declaration body */

"/*" {
  unionBODY.lex ().put (text ());
  yox_push_state (COMMENT, unionBODY);
}

"{" {
  nLbraces++;
  unionBODY.lex ().put ((char)chr ());
  return (parser::token::LBRACE);
}

"}" {
  if (--nLbraces == 0)
    {
      yyless (0);
      pop_state ();
    }
  else
    unionBODY.lex ().put ((char)chr ());
}

"//".*$|{WS}|{ID} {
  unionBODY.lex ().put (text ());
}

\n|. {
  unionBODY.put ((char)chr ());
}

<<EOF>> {
  unterminatedYaccUnion (yylloc, startPos.top ());
}

}



<BISONDEF>{

  /* YACC_DEFS 'subroutine': a Bison '%define' directive */

{ID}|"canonical-lr"|"union-directive" {
  yoxOutGdefs.lex ().put (text ());
  pop_state ();
  yylval.emplace<name_literal> (text ());
  return (parser::token::API_VALUE);
}

{API_NAME} {
  yoxOutGdefs.lex ().put (text ());
  yylval.emplace<name_literal> (text ());
  return (parser::token::API_VARIABLE);
}

{STRING_CONST} {
  yoxOutGdefs.lex ().put (text ());
  pop_state ();
  yylval.emplace<char_literal> (text ());
  return (parser::token::LITERAL_STRING);
}

{FILE_PATH} {
  yoxOutGdefs.lex ().put (text ());
  pop_state ();
  yylval.emplace<char_literal> (text ());
  return (parser::token::LITERAL_PATH);
}

"{" {
  yoxOutGdefs.lex ().put ((char)chr ());
  push_state (yylloc, TYPESPEC);
  nLbraces = 1;
  nLTbraces = 0;
  // normal return from TYPESPEC when nLbraces == 0, rescan "}"
  return (parser::token::LBRACE);
}

"}" {
  yoxOutGdefs.lex ().put ((char)chr ());
  pop_state ();
  return (parser::token::RBRACE);
}

{WS} {
  yoxOutGdefs.lex ().put (text ());
}

\n {
  yoxOutGdefs.lex ().put ((char)chr ());
  pop_state ();
}

. {
  yyless (0);
  pop_state ();
}

}



<INT_ARG>{

  /* YACC_DEFS 'subroutine': look for an int */

{WS} {
  yoxOutGdefs.lex ().put (text ());
}

{INTEGER} {
  yoxOutGdefs.lex ().put (text ());
  pop_state ();
  if (pass2parser)
    {
      pass2parser = false;

      const std::string& lexeme (text ());
      yylval.emplace<numberRep> (
          numberRep (lexeme, std::stoll (lexeme, nullptr, 0)));
      return (parser::token::NUM);
    }
}

\n {
  yoxOutGdefs.lex ().put ((char)chr ());
  pop_state ();
  pass2parser = false;
}

. {
  yyless (0);
  pop_state ();
  pass2parser = false;
}

}



<STRING_ARG>{

  /* YACC_DEFS 'subroutine': look for a string */

{WS} {
  yoxOutGdefs.lex ().put (text ());
}

{STRING_LIT} {
  yoxOutGdefs.lex ().put (text ());
  pop_state ();
  if (pass2parser)
    {
      pass2parser = false;

      const std::string& lexeme (text ());
      yylval.emplace<char_literal> (stringLookup (lexeme, true));
      return (parser::token::LITERAL_STRING);
    }
}

\n {
  yoxOutGdefs.lex ().put ((char)chr ());
  pop_state ();
  pass2parser = false;
}

. {
  yyless (0);
  pop_state ();
  pass2parser = false;
}

}



<USER_ARGS>{

  /* YACC_DEFS 'subroutine': user-defined additional parser/lexer arguments */

"{" {
  yoxOutGdefs.lex ().put ((char)chr ());
  push_state (yylloc, TYPESPEC);
  nLbraces = 1;
  nLTbraces = 0;
  // normal return from TYPESPEC when nLbraces == 0, rescan "}"
  return (parser::token::LBRACE);
}

"}" {
  yoxOutGdefs.lex ().put ((char)chr ());
  return (parser::token::RBRACE);
}

{WS} {
  yoxOutGdefs.lex ().put (text ());
}

\n {
  yoxOutGdefs.lex ().put ((char)chr ());
  pop_state ();
}

. {
  yyless (0);
  pop_state ();
}

}



<TYPESPEC,ATTR_DECL>{

  /* C/C++ type specifier rules common to ATTR_DECL and TYPESPEC */

"/*" {
  if (start () == TYPESPEC)
    {
      yoxOutGdefs.lex ().put (text ());
      yox_push_state (COMMENT, yoxOutGdefs);
    }
  else if (start () == ATTR_DECL)
    {
      attrDECL.lex ().put (text ());
      yox_push_state (COMMENT, attrDECL);
    }
}

{NUMBER} {
  if (start () == TYPESPEC)
    yoxOutGdefs.lex ().put (text ());
  else if (start () == ATTR_DECL)
    attrDECL.lex ().put (text ());

  const std::string& lexeme (text ());
  yylval.emplace<numberRep> (
      numberRep (lexeme, std::stoll (lexeme, nullptr, 0)));
  return (parser::token::NUM);
}

{ID} {
  if (start () == TYPESPEC)
    yoxOutGdefs.lex ().put (text ());
  else if (start () == ATTR_DECL)
    attrDECL.lex ().put (text ());

  const struct type_keyword* reserved =
      in_type_keyword_set (text (), (size_t)size ());

  if (reserved == nullptr)
    {
      const std::string& lexeme (text ());
      yylval.emplace<name_literal> (stringLookup (lexeme, true));
      return (parser::token::ID);
    }
  else
    {
      yylval.emplace<keyword> (reserved->name);
      return (reserved->token);
    }
}

{NAMESPACE_ID} {
  if (start () == TYPESPEC)
    yoxOutGdefs.lex ().put (text ());
  else if (start () == ATTR_DECL)
    attrDECL.lex ().put (text ());

  const std::string& lexeme (text ());
  yylval.emplace<name_literal> (stringLookup (lexeme, true));
  return (parser::token::NAMESPACE_ID);
}

"..." {
  if (start () == TYPESPEC)
    yoxOutGdefs.lex ().put (text ());
  else if (start () == ATTR_DECL)
    attrDECL.lex ().put (text ());
  return (parser::token::ELLIPSIS);
}

"," {
  if (start () == TYPESPEC)
    yoxOutGdefs.lex ().put ((char)chr ());
  else if (start () == ATTR_DECL)
    attrDECL.lex ().put ((char)chr ());
  return (parser::token::COMMA);
}

";" {
  if (start () == TYPESPEC)
    yoxOutGdefs.lex ().put ((char)chr ());
  else if (start () == ATTR_DECL)
    attrDECL.lex ().put ((char)chr ());
  return (parser::token::SEMICOLON);
}

":" {
  if (start () == TYPESPEC)
    yoxOutGdefs.lex ().put ((char)chr ());
  else if (start () == ATTR_DECL)
    attrDECL.lex ().put ((char)chr ());
  return (parser::token::COLON);
}

"*" {
  if (start () == TYPESPEC)
    yoxOutGdefs.lex ().put ((char)chr ());
  else if (start () == ATTR_DECL)
    attrDECL.lex ().put ((char)chr ());
  return (parser::token::ASTERISK);
}

"&" {
  if (start () == TYPESPEC)
    yoxOutGdefs.lex ().put ((char)chr ());
  else if (start () == ATTR_DECL)
    attrDECL.lex ().put ((char)chr ());
  return (parser::token::AMPERSAND);
}

"(" {
  if (start () == TYPESPEC)
    yoxOutGdefs.lex ().put ((char)chr ());
  else if (start () == ATTR_DECL)
    attrDECL.lex ().put ((char)chr ());
  return (parser::token::LPAREN);
}

")" {
  if (start () == TYPESPEC)
    yoxOutGdefs.lex ().put ((char)chr ());
  else if (start () == ATTR_DECL)
    attrDECL.lex ().put ((char)chr ());
  return (parser::token::RPAREN);
}

"[" {
  if (start () == TYPESPEC)
    yoxOutGdefs.lex ().put ((char)chr ());
  else if (start () == ATTR_DECL)
    attrDECL.lex ().put ((char)chr ());
  return (parser::token::LBRACKET);
}

"]" {
  if (start () == TYPESPEC)
    yoxOutGdefs.lex ().put ((char)chr ());
  else if (start () == ATTR_DECL)
    attrDECL.lex ().put ((char)chr ());
  return (parser::token::RBRACKET);
}

"<" {
  if (start () == TYPESPEC)
    yoxOutGdefs.lex ().put ((char)chr ());
  else if (start () == ATTR_DECL)
    attrDECL.lex ().put ((char)chr ());
  nLTbraces++;
  return (parser::token::LESSTHAN);
}

">" {
  if (--nLTbraces == 0)
    {
      // entered following "<" in start condition YACC_DEFS
      yyless (0);
      pop_state ();
    }
  else
    {
      if (start () == TYPESPEC)
        yoxOutGdefs.lex ().put ((char)chr ());
      else if (start () == ATTR_DECL)
        attrDECL.lex ().put ((char)chr ());
      return (parser::token::GTRTHAN);
    }
}

"//".*$ {
  if (start () == TYPESPEC)
    yoxOutGdefs.lex ().put (text ());
  else if (start () == ATTR_DECL)
    attrDECL.lex ().put (text ());
}

}



<TYPESPEC>{

  /* YACC_DEFS, BISONDEF 'subroutine': C/C++ type specification */

"{" {
  yoxOutGdefs.lex ().put ((char)chr ());
  nLbraces++;
  return (parser::token::LBRACE);
}

"}" {
  if (--nLbraces == 0)
    {
      // entered following "{" in start states BISONDEF & USER_ARGS
      yyless (0);
      pop_state ();
    }
  else
    {
      yoxOutGdefs.lex ().put ((char)chr ());
      return (parser::token::RBRACE);
    }
}

{WS} {
  yoxOutGdefs.lex ().put (text ());
}

\n {
  yoxOutGdefs.put ((char)chr ());
}

. {
  yyless (0);
  pop_state ();
}

}



<ATTR_DECL>{

  /* YACC_DEFS 'subroutine': process '@attributes' declaration; ended by a
     closing brace */

"@void" {
  // omit generation of '@void' attribute declarations
  push_state (yylloc, OXVOID);
  return (parser::token::OX_VOID);
}

"@autosyn" {
  // yylval not used in this context
  yylval.emplace<keyword> (stringLookup (str (), true));
  return (parser::token::OX_AUTOSYN);
}

"@autoinh" {
  // yylval not used in this context
  yylval.emplace<keyword> (stringLookup (str (), true));
  return (parser::token::OX_AUTOINH);
}

"@warn" {
  // yylval not used in this context
  yylval.emplace<keyword> (stringLookup (str (), true));
  return (parser::token::OX_WARN);
}

{OPTWS}"{" {
  if (nLbraces++ > 0)
    attrDECL.lex ().put (text ());
  return (parser::token::LBRACE);
}

"}" {
  if (--nLbraces == 0)
    start (ATTR_DECL_SYMS);
  else
    attrDECL.lex ().put ((char)chr ());
  return (parser::token::RBRACE);
}

{WS} {
  attrDECL.lex ().put (text ());
}

\n {
  attrDECL.put ((char)chr ());
  yoxOutGdefs.put ((char)chr ());
}

. {
  unexpectedChar (yylloc, (char)chr (), "in @attribute declaration");
}

<<EOF>> {
  unterminatedAttrDeclaration (yylloc, startPos.top ());
}

}



<TRAV_DECL>{

  /* YACC_DEFS 'subroutine': traversal declaration */

"/*" {
  yoxOutGdefs.lex ().put (text ());
  yox_push_state (COMMENT, yoxOutGdefs);
}

"@postorder" {
  return (parser::token::OX_POSTORDER);
}

"@preorder" {
  return (parser::token::OX_PREORDER);
}

"@lefttoright" {
  return (parser::token::OX_LEFTTORIGHT);
}

"@righttoleft" {
  return (parser::token::OX_RIGHTTOLEFT);
}

"@disable" {
  return (parser::token::OX_DISABLE);
}

{ID} {
  yylval.emplace<name_literal> (stringLookup (str (), true));
  return (parser::token::ID);
}

"//".*$ {
  yoxOutGdefs.lex ().put (text ());
}

{WS} {
  ; // no-op
}

^{OPTWS}"%%" {
  yyless (0);
  yy_set_bol (true); // rescan @ beginning of line
  pop_state ();
}

\n {
  yoxOutGdefs.put ((char)chr ());
}

. {
  yyless (0);
  pop_state ();
}

}



<ATTR_DECL_SYMS>{

  /* return from ATTR_DECL back to YACC_DEFS; avoid echoing grammar symbols
     in an attribute declaration symbol list */

"/*" {
  yox_push_state (SKIP_COMMENT, yoxOutGdefs);
}

{CHAR_LIT} {
  yylval.emplace<char_literal> (stringLookup (str (), true));
  return (parser::token::LITERAL_CHAR);
}

{STRING_LIT} {
  yylval.emplace<char_literal> (stringLookup (str (), true));
  return (parser::token::LITERAL_STRING);
}

{DOT_NAME} {
  yylval.emplace<name_literal> (stringLookup (str (), true));
  return (parser::token::GS_ID);
}

"//".*$|{WS} {
  ; // no-op
}

^{OPTWS}"%%" {
  yyless (0);
  yy_set_bol (true); // rescan @ beginning of line
  pop_state ();
}

\n {
  yoxOutGdefs.put ((char)chr ());
}

. {
  yyless (0);
  pop_state ();
}

}



<YACC_CCODE>{

  /* YACC_DEFS 'subroutine': echo (or accumulate) C/C++ code between
     "%{" and "%}" */

"/*" {
  if (yaccUnionDefined != 0)
    {
      yaccCODE.lex ().put (text ());
      yox_push_state (COMMENT, yaccCODE);
    }
  else
    {
      yoxOutGdefs.lex ().put (text ());
      yox_push_state (COMMENT, yoxOutGdefs);
    }
}

"L"?"\"" {
  if (yaccUnionDefined != 0)
    {
      yaccCODE.lex ().put (text ());
      yox_push_state (STRING, yaccCODE);
    }
  else
    {
      yoxOutGdefs.lex ().put (text ());
      yox_push_state (STRING, yoxOutGdefs);
    }
}

"{" {
  if (yaccUnionDefined != 0)
    yaccCODE.lex ().put ((char)chr ());
  else
    yoxOutGdefs.lex ().put ((char)chr ());
  return (parser::token::LBRACE);
}

"}" {
  if (yaccUnionDefined != 0)
    yaccCODE.lex ().put ((char)chr ());
  else
    yoxOutGdefs.lex ().put ((char)chr ());
  return (parser::token::RBRACE);
}

"%}" {
  if (yaccUnionDefined != 0)
    yaccCODE.lex ().put (text ());
  else
    yoxOutGdefs.lex ().put (text ());
  pop_state ();
  return (parser::token::RCURL);
}

"//".*$|{WS}|{NUMBER}|{ID}|{CHAR_CONST} {
  if (yaccUnionDefined != 0)
    yaccCODE.put (text ());
  else
    yoxOutGdefs.put (text ());
}

\n {
  if (yaccUnionDefined != 0)
    yaccCODE.put ((char)chr ());
  yoxOutGdefs.put ((char)chr ());
}

. {
  if (yaccUnionDefined != 0)
    yaccCODE.put ((char)chr ());
  else
    yoxOutGdefs.put ((char)chr ());
}

<<EOF>> {
  unterminatedCode (yylloc, startPos.top ());
}

}



<OXVOID>{

  /* ATTR_DECL 'subroutine': don't output '@void' attribute declarations */

"/*" {
  attrDECL.lex ().put (text ());
  yox_push_state (COMMENT, attrDECL);
}

{ID} {
  yylval.emplace<name_literal> (stringLookup (str (), true));
  return (parser::token::ID);
}

"," {
  return (parser::token::COMMA);
}

";" {
  pop_state ();
  return (parser::token::SEMICOLON);
}

"//".*$ {
  attrDECL.lex ().put (text ());
}

\n {
  attrDECL.put ((char)chr ());
  yoxOutGdefs.put ((char)chr ());
}

. {
  attrDECL.put ((char)chr ());
}

}



<MACRO_HEAD>{

  /* macro formal(s): transition from YACC_DEFS to MACRO_BODY */

"/*" {
  yoxOutGdefs.lex ().put (text ());
  yox_push_state (COMMENT, yoxOutGdefs);
}

{ID} {
  yylval.emplace<name_literal> (stringLookup (str (), true));
  return (parser::token::ID);
}

"(" {
  return (parser::token::LPAREN);
}

"," {
  return (parser::token::COMMA);
}

")"{OPTWS}\n? {
  if (yylloc.end.column == 0)
    {
      // macro header ends with '\n': include it in the output Y-file
      yoxOutGdefs.lex ().put ('\n');
      yylloc.end.column = 1; // body starts on 'end' line @column 1
    }
  macro.setOrigin (yylloc.end);
  yylloc.end = yylloc.begin + 1;
  start (MACRO_BODY);
  return (parser::token::RPAREN);
}

"//".*$ {
  yoxOutGdefs.lex ().put (text ());
}

{WS} {
  ; // no-op
}

\n {
  yoxOutGdefs.put ((char)chr ());
}

. {
  unexpectedChar (yylloc, (char)chr (), "in macro heading");
}

}



<MACRO_BODY>{

  /* macro body: transition from MACRO_HEAD to YACC_DEFS */

"/*" {
  macro.lex ().put (text ());
  yox_push_state (COMMENT, macro);
}

"L"?"\"" {
  macro.lex ().put (text ());
  yox_push_state (STRING, macro);
}

{ID} {
  macro.put (stringLookup (str (), true));
}

"\\@macro" {
  // nested macro definition start
  macro.lex ().put ("@macro");
}

"\\@end" {
  // nested macro definition end
  macro.lex ().put ("@end");
}

"@end"{OPTWS} {
  if (text ()[size () - 1] != 'd')
    yylloc.end = yylloc.begin + 4; // trim trailing whitespace
  pop_state (); // -> YACC_DEFS
  macro.close ();
  return (parser::token::OX_END);
}

"//".*$ {
  macro.lex ().put (text ());
}

\n {
  macro.put ((char)chr ());
  yoxOutGdefs.put ((char)chr ());
}

. {
  macro.put ((char)chr ());
}

}



<YACC_RULES>{

  /* Y-file Rules Section */

{WS}"@{" {
  // throw away whitespace preceding "@{" "}"<-balance lbrace
  yyless (size () - 2); // rescan "@{"
  notLHS ();
}

"@{" {
  flushWS (possibleLHS);
  if (nATLbraces++ > 0)
    unexpectedATLBRACE (yylloc);
  else
    {
      push_state (yylloc, YATTR_REF);
      return (parser::token::OX_LBRACE);
    }
}

"/*" {
  flushWS (possibleLHS);
  possibleLHS.lex ().put (text ());
  yox_push_state (COMMENT, possibleLHS);
}

{CHAR_LIT} {
  notLastSemAction ();
  notLHS ();
  flushWS (yoxOutG);
  yoxOutG.lex ().put (text ());

  const std::string& lexeme (text ());
  yylval.emplace<char_literal> (stringLookup (lexeme, true));
  return (parser::token::LITERAL_CHAR);
}

{STRING_LIT} {
  notLastSemAction ();
  notLHS ();
  flushWS (yoxOutG);
  yoxOutG.lex ().put (text ());

  const std::string& lexeme (text ());
  yylval.emplace<char_literal> (stringLookup (lexeme, true));
  return (parser::token::LITERAL_STRING);
}

"L"?"\"" {
  notLastSemAction ();
  notLHS ();
  flushWS (yoxOutG);
  yoxOutG.lex ().put (text ());
  yox_push_state (STRING, yoxOutG);
}

{NUMBER} {
  notLastSemAction ();
  notLHS ();
  flushWS (yoxOutG);
  yoxOutG.lex ().put (text ());

  const std::string& lexeme (text ());
  yylval.emplace<numberRep> (
      numberRep (lexeme, std::stoll (lexeme, nullptr, 0)));
  return (parser::token::NUM);
}

{ID} {
  notLastSemAction ();
  notLHS ();
  flushWS (possibleLHS);
  possibleLHS.lex ().put (text ());

  if (ctrl.targetParserGen == btyaccParserGen ||
      ctrl.targetParserGen == byaccParserGen)
    {
      // BtYacc & BYacc extension: possibly followed by inherited attributes
      nLparens = 0;
      push_state (yylloc, INHERIT_ATTR);
    }

  const std::string& lexeme (text ());
  yylval.emplace<name_literal> (stringLookup (lexeme, true));
  return (parser::token::ID);
}

{DOT_NAME} {
  notLastSemAction ();
  notLHS ();
  flushWS (possibleLHS);
  possibleLHS.lex ().put (text ());

  if (ctrl.targetParserGen == btyaccParserGen ||
      ctrl.targetParserGen == byaccParserGen)
    {
      // BtYacc & BYacc extension: possibly followed by inherited attributes
      nLparens = 0;
      push_state (yylloc, INHERIT_ATTR);
    }

  const std::string& lexeme (text ());
  yylval.emplace<name_literal> (stringLookup (lexeme, true));
  return (parser::token::GS_ID);
}

{YOX_DIRECTIVE} {
  const struct yox_directive* directive =
      in_yox_directive_set (text (), (size_t)size ());

  if (directive == nullptr)
    {
      // "%<directive>" not important to Ox, just pass it on through
      if ((char)chr () == '%')
        {
          notLastSemAction ();
          notLHS ();
          flushWS (yoxOutG);
          yoxOutG.lex ().put (text ());
        }
      else if ((char)chr () == '@')
        {
          // found "@{ID}" -> will be a syntax error
          const std::string& lexeme (text ());
          yylval.emplace<name_literal> (stringLookup (lexeme.substr (1), true));
          return (parser::token::OX_ID);
        }
    }
  else
    {
      yylval.emplace<keyword> (directive->name);

      switch (directive->token)
        {
        case parser::token::EMPTY:
          notLastSemAction ();
          notLHS ();
          flushWS (yoxOutG);
          yoxOutG.lex ().put (text ());
          return (parser::token::EMPTY);
        case parser::token::DPREC:
        case parser::token::EXPECT:
        case parser::token::EXPECT_RR:
        case parser::token::LOOKAHEAD:
        case parser::token::MERGE:
        case parser::token::PREC:
          if (!actionCLOSER.empty ())
            {
              push_state (yylloc, ACTIONCLOSER_DIRECTIVE);
              actionCLOSER.append (possibleLHS);
              possibleLHS.clear ();
              flushWS (actionCLOSER);
              actionCLOSER.lex ().put (text ());
            }
          else
            {
              notLHS ();
              push_state (yylloc, YOXOUTG_DIRECTIVE);
              flushWS (yoxOutG);
              yoxOutG.lex ().put (text ());
            }
          return (directive->token);
        default: // anything that makes it here will be a syntax error
          notLastSemAction ();
          notLHS ();
          flushWS (yoxOutG);
          yoxOutG.lex ().put (text ());
          return (directive->token);
        }
    }
}

^{OPTWS}"%%".*\n {
  yylloc.begin += std::strcspn (text (), "%");
  yylloc.end = yylloc.begin + 2;
  flushWS (possibleLHS);
  notLastSemAction ();
  notLHS ();
  separatorMARK.put (text ());
  start (YACC_USER);
  return (parser::token::MARK);
}

"|" {
  // found the end a rule:

  if (!actionCLOSER.empty ())
    {
      // 'actionCLOSER' holds the closing '}' of the last Yacc rule action;
      // append whatever might be in 'possibleLHS' (s.b. empty)
      actionCLOSER.append (possibleLHS);
      possibleLHS.clear ();
    }
  else
    notLHS ();

  flushWS (possibleLHS);
  possibleLHS.lex ().put ((char)chr ());
  return (parser::token::VBAR);
}

";" {
  // found the end of a rule:

  if (!actionCLOSER.empty ())
    {
      // 'actionCLOSER' holds the closing '}' of the last Yacc rule action;
      // append whatever might be in 'possibleLHS' (s.b. empty)
      actionCLOSER.append (possibleLHS);
      possibleLHS.clear ();
    }
  else
    notLHS ();

  flushWS (possibleLHS);
  possibleLHS.lex ().put ((char)chr ());
  return (parser::token::SEMICOLON);
}

"(" {
  notLastSemAction ();
  notLHS ();
  flushWS (yoxOutG);
  yoxOutG.lex ().put ((char)chr ());
  return (parser::token::LPAREN);
}

")" {
  notLastSemAction ();
  notLHS ();
  flushWS (yoxOutG);
  yoxOutG.lex ().put ((char)chr ());
  return (parser::token::RPAREN);
}

"/" {
  notLastSemAction ();
  notLHS ();
  flushWS (yoxOutG);
  yoxOutG.lex ().put ((char)chr ());
  return (parser::token::SLASH);
}

"*" {
  notLastSemAction ();
  notLHS ();
  flushWS (yoxOutG);
  yoxOutG.lex ().put ((char)chr ());
  return (parser::token::ASTERISK);
}

"-" {
  notLastSemAction ();
  notLHS ();
  flushWS (yoxOutG);
  yoxOutG.lex ().put ((char)chr ());
  return (parser::token::MINUS);
}

"+" {
  notLastSemAction ();
  notLHS ();
  flushWS (yoxOutG);
  yoxOutG.lex ().put ((char)chr ());
  return (parser::token::PLUS);
}

"@" {
  notLastSemAction ();
  notLHS ();
  flushWS (yoxOutG);
  yoxOutG.lex ().put ((char)chr ());
  return (parser::token::AT);
}

"{" {
  notLastSemAction ();
  notLHS ();
  flushWS (yoxOutG);
  yoxOutG.lex ().put ((char)chr ());
  nLbraces = 1;
  push_state (yylloc, YACC_ACT);
  return (parser::token::LBRACE);
}

"[" {
  notLastSemAction ();
  notLHS ();
  flushWS (yoxOutG);
  yoxOutG.lex ().put ((char)chr ());
  if (ctrl.targetParserGen == btyaccParserGen ||
      ctrl.targetParserGen == byaccParserGen)
    {
      // BtYacc & BYacc extension: start trial & normal mode action
      nLbrackets = 1;
      push_state (yylloc, BTYACC_ACT);
      return (parser::token::LBRACKET);
    }
  else if (ctrl.targetParserGen == bisonParserGen ||
           ctrl.targetParserGen == mstaParserGen)
    {
      // Bison extension: start explicit named reference definition
      // MSTA extension: start EBNF optional construct
      return (parser::token::LBRACKET);
    }
  else
    unexpectedChar (yylloc, (char)chr (), "in rules section");
}

"]" {
  notLastSemAction ();
  notLHS ();
  flushWS (yoxOutG);
  yoxOutG.lex ().put ((char)chr ());
  return (parser::token::RBRACKET);
}

"<" {
  notLastSemAction ();
  notLHS ();
  flushWS (yoxOutG);
  yoxOutG.lex ().put ((char)chr ());
  return (parser::token::LESSTHAN);
}

">" {
  notLastSemAction ();
  notLHS ();
  flushWS (yoxOutG);
  yoxOutG.lex ().put ((char)chr ());
  return (parser::token::GTRTHAN);
}

"<-" {
  notLastSemAction ();
  notLHS ();
  flushWS (yoxOutG);
  yoxOutG.lex ().put (text ());
  return (parser::token::LTARROW);
}

"->" {
  notLastSemAction ();
  notLHS ();
  flushWS (yoxOutG);
  yoxOutG.lex ().put (text ());
  return (parser::token::RTARROW);
}

"<->" {
  notLastSemAction ();
  notLHS ();
  flushWS (yoxOutG);
  yoxOutG.lex ().put (text ());
  return (parser::token::DBLARROW);
}

"%?" {
  notLastSemAction ();
  notLHS ();
  flushWS (yoxOutG);
  yoxOutG.lex ().put (text ());
  return (parser::token::PREDICATE);
}

":" {
  flushWS (possibleLHS);
  possibleLHS.lex ().put ((char)chr ());
  return (parser::token::COLON);
}

{WS} {
  bufferWS ();
}

"//".*$ {
  flushWS (possibleLHS);
  possibleLHS.lex ().put (text ());
}

\n {
  flushWS (possibleLHS);
  possibleLHS.lex ().put ((char)chr ());
}

. {
  notLastSemAction ();
  notLHS ();
  flushWS (yoxOutG);
  unexpectedChar (yylloc, (char)chr (), "in rules section");
}

}



<YATTR_REF,ATTR_DEPS>{

  /* YATTR_REF: YACC_RULES 'subroutine': Y-file attribute reference section */
  /* ATTR_DEPS: YATTR_REF 'subroutine': explicit/mixed attribute occurrences */

"/*" {
  rsBuffer.lex ().put (text ());
  yox_push_state (COMMENT, rsBuffer);
}

"L"?"\"" {
  rsBuffer.lex ().put (text ());
  yox_push_state (STRING, rsBuffer);
}

{ID} {
  const std::string& lexeme (text ());
  const std::string& Name (stringLookup (lexeme, true));

  do
    {
      if (expandingMacro ())
        {
          // check to see if 'Name' is a macro formal parameter name
          int argIndex = getCurMacro ().macroFormalLookup (Name);
          if (argIndex != -1)
            {
              std::string substValue (substFormal (argIndex));
              position arg_origin (yylloc.begin);

              // setup new input buffer for macro argument
              YY_BUFFER_STATE newbuf =
                  yy_create_buffer (new std::istringstream (substValue),
                                    YY_BUF_SIZE);
              pushbuffer (newbuf, arg_origin); // make it current
              setupFormalExpansion (yylloc);
              break;
            }
        }

      if (macroLookup (Name))
        {
          nLparensM = 0;
          push_state (yylloc, EXPAND_MACRO);
          expandMacro (Name, yylloc.begin);
        }
      else
        {
          rsBuffer.lex ().put (Name);
          yylval.emplace<name_literal> (Name);
          return (parser::token::ID);
        }
    }
  while (false);
}

{CPP_DIRECTIVE} {
  rsBuffer.lex ().put (text ());

  if (text ()[size () - 1] == '\\')
    {
      // multi-line directive; include the following '\n'
      rsBuffer.lex ().put ((char)yyinput ());
    }
}

"@{" {
  unexpectedATLBRACE (yylloc);
}

"@}" {
  --nATLbraces;

  if (start () == ATTR_DEPS)
    pop_state ();

  pop_state ();
  return (parser::token::OX_RBRACE);
}

"@decorate" {
  return (parser::token::OX_DECORATE);
}

"@revorder" {
  return (parser::token::OX_REVORDER);
}

"@revdirection" {
  return (parser::token::OX_REVDIR);
}

"@"{ID} {
  const std::string& lexeme (text ());
  std::string Name (stringLookup (lexeme.substr (1), true));

  if (start () == ATTR_DEPS)
    pop_state ();

  if (expandingMacro ())
    {
      // check to see if 'Name' is a macro formal parameter name
      int argIndex = getCurMacro ().macroFormalLookup (Name);
      if (argIndex != -1)
        Name = stringLookup (substFormal (argIndex), true);
    }

  if (Name.length () == 1)
    {
      // could be an attribute definition annunciator
      rsBuffer.init (yylloc.begin);
      switch (text ()[1])
        {
        case 'e':
          push_state (yylloc, ATTR_DEPS);
          return (parser::token::OX_E);
        case 'i':
          return (parser::token::OX_I);
        case 'm':
          push_state (yylloc, ATTR_DEPS);
          return (parser::token::OX_M);
        default:
          break; // nope
        }
    }

  // otherwise, this should be a traversal mode annunciator

  currentTrav = travLookup (Name);
  if (currentTrav == -1)
    {
      if (size () == 2)
        {
          // typo: corrupted "definition mode annunciator"?
          yox_error (yylloc, "unknown traversal: \"@" + Name + "\"\n"
                             "   assuming a corrupted `@i' definition"
                             " mode annunciator.\n");

          return (parser::token::OX_I);
        }
      else
        yox_error (yylloc, "unknown traversal: \"@" + Name + "\".\n");
    }

  // it is a traversal mode annunciator
  yylval.emplace<name_literal> (Name);
  return (parser::token::OX_ID);
}

{PVAR_REF} {
  // does not check for macro formal parameter occurrences
  rsBuffer.lex ().put (text ());
  if (!unionDefined ())
    notSupported (yylloc, "Yacc pseudo variable with no '%union'");
  hasYaccPvarRefGlob = true;
}

{PLOC_REF} {
  // does not check for macro formal parameter occurrences
  rsBuffer.lex ().put (text ());
  hasYaccPvarRefGlob = true;
}

{ATTRB_REF} {
  int nAt = 1; // # of '@'
  if (text ()[(int)size () - 1] == '@')
    ++nAt;
  else
    yox_warning (yylloc, "missing final '@': \"" + str () + "\".\n");

  std::string attrb_occur (text (), 1, size () - nAt);
  if (expandingMacro ())
    expandFormalsInAttrOccr (attrb_occur);

  yylval.emplace<attrbTriple> (scanAttrbName (yylloc, attrb_occur));
  return (parser::token::OX_ATTRB_REF);
}

"//".*$|{WS} {
  rsBuffer.lex ().put (text ());
}

{CHAR_CONST} {
  rsBuffer.lex ().put (text ());

  const std::string& lexeme (text ());
  yylval.emplace<char_literal> (stringLookup (lexeme, true));
  return (parser::token::LITERAL_CHAR);
}

{STRING_CONST} {
  rsBuffer.lex ().put (text ());

  const std::string& lexeme (text ());
  yylval.emplace<char_literal> (stringLookup (lexeme, true));
  return (parser::token::LITERAL_STRING);
}

{NUMBER} {
  rsBuffer.lex ().put (text ());

  const std::string& lexeme (text ());
  yylval.emplace<numberRep> (
      numberRep (lexeme, std::stoll (lexeme, nullptr, 0)));
  return (parser::token::NUM);
}

^{OPTWS}"%%".*$ {
  yylloc.begin += std::strcspn (text (), "%");
  yylloc.end = yylloc.begin + 2;
  // "%%" -> will be a syntax error
  return (parser::token::MARK);
}

"(" {
  rsBuffer.lex ().put ((char)chr ());
  return (parser::token::LPAREN);
}

")" {
  rsBuffer.lex ().put ((char)chr ());
  return (parser::token::RPAREN);
}

"[" {
  rsBuffer.lex ().put ((char)chr ());
  return (parser::token::LBRACKET);
}

"]" {
  rsBuffer.lex ().put ((char)chr ());
  return (parser::token::RBRACKET);
}

"{" {
  rsBuffer.lex ().put ((char)chr ());
  return (parser::token::LBRACE);
}

"}" {
  rsBuffer.lex ().put ((char)chr ());
  return (parser::token::RBRACE);
}

";" {
  if (start () == ATTR_DEPS)
    pop_state ();
  else
    rsBuffer.lex ().put ((char)chr ());
  return (parser::token::SEMICOLON);
}

"." {
  rsBuffer.lex ().put ((char)chr ());
  return (parser::token::PERIOD);
}

"->" {
  rsBuffer.lex ().put (text ());
  return (parser::token::RTARROW);
}

"++" {
  rsBuffer.lex ().put (text ());
  return (parser::token::PLUSPLUS);
}

"--" {
  rsBuffer.lex ().put (text ());
  return (parser::token::MINUSMINUS);
}

"~" {
  rsBuffer.lex ().put ((char)chr ());
  return (parser::token::TILDE);
}

"!" {
  rsBuffer.lex ().put ((char)chr ());
  return (parser::token::EXCLAMATION);
}

"-" {
  rsBuffer.lex ().put ((char)chr ());
  return (parser::token::MINUS);
}

"+" {
  rsBuffer.lex ().put ((char)chr ());
  return (parser::token::PLUS);
}

"&" {
  rsBuffer.lex ().put ((char)chr ());
  return (parser::token::AMPERSAND);
}

"*" {
  rsBuffer.lex ().put ((char)chr ());
  return (parser::token::ASTERISK);
}

"/" {
  rsBuffer.lex ().put ((char)chr ());
  return (parser::token::SLASH);
}

"%" {
  rsBuffer.lex ().put ((char)chr ());
  return (parser::token::PERCENT);
}

"<<" {
  rsBuffer.lex ().put (text ());
  return (parser::token::LTSHIFT);
}

">>" {
  rsBuffer.lex ().put (text ());
  return (parser::token::RTSHIFT);
}

"<" {
  rsBuffer.lex ().put ((char)chr ());
  return (parser::token::LESSTHAN);
}

">" {
  rsBuffer.lex ().put ((char)chr ());
  return (parser::token::GTRTHAN);
}

"<=" {
  rsBuffer.lex ().put (text ());
  return (parser::token::LTOREQ);
}

">=" {
  rsBuffer.lex ().put (text ());
  return (parser::token::GTOREQ);
}

"==" {
  rsBuffer.lex ().put (text ());
  return (parser::token::EQUAL);
}

"!=" {
  rsBuffer.lex ().put (text ());
  return (parser::token::NOTEQUAL);
}

"^" {
  rsBuffer.lex ().put ((char)chr ());
  return (parser::token::UPARROW);
}

"|" {
  rsBuffer.lex ().put ((char)chr ());
  return (parser::token::VBAR);
}

"&&" {
  rsBuffer.lex ().put (text ());
  return (parser::token::LOGAND);
}

"||" {
  rsBuffer.lex ().put (text ());
  return (parser::token::LOGOR);
}

"?" {
  rsBuffer.lex ().put ((char)chr ());
  return (parser::token::QUESTION);
}

":" {
  if (start () != ATTR_DEPS)
    {
      // skip output of ':' in (explicit) attribute definition
      rsBuffer.lex ().put ((char)chr ());
    }
  return (parser::token::COLON);
}

"=" {
  rsBuffer.lex ().put ((char)chr ());
  return (parser::token::ASSIGN);
}

"+=" {
  rsBuffer.lex ().put (text ());
  return (parser::token::PLUSEQ);
}

"-=" {
  rsBuffer.lex ().put (text ());
  return (parser::token::MINUSEQ);
}

"*=" {
  rsBuffer.lex ().put (text ());
  return (parser::token::TIMESEQ);
}

"/=" {
  rsBuffer.lex ().put (text ());
  return (parser::token::DIVEQ);
}

"%=" {
  rsBuffer.lex ().put (text ());
  return (parser::token::REMEQ);
}

"<<=" {
  rsBuffer.lex ().put (text ());
  return (parser::token::LTSHIFTEQ);
}

">>=" {
  rsBuffer.lex ().put (text ());
  return (parser::token::RTSHIFTEQ);
}

"&=" {
  rsBuffer.lex ().put (text ());
  return (parser::token::BITANDEQ);
}

"^=" {
  rsBuffer.lex ().put (text ());
  return (parser::token::BITXOREQ);
}

"|=" {
  rsBuffer.lex ().put (text ());
  return (parser::token::BITOREQ);
}

"," {
  rsBuffer.lex ().put ((char)chr ());
  return (parser::token::COMMA);
}

"->*" {
  rsBuffer.lex ().put (text ());
  return (parser::token::PTRTOMEMBER);
}

".*" {
  rsBuffer.lex ().put (text ());
  return (parser::token::MEMBERSELECT);
}

"::" {
  rsBuffer.lex ().put (text ());
  return (parser::token::SCOPE);
}

\n {
  rsBuffer.lex ().put ((char)chr ());
}

. {
  unexpectedChar (yylloc, (char)chr (), "in attribute reference section");
}

<<EOF>> {
  if (replacingFormal ())
    {
      // current macro argument complete
      popbuffer; // revert to previous input buffer
      macroFormalDone ();
    }
  else if (expandingMacro ())
    {
      // current macro expansion complete
      popbuffer; // revert to previous input buffer
      macroDone ();
    }
  else
    unterminatedAttrRefSection (yylloc, startPos.top ());
}

}



<ATTR_DEPS>{

  /* YATTR_REF 'subroutine': handle explicit and mixed attribute occurrences */

{ATTRB_OCCUR} {
  std::string attrb_occur (text ());
  if (expandingMacro ())
    expandFormalsInAttrOccr (attrb_occur);

  yylval.emplace<attrbTriple> (scanAttrbName (yylloc, attrb_occur));
  return (parser::token::OX_ATTRB_OCCUR);
}

}



<INHERIT_ATTR>{

  /* YACC_RULES 'subroutine': BtYacc inherited attribute(s) following a grammar
     symbol */

{WS}"@{" {
  // grammar symbol, NOT action, precedes attribute reference section

  // throw away whitespace preceding "@{" "}"<-balance lbrace
  yyless (size () - 2); // rescan "@{"

  notLastSemAction ();
  notLHS ();
  pop_state ();
}

"/*" {
  flushWS (possibleLHS);
  possibleLHS.lex ().put (text ());
  yox_push_state (COMMENT, possibleLHS);
}

":" {
  if (nLparens == 0)
    {
      // ':' is the Yacc end of LHS indicator
      yyless (0);
      pop_state ();
    }
  else
    {
      // ':' is part of a C/C++ '?:' operator in a BtYacc RHS inherited
      // attribute expression
      flushWS (possibleLHS);
      possibleLHS.lex ().put ((char)chr ());
      return (parser::token::COLON);
    }
}

"(" {
  flushWS (possibleLHS);
  possibleLHS.lex ().put ((char)chr ());
  ++nLparens;
  return (parser::token::LPAREN);
}

")" {
  flushWS (possibleLHS);
  possibleLHS.lex ().put ((char)chr ());
  if (--nLparens == 0)
    pop_state ();
  return (parser::token::RPAREN);
}

"," {
  if (nLparens == 0)
    {
      // ',' will be a syntax error
      yyless (0);
      pop_state ();
    }
  else
    {
      flushWS (possibleLHS);
      possibleLHS.lex ().put ((char)chr ());
      return (parser::token::COMMA);
    }
}

{ID}|{PVAR_REF} {
  if (nLparens == 0)
    {
      yyless (0);
      pop_state ();
    }
  else
    {
      flushWS (possibleLHS);
      possibleLHS.lex ().put (text ());

      const std::string& lexeme (text ());
      yylval.emplace<name_literal> (stringLookup (lexeme, true));
      if ((char)chr () == '$')
        return (parser::token::INHATTR_ID);
      else
        return (parser::token::ID);
    }
}

{WS} {
  bufferWS ();
}

"//".*$ {
  flushWS (possibleLHS);
  possibleLHS.lex ().put (text ());
}

\n {
  flushWS (possibleLHS);
  possibleLHS.lex ().put ((char)chr ());
}

. {
  if (nLparens == 0)
    {
      yyless (0);
      pop_state ();
    }
  else
    {
      // expression component of a BtYacc RHS inherited attribute value
      flushWS (possibleLHS);
      possibleLHS.put ((char)chr ());
    }
}

}



<ACTIONCLOSER_DIRECTIVE,YOXOUTG_DIRECTIVE>{

  /* YACC_RULES 'subroutine': ensure that any rule directives immediately
     following the last Yacc action in a RHS follow the closing brace of the
     modified action */

"/*" {
  flushWS (ruleDIRECTIVE);
  ruleDIRECTIVE.lex ().put (text ());
  yox_push_state (COMMENT, ruleDIRECTIVE);
}

{NUMBER} {
  // Bison, BYacc, MSTA: %expect N
  // Bison, BYacc: %expect-rr N
  // Bison: %dprec N
  // MSTA: %la N
  flushWS (ruleDIRECTIVE);
  ruleDIRECTIVE.lex ().put (text ());

  const std::string& lexeme (text ());
  yylval.emplace<numberRep> (
      numberRep (lexeme, std::stoll (lexeme, nullptr, 0)));
  return (parser::token::NUM);
}

{ID} {
  // Bison: %prec 'terminal grammar symbol'
  flushWS (ruleDIRECTIVE);
  ruleDIRECTIVE.lex ().put (text ());

  const std::string& lexeme (text ());
  yylval.emplace<name_literal> (stringLookup (lexeme, true));
  return (parser::token::ID);
}

{DOT_NAME} {
  // Bison: %prec 'terminal grammar symbol'
  flushWS (ruleDIRECTIVE);
  ruleDIRECTIVE.lex ().put (text ());

  const std::string& lexeme (text ());
  yylval.emplace<name_literal> (stringLookup (lexeme, true));
  return (parser::token::GS_ID);
}

"<" {
  // Bison: %merge < 'function name' >
  flushWS (ruleDIRECTIVE);
  ruleDIRECTIVE.lex ().put ((char)chr ());
  return (parser::token::LESSTHAN);
}

">" {
  // Bison: %merge < ... >
  flushWS (ruleDIRECTIVE);
  ruleDIRECTIVE.lex ().put ((char)chr ());
  return (parser::token::GTRTHAN);
}

{WS} {
  bufferWS ();
}

"//".*$ {
  flushWS (ruleDIRECTIVE);
  ruleDIRECTIVE.lex ().put (text ());
}

\n {
  flushWS (ruleDIRECTIVE);
  ruleDIRECTIVE.lex ().put ((char)chr ());
}

}



<ACTIONCLOSER_DIRECTIVE>{

. {
  flushWS (ruleDIRECTIVE);
  actionCLOSER.append (ruleDIRECTIVE);
  ruleDIRECTIVE.clear ();
  yyless (0);
  pop_state ();
}

}



<YOXOUTG_DIRECTIVE>{

. {
  flushWS (ruleDIRECTIVE);
  yoxOutG.append (ruleDIRECTIVE);
  ruleDIRECTIVE.clear ();
  yyless (0);
  pop_state ();
}

}



<YACC_ACT>{

  /* YACC_RULES 'subroutine': copy a Yacc {}-action until the opening brace is
     paired, ignoring braces in comments, strings and character constants */

"{" {
  yoxOutG.lex ().put ((char)chr ());
  nLbraces++;
  return (parser::token::LBRACE);
}

{WS}"}" {
  yyless (size () - 1); // rescan '}'
  actionCLOSER.lex ().put (text ()); // accumulate the whitespace
}

"}"{OPTWS}\n? {
  yylloc.end = yylloc.begin + 1;
  actionCLOSER.lex ().put (text ());
  if (--nLbraces == 0)
    pop_state ();
  else
    notLastSemAction ();
  return (parser::token::RBRACE);
}

<<EOF>> {
  unterminatedAction (yylloc, startPos.top ());
}

}



<BTYACC_ACT>{

  /* YACC_RULES 'subroutine': copy a BtYacc []-action until the opening bracket
     is paired, ignoring brackets in comments, strings & character constants */

"[" {
  yoxOutG.lex ().put ((char)chr ());
  nLbrackets++;
  return (parser::token::LBRACKET);
}

"]" {
  yoxOutG.lex ().put ((char)chr ());
  if (--nLbrackets == 0)
    pop_state ();
  return (parser::token::RBRACKET);
}

"{" {
  yoxOutG.lex ().put ((char)chr ());
  return (parser::token::LBRACE);
}

"}" {
  yoxOutG.lex ().put ((char)chr ());
  return (parser::token::RBRACE);
}

<<EOF>> {
  unterminatedAction (yylloc, startPos.top ());
}

}



<YACC_ACT,BTYACC_ACT>{

"/*" {
  yoxOutG.lex ().put (text ());
  yox_push_state (COMMENT, yoxOutG);
}

"L"?"\"" {
  yoxOutG.lex ().put (text ());
  yox_push_state (STRING, yoxOutG);
}

{PVAR_REF} {
  yoxOutG.lex ().put (text ());
  if (!unionDefined ())
    notSupported (yylloc, "Yacc pseudo variable with no `%union'");
}

"//".*$|{WS}|{NUMBER}|{ID}|{CHAR_CONST} {
  yoxOutG.put (text ());
}

\n|. {
  yoxOutG.put ((char)chr ());
}

}



<EXPAND_MACRO>{

  /* YATTR_REF, ATTR_DEPS 'subroutine': expand a macro */

"/*" {
  rsBuffer.lex ().put (text ());
  yox_push_state (COMMENT, rsBuffer);
}

"L"?"\"" {
  rsBuffer.lex ().put (text ());
  yox_push_state (STRING, rsBuffer);
}

"(" {
  if (++nLparensM == 1)
    {
      // start of a new argument
      macroARG = macroArg (yylloc.end); // -> parser output
      push_state (yylloc, MACRO_ARG);
    }
  else
    unbalancedParensInMacro (yylloc);
}

")" {
  if (nLparensM == 0)
    {
      macroARG.terminate ();

      std::string macroBody;
      position body_origin;
      setupMacroExpansion (yylloc, macroBody, body_origin);

      // setup new input buffer for macro body
      YY_BUFFER_STATE newbuf =
          yy_create_buffer (new std::istringstream (macroBody), YY_BUF_SIZE);
      pushbuffer (newbuf, body_origin); // make it current
      pop_state ();
    }
  else
    unbalancedParensInMacro (yylloc);
}

"," {
  macroARG.terminate ();
  // start of a new argument
  macroARG = macroArg (yylloc.end); // -> parser output
  push_state (yylloc, MACRO_ARG);
}

"//".*$|{WS}|{NUMBER}|{ID}|{CHAR_CONST} {
  rsBuffer.lex ().put (text ());
}

\n|. {
  rsBuffer.put ((char)chr ());
}

}



<MACRO_ARG>{

  /* EXPAND_MACRO 'subroutine': process a macro argument */

"/*" {
  macroARG.put (text ());
  yox_push_state (COMMENT, macroARG);
}

"L"?"\"" {
  macroARG.put (text ());
  yox_push_state (STRING, macroARG);
}

"(" {
  if (++nLparensM > 1)
    {
      // nested expression in the argument
      macroARG.put ((char)chr ());
    }
  else
    unbalancedParensInMacro (yylloc);
}

")" {
  if (--nLparensM > 0)
    macroARG.put ((char)chr ());
  else if (nLparensM == 0)
    {
      yyless (0);
      pop_state ();
    }
  else
    unbalancedParensInMacro (yylloc);
}

"," {
  if (nLparensM > 1)
    macroARG.put ((char)chr ());
  else if (nLparensM == 1)
    {
      yyless (0);
      pop_state ();
    }
  else
    unbalancedParensInMacro (yylloc);
}

\\\\ {
  macroARG.put ('\\');
}

\\, {
  macroARG.put (',');
}

\\\( {
  macroARG.put ('(');
}

\\\) {
  macroARG.put (')');
}

\\ {
  escapeCharMisuse (yylloc);
}

"//".*$|{WS}|{NUMBER}|{ID}|{CHAR_CONST} {
  macroARG.put (text ());
}

\n|. {
  macroARG.put ((char)chr ());
}

}



<YACC_USER>{

  /* start of the optional user section following the second "%%" in the Y-file,
     if present */

\n|. {
  yoxOutG.lex ().put ((char)chr ());
  start (YACC_TRAILER);
}

}



<YACC_TRAILER>{

  /* echo the optional user section */

"/*" {
  yoxOutG.put (text ());
  yox_push_state (COMMENT, yoxOutG);
}

"L"?"\"" {
  yoxOutG.put (text ());
  yox_push_state (STRING, yoxOutG);
}

"//".*$|{WS}|{NUMBER}|{ID}|{CHAR_CONST} {
  yoxOutG.put (text ());
}

\n|. {
  yoxOutG.put ((char)chr ());
}

}



<COMMENT>{

  /* 'subroutine': handle multi-line C comments */

"*/" {
  whereTo->put (text ());
  pop_state ();
}

[^\*\n]+ {
  whereTo->put (text ());
}

\n|. {
  whereTo->put ((char)chr ());
}

<<EOF>> {
  unterminatedComment (yylloc, startPos.top ());
}

}



<SKIP_COMMENT>{

  /* 'subroutine': skip multi-line C comments */

"*/" {
  pop_state ();
}

[^\*\n]+|\n|. {
  ; // no-op
}

<<EOF>> {
  unterminatedComment (yylloc, startPos.top ());
}

}



<STRING>{

  /* 'subroutine': handle multi-line C/C++ string literals */

"\"" {
  whereTo->put ((char)chr ());
  if (stringHasBareNewline)
    stringContainsNewline (yylloc, startPos.top ());
  pop_state ();
}

[^"\\\n]+|\\(.|\n) {
  whereTo->put (text ());
}

\n {
  stringHasBareNewline = true;
  whereTo->put ((char)chr ());
}

. {
  whereTo->put ((char)chr ());
}

<<EOF>> {
  unterminatedString (yylloc, startPos.top ());
}

}



<CPP_DIRECTIVE>{

  /* 'subroutine': handle C/C++ directives */

\\\n {
  // C/C++ directive spans multiple source lines
  whereTo->put (text ());
}

\n {
  whereTo->put ((char)chr ());
  pop_state ();
}

. {
  whereTo->put ((char)chr ());
}

<<EOF>> {
  unterminatedCPPdirective (yylloc, startPos.top ());
}

}



<<EOF>> {
  return parser::token::END;
}

%%

// Y-file lexer constructor
AGC::YFILE::Lexer::Lexer (Yfile_specIO& fR, std::ifstream& yoxInG)
    : theFiles (fR),
      EOLstate (EOLctx (position (&fR.inName, 0, 0))),
      ctrl (outMgr::ctrl)
{
  yycolumn = 1;

  nLbraces = 0;
  nLTbraces = 0;
  nLbrackets = 0;
  nLparens = 0;
  nLparensM = 0;
  nATLbraces = 0;

  yyrestart (yoxInG);

  start (YACC_DEFS);
}

namespace AGC
{
namespace YFILE
{
#if defined(_WIN32)
#include "YY_INPUT.hh"
#endif

void
Lexer::yox_push_state (int lexState, outMgr& whTo)
{
  whereTo = &whTo;
  if (lexState == STRING)
    stringHasBareNewline = false;
  push_state (*whereTo->getCurPos (), lexState);
}

}
}
