/* $Id: mac.hh 2834 2022-12-10 21:33:22Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 1992-1994  Kurt M. Bischoff
   Copyright (C) 2004-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#ifndef MAC_HH_INCLUDED
#define MAC_HH_INCLUDED

#include <vector>

#include "location.hh"

#include "memMgr.hh"
#include "outMgr.hh"


typedef std::vector<std::string> formalList;

class macroDef : public memMgr
{
public:

  macroDef ()
      : name (std::string ()),
        formals (formalList ())
  {}

  macroDef (const std::string n)
      : name (n),
        formals (formalList ())
  {}

  int macroFormalLookup (const std::string&) const;

  bool installMacroFormal (const std::string&);

  void close (void);

  std::string name;
  formalList formals;
  std::string body;

protected:
private:
};

extern macroDef macro;


extern const macroDef& getCurMacro (void);


extern bool macroLookup (const std::string&);



class macroArg : public memMgr
{
public:

  macroArg (bool forLexer = false)
      : memMgr (forLexer)
  {}

  macroArg (const AGC::position pos, bool forLexer = false)
      : memMgr (forLexer)
  {
    origin = pos;
  }

  void terminate (void);

protected:
private:
};



extern void expandMacro (const std::string&, const AGC::position);

extern void setupMacroExpansion (const AGC::location&,
                                 std::string&,
                                 AGC::position&);
extern bool expandingMacro (void);
extern void macroDone (void);

extern const std::string& substFormal (int);

extern void expandFormalsInAttrOccr (std::string&);

extern void setupFormalExpansion (const AGC::location&);
extern bool replacingFormal (void);
extern void macroFormalDone (void);


extern const std::string& macroName (void);
extern int macroInvokedFrom (void);
extern const AGC::location& formalLoc (void);

extern bool initMacroDef (const std::string&);

#endif // !MAC_HH_INCLUDED
