/* Infix notation calculator--calc */
%pure-parser

%{
#include "types.h"
%}

%union{
/* `bison' v2.3 hack:
   Ox 1.5 ensures that any literal code blocks that follow a '%union'
   declaration in the parser specification also occur after the
   Ox-generated '%union' declaration.  In the case of old versions of
   `bison' , this dummy '%union' declaration ensures that the code block
   below which needs 'YYSTYPE' will occur in the generated parser
   C source following the '#include "parser.h"'. */
#ifndef __GNUC__
int dummy; /* avoid empty union if not GNU C/C++ */
#endif
}

%{
#include <assert.h>

static int power (int base, int exponent);

#include <stdio.h>

static void yyerror (const char *msg);
extern int yylex (YYSTYPE *lvalp);
%}

/* Grammar Declarations */
%token ERROR
%token NUM

%nonassoc '='   /* comparison            */
%left '-' '+'
%left '*' '/'
%nonassoc NEG   /* negation--unary minus */
%right '^'      /* exponentiation        */

@attributes {semantic_error eVal;} ERROR
@attributes {semantic_value iVal;} exp NUM
@attributes {int print;} line

%%
input	:	line
	|	input line
	;

line	:	'\n'		@{ @i @line.print@ = 0; @}
	|	exp '\n'	@{ @decorate
				   @i @line.print@ = fprintf(stderr, "semval = %d\n", @exp.iVal@);
				@}
	;

exp	:	NUM		@{ @i @exp.iVal@ = @NUM.iVal@; @}
	|	exp '=' exp	@{ @i @exp.0.iVal@ = (@exp.1.iVal@ == @exp.2.iVal@); @}
	|	exp '+' exp	@{ @i @exp.0.iVal@ = @exp.1.iVal@ + @exp.2.iVal@; @}
	|	exp '-' exp	@{ @i @exp.0.iVal@ = @exp.1.iVal@ - @exp.2.iVal@; @}
	|	exp '*' exp	@{ @i @exp.0.iVal@ = @exp.1.iVal@ * @exp.2.iVal@; @}
	|	exp '/' exp	@{ @i @exp.0.iVal@ = @exp.1.iVal@ / @exp.2.iVal@; @}
	|	'-' exp %prec NEG @{ @i @exp.0.iVal@ = -@exp.1.iVal@; @}
	|	exp '^' exp	@{ @i @exp.0.iVal@ = power(@exp.1.iVal@, @exp.2.iVal@); @}
	|	'(' exp ')'	@{ @i @exp.0.iVal@ = @exp.1.iVal@; @}
	|	ERROR		@{ @e exp.iVal : ERROR.eVal ;
				   {@exp.iVal@ = 1111;
				    fprintf(stderr, "gc: invalid input: [%c]\n", @ERROR.eVal@);}
				@}
	;
%%

static int
power (int base, int exponent)
{
  int res = 1;
  assert (0 <= exponent);
  for (/* Niente */; exponent; --exponent)
    res *= base;
  return res;
}

static
void yyerror (const char *msg)
{
  fprintf(stderr, "%s\n", msg);
}
