%language "C++"
%require "3.5.4"
%defines

%glr-parser

%define api.token.raw

%define api.value.type union
%define parse.assert

%code requires { // -> header file, before the definition of "class parser"

#include <string>

class driver;

}

%param { driver& drv } // The parsing context

%locations
%define api.location.file "location.hh"

%define parse.trace
%define parse.error verbose

%code {

#include "driver.hh"

// `yylex()' for the parser's use
#if defined (USE_FLEXCXX)

#define yylex drv.yylex

#else

#include "lexer.hh"

extern YY_DECL;

#endif

}

%define api.token.prefix {TOK_}
%token
  END  0  "end of file"
  ASSIGN  ":="
  MINUS   "-"
  PLUS    "+"
  STAR    "*"
  SLASH   "/"
  LPAREN  "("
  RPAREN  ")"
;

%token IDENTIFIER "identifier"
%token <int> NUMBER "number"
%nterm <int> exp

%printer { yyo << $$; } <*>;

%start unit

%left "+" "-"
%left "*" "/"

@attributes { int Val; } exp "number"
@attributes { std::string Name; } "identifier"
@attributes { @void stg; } assignment unit

%%

unit: assignments exp  @{ @m @unit.stg@; drv.result = @exp.Val@; @};

assignments:
  %empty
| assignments assignment;

assignment:
  "identifier" ":=" exp @{ @m @assignment.stg@; drv.variables[@"identifier".Name@] = @exp.Val@; @};

exp:
  "number"	@{ @i @exp.Val@ = @"number".Val@; @}
| "identifier"  @{ @i @exp.Val@ = drv.variables[@"identifier".Name@]; @}
| exp "+" exp   @{ @i @exp.0.Val@ = (@exp.1.Val@ + @exp.2.Val@); @}
| exp "-" exp   @{ @i @exp.0.Val@ = (@exp.1.Val@ - @exp.2.Val@); @}
| exp "*" exp   @{ @i @exp.0.Val@ = (@exp.1.Val@ * @exp.2.Val@); @}
| exp "/" exp   { 
		  if ($3 == 0)
		    yyparser.error
		      (@3, "invalid division by zero");
		}
		@{ @i @exp.0.Val@ = (@exp.1.Val@ / @exp.2.Val@); @}
| "(" exp ")"   @{ @i @exp.0.Val@ = @exp.1.Val@; @};
%%

#include <iostream>

void yy::parser::error (const YYLTYPE& l, const std::string& m) {

  std::cerr << l << ": " << m << '\n';

}
