%option nodefault noinput nounput noyywrap

%{
#include <errno.h>
#include <limits.h>
#include <stdlib.h>

#include "parser.h"
%}

%%

"="	return TOK_EQ;
"+"	return TOK_PLUS;
"-"	return TOK_MINUS;
"*"	return TOK_STAR;
"/"	return TOK_SLASH;
"^"	return TOK_UP;

"("	return TOK_LPAREN;
")"	return TOK_RPAREN;

[0-9]+	{
	 errno = 0;
	 long n = strtol (yytext, NULL, 10);
	 if (! (INT_MIN <= n && n <= INT_MAX && errno != ERANGE))
	   {yyerror (nerrs, "integer is out of range");
	    if (n < 0) n = INT_MIN; else n = INT_MAX;
	   }
	 yylval->TOK_NUM = (int) n;
	 return TOK_NUM;
	 @{ @TOK_NUM.iVal@ = (int) n; @}
	}

[ \t\f]+ continue;

"\n"	return TOK_EOL;

.        yyerror (nerrs, "syntax error, invalid character: %c", yytext[0]);

%%
