%{
#include <cstdlib>
#include <cstring>

#include "types.hh"
#include "util.hh"
#include "parser.hh"
%}

IDENT ([a-zA-Z_][a-zA-Z0-9_]*) 

%% 
global         { yylval.cVal = cGLOBAL; return(GLOBAL);
                 @{ @GLOBAL.cVal@  = cGLOBAL; @}
               }
local          { yylval.cVal = cLOCAL; return(LOCAL);
                 @{ @LOCAL.cVal@   = cLOCAL; @}
               }
real           { yylval.tVal = tREAL; return(REAL);
                 @{ @REAL.tVal@    = tREAL; @}
               }
integer        { yylval.tVal = tINTEGER; return(INTEGER);
                 @{ @INTEGER.tVal@ = tINTEGER; @}
               }
\;             return(';');
[ \n\t\f]*     ;  
{IDENT}        { yylval.id = strdup(yytext); return(NAME);
                 @{ @NAME.id@ = strdup(yytext); @}
               }
.              fprintf(stderr,"line %d: illegal character '%c' [%#x]\n",yylineno,yytext[0],yytext[0]); 
%%
int yywrap(void){return 1;}

extern int yydebug;
extern int yyparse(void);

int
main (int argc, char *argv[])
{
#if YYDEBUG
  if (getenv("YYDEBUG") != 0) yydebug = 1;
#endif

  if (argc > 1)
    yyin = fopen (argv[argc-1], "r");
  else
    yyin = stdin;

  if (yyparse() == 0) {
    if (cmpYaccToOxDeclList()) {
      fputs("Yacc and Ox declaration_lists are identical\n",stdout);
    } else {
      dumpDeclList(btYaccDeclList);
      dumpDeclList(oxDeclList);
    }
  }
}
