%top {

// ensure that "system" includes precede
// includes in the lexer skeleton

#include <cerrno>
#include <climits>
#include <cstdlib>
#include <string>

}

%{
#include "driver.hh"
#include "parser.hh"

void yyerror (const char *, ...);
%}

%option noyywrap nounput noinput batch debug
%option yylineno

id    [a-zA-Z][a-zA-Z_0-9]*
int   [0-9]+
blank [ \t\r]

%%

{blank}+	;
\n+		;

"-"		return '-';
"+"		return '+';
"*"		return '*';
"/"		return '/';
"("		return '(';
")"		return ')';
":="		return ASSIGN;

{int}		{ long n = strtol (yytext, NULL, 10);
		  if (! (INT_MIN <= n && n <= INT_MAX && errno != ERANGE))
		    { yyerror (("integer is out of range: " +
			       std::string(yytext)).c_str());
		      if (n < 0) n = INT_MIN; else n = INT_MAX;
		    }
		  yylval.val =  (int) n;
		  return NUMBER;
		  @{ @NUMBER.Val@ = (int) n; @}
 		}
{id}		{ return IDENTIFIER;
		  @{ @IDENTIFIER.Name@ = strdup (yytext); @}
		}
.		{ yyerror (("invalid character: " +
			   std::string(yytext)).c_str());
		}
%%

#include <iostream>

void
yyerror(const char *s, ...)
{
#if USE_FLEXCXX

 std::cerr << ::drv.file << ':' << ::drv.theScanner->lineno () << ": " << s << '\n';

#else

 std::cerr << ::drv.file << ':' << yylineno << ": " << s << '\n';

#endif

 yyyabort (); 
}
