%option bison-bridge bison-locations
%{
#include <string.h>

#include "types.h"
#include "parser.h"
#include "util.h"

static int yycolno=1;

#define set_yylloc(width) {yylloc->first_line    = yylineno; \
                           yylloc->first_column  = yycolno; \
                           yylloc->last_line     = yylineno; \
                           yylloc->last_column   = ((yycolno+=width)-1);}
%}

IDENT ([a-zA-Z_][a-zA-Z0-9_]*) 

%% 

\n         yycolno=1;  
[ \t\f]    yycolno++;  
const      %{set_yylloc(yyleng); return(CONST);%}
type       %{set_yylloc(yyleng); return(TYPE);%}
[0-9]+     %{set_yylloc(strlen(yytext));
            sscanf(yytext,"%d",&yylval->iVal);
            return(NUM);
            @{ sscanf(yytext,"%d",&@NUM.iVal@);
               @NUM.loc@ = *yylloc;
            @}
           %}
\(         %{set_yylloc(1); return('(');
            @{ @'('.loc@ = *yylloc; @}
           %}
\)         %{set_yylloc(1); return(')');
            @{ @')'.loc@ = *yylloc; @}
           %}
\,         %{set_yylloc(1); return(',');%}
\;         %{set_yylloc(1); return(';');%}
\+         %{set_yylloc(1); return('+');
            @{ @'+'.loc@ = *yylloc; @}
           %}
\-         %{set_yylloc(1); return('-');
            @{ @'-'.loc@ = *yylloc; @}
           %}
\=         %{set_yylloc(1); return('=');%}
\.\.       %{set_yylloc(2); return(DOTDOT);%}
{IDENT}    %{set_yylloc(strlen(yytext));
            yylval->sVal = addYaccSymbolIfNew(yylloc,yytext);
            return(ID);
            @{@ID.loc@ = *yylloc;
              @ID.sym@ = addOxSymbolIfNew(g_SymbolTable(),yylloc,yytext);
            @}
           %}
\{         eat_comment("}",&yycolno,&yylineno);
\(\*       eat_comment("*)",&yycolno,&yylineno);
.          %{fprintf(stderr,"line %d:%d illegal character\n",yylineno,yycolno); yycolno++;%}

%%
#if !defined(FLEX_SCANNER) || !defined(__cplusplus)
int yyyinput(void){return input();} /* `lex', or `flex' with C compiler */
#else
int yyyinput(void){return yyinput();} /* `flex' with C++ compiler */
#endif

int yywrap(void){return 1;}

extern int yydebug;
extern int yyparse(void);

int
main (int argc, char *argv[])
{
#if YYDEBUG
  if (getenv("YYDEBUG") != 0) yydebug = 1;
#endif

  if (argc > 1)
    yyin = fopen (argv[argc-1], "r");
  else
    yyin = stdin;

  if (yyparse() == 0) {
    if (cmpYaccToOxSymbolTable()) {
      fputs("Yacc and Ox symbol tables are identical\n",stdout);
    } else {
      dumpYaccSymbolTable();
      dumpOxSymbolTable();
    }
  }
}
