/* RCS Log: gram.y,v
 * Revision 1.5  1990/04/28  11:58:42  kelvin
 * Modified the dclstate DFA to allow pointers to functions as parameters.
 *
 * Revision 1.4  89/03/31  10:43:46  kelvin
 * added production for anonymous structure definitions.
 * 
 * Revision 1.3  89/03/26  18:31:48  kelvin
 * Changed Header to Log.
 * 
 */

/*
  tiny - C
*/

/* types of identifiers */
%token TK_IDENTIFIER

/* types of constants */
%token <i> TK_ICONSTANT 

/* simple operators */
%token TK_PTR TK_INC TK_DEC TK_LE TK_GE TK_EQ TK_NE


/* keywords */
%token TK_EXTERN 
%token TK_INT 
%token TK_STRUCT
%token TK_IF TK_ELSE TK_WHILE TK_DO TK_FOR TK_RETURN TK_BREAK


%{
#include <ctype.h>

#include "sym.h"

#if YYBTYACC && !YYBYACC
/* the `btyacc' 3.0 skeleton omits the following */
#include "parser.h"
#endif
  
/* The grammar actually allows a larger language than is legal,
in particular function declarations are slipped in with other declarations.
The following are states for the finite automata that will keep track
of function declarations, making sure everything is legal */

enum dclcls {globaldcl, varsdcl, posfundcl, fundcl, funbody};

/* initial state is looking at global declarations */
enum dclcls dclstate = globaldcl;

/* ---- put your global declarations here ------ */

void check_fstate(void);
void check_dcl(enum dclcls);

extern int yylex(void);

#if YYBTYACC && !YYBYACC
extern void yyerror(const char *, ...);
#else
extern void yyerror(const char *);
#endif
%}

/* This declares the attribute stack entry */
%union {
	double			d;
	int			i;
	char		       *s;
	struct typestruct      *t;
	unsigned int		u;
	struct sym             *y;
	struct node            *n;
	}

%type <i> constant_expr

%start file		/* start symbol is file */
%%

file
	: external_definition
	| file external_definition
	;

external_definition
	: declaration			{ dclstate = globaldcl; }
	| function_definition		{ dclstate = globaldcl; }
	;

declaration
	/* This syntax is only permitted when declaring struct/union */
	: declaration_specifiers ';'
	| declaration_specifiers init_declarator_list ';'
	;

function_definition
	: function_head function_body
	;

function_head
	: declarator
		{ check_fstate(); }
	| declaration_specifiers declarator
		{ check_fstate(); }
	;

				/* returns a base type */
declaration_specifiers
	: ssc_specifier		/* this is only permitted when
				 *  dclstate == globaldcl
				 *  and ssc_specifier is extern
				 *  type defaults to int
				 */
		{ 
		  if (dclstate != globaldcl)
		    yyerror("illegal omission of type specifier");
		}
	| ssc_specifier type_specifier_list
	| type_specifier_list
	;

ssc_specifier
	: TK_EXTERN
	;

type_specifier_list
	: type_specifier
	;

type_specifier
	: TK_INT
	| struct_or_union_specifier
	;

/* struct_or_union_specifier should also accept struct_or_union '{' ... '}' */
struct_or_union_specifier
	: struct_def struct_declaration_list '}'
	| struct_or_union TK_IDENTIFIER
	;

struct_def
	: struct_or_union TK_IDENTIFIER '{'
	| struct_or_union '{'
	;

/* should permit unions too */
struct_or_union
	: TK_STRUCT
	;

struct_declaration_list
	: struct_declaration
	| struct_declaration_list struct_declaration
	;

struct_declaration
	: stype struct_declarator_list ';'
	;

stype
	: type_specifier_list
	;

struct_declarator_list
	: struct_declarator
	| struct_declarator_list ',' struct_declarator
	;

struct_declarator
	: sdeclarator
	;

sdeclarator
	: stars sdeclarator2
	;

stars
	: '*' stars
	|				/* null string */
	;

sdeclarator2
	: TK_IDENTIFIER
	| leftparen sdeclarator ')'
	| sdeclarator2 '[' constant_expr ']'
	;

leftparen
	: '('
	;

init_declarator_list
	: init_declarator
	| init_declarator_list ',' init_declarator      
		{
		  if (dclstate != fundcl)
		    check_dcl(varsdcl);
		}
	;

init_declarator
	: declarator
	;

declarator
	: declarator1
	;

declarator1
	: stars declarator2
	;

declarator2
	: TK_IDENTIFIER
	| leftparen declarator ')'
	| declarator2 '[' ']'
	| declarator2 '[' constant_expr ']'
	| declarator2 '(' ')'		
			{ check_dcl(posfundcl); }
	| declarator2 '(' parameter_declaration_list ')'
			{ check_dcl(fundcl); }
	;

parameter_declaration_list
	: identifier_list
	;

identifier_list
	: TK_IDENTIFIER
	| identifier_list ',' TK_IDENTIFIER
	;

function_body
	: param_declarations function_statement
	;

param_declarations
	:  /* nothing */		{ dclstate = funbody; }
	|
		{
		  dclstate = funbody;
		}
          declaration_list
	;

declaration_list
	: declaration
	| declaration_list declaration
	;

function_statement
	: '{' function_declarations statement_list '}'
	;

function_declarations
	: /* empty */
	| declaration_list
	;

statement_list
	:			/* empty string */
	| statement_list statement
	;

statement
	: compound_statement
	| expression_statement
	| selection_statement
	| iteration_statement
	| jump_statement
	;

compound_statement
	: '{' statement_list '}'
	;

expression_statement
	: ';'
	| expr ';'
	;

selection_statement
	: TK_IF '(' expr ')' statement
	| TK_IF '(' expr ')' statement TK_ELSE statement
	;

iteration_statement
	: TK_WHILE '(' expr ')' statement
	| TK_DO statement TK_WHILE '(' expr ')' ';'
	| TK_FOR '(' opte ';' opte ';' opte ')' statement
	;

opte
	: /* nothing */
	| expr
	;

jump_statement
	: TK_RETURN ';'
  	| TK_RETURN expr ';'
      	| TK_BREAK ';'
	;

primary_expr
	: TK_IDENTIFIER
	  		{
			  /* dump the symbol */
			  struct sym *vp;
			  extern struct sym *lookup();
			  extern void vdisplay(struct sym *);

			  if ((vp = lookup($<s>1)))
			    vdisplay(vp);
			  else
			    fprintf(stderr, "%s: not defined\n", $<s>1);
			}
	| TK_ICONSTANT
	| '(' expr ')'
	| primary_expr '[' expr ']'
	| primary_expr '(' ')'
	| primary_expr '(' argument_expr_list ')'
	| primary_expr '.' TK_IDENTIFIER
	| primary_expr TK_PTR TK_IDENTIFIER
	;

argument_expr_list
	: assignment_expr
	| argument_expr_list ',' assignment_expr
	;

postfix_expr
	: primary_expr
	| primary_expr TK_INC
	| primary_expr TK_DEC
	;

unary_expr
	: postfix_expr
	| TK_INC unary_expr
	| TK_DEC unary_expr
	| unary_operator cast_expr
	;

unary_operator
	: '&'
	| '*'
	| '+'
	| '-'
	| '!'
	;

cast_expr
	: unary_expr
	;

multiplicative_expr
	: cast_expr
	| multiplicative_expr '*' cast_expr
	| multiplicative_expr '/' cast_expr
	| multiplicative_expr '%' cast_expr
	;

additive_expr
	: multiplicative_expr
	| additive_expr '+' multiplicative_expr
	| additive_expr '-' multiplicative_expr
	;

shift_expr
	: additive_expr
	;

relational_expr
	: shift_expr
	| relational_expr '<' shift_expr
	| relational_expr '>' shift_expr
	| relational_expr TK_LE shift_expr
	| relational_expr TK_GE shift_expr
	;

equality_expr
	: relational_expr
	| equality_expr TK_EQ relational_expr
	| equality_expr TK_NE relational_expr
	;

and_expr
	: equality_expr
	;

exclusive_or_expr
	: and_expr
	;

inclusive_or_expr
	: exclusive_or_expr
	;

logical_and_expr
	: inclusive_or_expr
	;

logical_or_expr
	: logical_and_expr
	;

conditional_expr
	: logical_or_expr
	;

assignment_expr
	: conditional_expr
	| unary_expr assignment_operator assignment_expr
	;

/* should include all augmented assignment operators */
assignment_operator
	: '='
	;

expr
	: assignment_expr
	;

constant_expr
	: TK_ICONSTANT
	;

%%

#include <stdio.h>

extern char yytext[];

static int errcount = 0;

extern void init_sym(void);
int main(void)
{
	/* put any initialization here */

  init_sym();

  yyparse();
}

/* check_dcl - see if a declaration state transition is legal */
void check_dcl(enum dclcls newstate)
{
  switch(newstate) {
  case globaldcl: break; /* doesn't occur */
  case funbody: break; /* doesn't occur */
  case varsdcl:
    if (dclstate == funbody)
      dclstate = funbody;
    else if ((dclstate == globaldcl) || (dclstate == varsdcl))
      dclstate = varsdcl;
    else
      yyerror("invalid function declaration");
    break;
  case posfundcl:
    if (dclstate == funbody)
      dclstate = funbody;
    else if (dclstate == varsdcl)
      dclstate = varsdcl;
    else if (dclstate == globaldcl || dclstate == posfundcl)
      dclstate = posfundcl;
    else
      yyerror("invalid function declaration");
    break;
  case fundcl:
    if (dclstate == globaldcl)
      dclstate = fundcl;
    else
      yyerror("invalid function declaration");
  }
}

/* check fstate - make sure it is legal for a function to begin */
void check_fstate(void)
{
  if ((dclstate != posfundcl) && (dclstate != fundcl))
    yyerror("invalid function declaration (missing semicolon?)");
  else
    dclstate = fundcl;
}


