/* $Id: code.c 2579 2022-06-23 17:54:48Z teshields $ */

#include "gppl_glob.h"
#include "parser.h"
#include "util.h"

#define ALIGNSIZE   4
#define STACKSIZE1 500000L
#define STACKSIZE  (ceiling(STACKSIZE1,ALIGNSIZE)) 

#define nFREGS 3 
#define RETREGADDR1   -2    /* relative to the callee's fp */
#define RETREGADDR2   -4    /* relative to sp upon return  */ 

/* RA, VA, PA are for calculations of frame-pointer-relative addresses */
#define RA(regNum)       (-(nFREGS + (regNum)))
#define VA(variableAddr) (-(nFREGS + dumEnv->nRegs + (variableAddr))) 
#define PA(paramAddr)    ((paramAddr) + 1 + 1)


enum pred doCodeGen = True;

FILE *tempOut; 



long stgSize(enum symType type, long howMany) 
  {return(howMany);
  } 



static labelType labelSeq = 0; 
static labelType returnLabelSeq = 0; 



labelType newLabelNum(void) 
  {return (labelSeq++); 
  }



labelType newReturnLabelNum(void) 
  {return (returnLabelSeq++); 
  }

 

static const char *cHeaderCode[] =
{ 
"\n",
"#include <stdlib.h>\n",
"#include <stdio.h>\n",
"\n",
"union stackItem {\n",
"   int i; \n",
"   float f;\n",
"   int b; \n",
"   int *ip;\n",
"   float *fp;\n",
"   int *bp; \n",
"   union stackItem *up;\n",
"  };\n",
"\n",
"\n",
"int main()\n",
"{\n",
" register union stackItem *sp,*fp;\n",
"\n",
" if ((sp = (union stackItem *)malloc(STACKSIZE * sizeof(union stackItem)))\n",
"     ==\n",
"     ((union stackItem *)NULL)\n",
"    )\n",
"    fprintf(stderr,\"malloc error in stack allocation\\n\");\n",
" sp = sp + STACKSIZE;\n",
"\n",
}; 



void cgHeader(void) 
  {if (doCodeGen == False) return; 
   fprintf(tempOut,"#define STACKSIZE %ld\n",STACKSIZE); 
   printStringArray(cHeaderCode,sizeof(cHeaderCode),tempOut); 
  }



void cgTrailer(void) 
  {long i;
   
   if (doCodeGen == False) return; 
   fprintf(tempOut,"\n\texit(0);\n");  
   fprintf(tempOut,"rts:\n");
   fprintf(tempOut,"  switch ((*(sp++)).i)  \n    {\n");
   for (i=0; i<returnLabelSeq; i++) 
       fprintf(tempOut,"     case %4ld: goto r%ld;\n",i,i); 
   fprintf(tempOut,"    } /* rts switch */ \n"); 
   fprintf(tempOut,"IOBtrap:\n"); 
   fprintf(tempOut,"\tfprintf(stderr,\"array index out of bounds.\\n\");\n"); 
   fprintf(tempOut,"\texit(-1);\n"); 
   fprintf(tempOut,"} /* main */ \n"); 
  }



void cgFuncCall(struct env *env,
                struct sym *funcSym,
                regNumType  regNum,
                enum pred   isEmptyActList,
                lineNumType line) 
  {int retLabel; 
   int staticDist,i;
   struct sym *dum;

   if ((doCodeGen == False) || (funcSym == lambdaSym)) return; 
   if (isEmptyActList)
      {dum = funcSym->paramLoc;
       while (dum != lambdaSym)
         {fprintf(stderr,"missing actual parameter in line %d's call to %s.\n",
                         line,funcSym->id
                 ); 
          handleSemError();
          dum = dum->next;
         }
      }
   staticDist = env->level - funcSym->level;
   retLabel = newReturnLabelNum(); 
   fprintf(tempOut,"\t(*--sp).i = %d;\n",retLabel);  
   fprintf(tempOut,"\t(*--sp).up = fp;\n");  
   for (i=0; i<staticDist; i++) 
       fprintf(tempOut,"\t*sp = *(*sp).up;\n");  
   fprintf(tempOut,"\tgoto f%d;\n",funcSym->label);  
   fprintf(tempOut,"r%d:\n",retLabel); 
   if (funcSym->type != VOIDT)
      fprintf(tempOut,"\tfp[%d] = sp[%d];\n",RA(regNum),RETREGADDR2); 
   fprintf(tempOut,"\tsp += %ld;\n",funcSym->paramsSize);  
  }



void cgBlockEnter(struct env *env) 
  {if (doCodeGen == False) return; 
   if (env != lambdaEnv) 
      {if (env->funcSym == lambdaSym) 
          fprintf(tempOut,"\t(*--sp).up = fp;\n"); 
       fprintf(tempOut,"\t(*--sp).up = fp;\n"); 
       fprintf(tempOut,"\tfp = sp + 1;\n"); 
       fprintf(tempOut,"\tsp -= %ld;\n",
               1 + env->nRegs + env->varsSize 
              ); 
      }
  }



void cgBlockExit(struct env *env) 
  {if (doCodeGen == False) return; 
   if (env == lambdaEnv) return;
   fprintf(tempOut,"BE%d:\tsp = fp;\n",env->exitLabel); 
   fprintf(tempOut,"\tfp = (*(sp++ - 1)).up;\n"); 
  }



void cgFuncAddress(struct sym *sym) 
  {if (doCodeGen == False) return; 
   if (sym != lambdaSym) 
      {fprintf(tempOut,"\tgoto j%d;\n",sym->label); 
       fprintf(tempOut,"f%d:\t\t\t\t\t\t/* %s */\n",sym->label,sym->id); 
      }
  }



void cgFuncTrailer(struct sym *sym) 
  {if (doCodeGen == False) return; 
   fprintf(tempOut,"\tgoto rts;\n"); 
   fprintf(tempOut,"j%d:\n",sym->label); 
  }



void cgBinArithExpr(int destReg, int op, enum symType type) 
  {const char *typeExt;  

   if (doCodeGen == False) return; 
   switch(type) 
     {case INTT:    typeExt = "i";  break;
      case FLOATT:  typeExt = "f";  break;
      default:                      break;
     }
   fprintf(tempOut,"\tfp[%d].%s %c= fp[%d].%s;\n",
           RA(destReg),
           typeExt,
           op,
           RA(destReg + 1),
           typeExt 
          );
  }



void cgRelExpr(int destReg, int op, enum symType type) 
  {const char *typeExt,*opStr;  
   labelType label;

   if (doCodeGen == False) return; 
   switch(type) 
     {case INTT:    typeExt = "i";  break;
      case FLOATT:  typeExt = "f";  break;
      default:                      break;
     }
   switch (op) 
     {case EQ:  opStr = "=="; break;
      case LEQ: opStr = "<="; break;
      case GEQ: opStr = ">="; break;
      case NEQ: opStr = "!="; break;
      case LT:  opStr = "<"; break;
      case GT:  opStr = ">"; break;
     } 
   label = newLabelNum(); 
   fprintf(tempOut,"\tif (fp[%d].%s %s fp[%d].%s) goto b%d;\n",
           RA(destReg),typeExt,opStr,
           RA(destReg + 1),typeExt,label
          );
   fprintf(tempOut,"\tfp[%d].i = 0;\n",RA(destReg)); 
   fprintf(tempOut,"\tgoto B%d;\n",label); 
   fprintf(tempOut,"b%d:\tfp[%d].i = 1;\n",label,RA(destReg)); 
   fprintf(tempOut,"B%d:\n",label); 
  }



void cgBinLogExpr(int destReg, int op) 
  {labelType label;
   char unrOp,rslt1,rslt2; 

   if (doCodeGen == False) return; 
   label = newLabelNum(); 
   switch(op)
     {case AND: unrOp = '!'; rslt1 = '1'; rslt2 = '0'; break;  
      case OR : unrOp = ' '; rslt1 = '0'; rslt2 = '1'; break;  
     }
   fprintf(tempOut,"\tif (%c(fp[%d].i)) goto b%d;\n",
           unrOp,RA(destReg),label
          );
   fprintf(tempOut,"\tif (%c(fp[%d].i)) goto b%d;\n",
           unrOp,RA(destReg + 1),label
          );
   fprintf(tempOut,"\tfp[%d].i = %c;\n",RA(destReg),rslt1); 
   fprintf(tempOut,"\tgoto B%d;\n",label); 
   fprintf(tempOut,"b%d:\tfp[%d].i = %c;\n",label,RA(destReg),rslt2); 
   fprintf(tempOut,"B%d:\n",label); 
  }



void cgUnrArithExpr(int destReg, int op, enum symType type) 
  {const char *typeExt;  

   if (doCodeGen == False) return; 
   if (op == '+') return; 
   switch(type) 
     {case INTT:    typeExt = "i";  break;
      case FLOATT:  typeExt = "f";  break;
      default:                      break;
     }
   fprintf(tempOut,"\tfp[%d].%s = -fp[%d].%s;\n",
           RA(destReg),typeExt,RA(destReg),typeExt 
          );
  }



void cgLogNegExpr(int regNum) 
  {labelType label;
   
   if (doCodeGen == False) return; 
   label = newLabelNum(); 
   fprintf(tempOut,"\tif ((fp[%d].i) == 0) goto b%d;\n",
           RA(regNum),label
          );
   fprintf(tempOut,"\tfp[%d].i = 0;\n",RA(regNum)); 
   fprintf(tempOut,"\tgoto B%d;\n",label); 
   fprintf(tempOut,"b%d:\tfp[%d].i = 1;\n",label,RA(regNum)); 
   fprintf(tempOut,"B%d:\n",label); 
  }



void cgEmptyIndexSeq(struct env *env,
                     struct sym *sym,
                     lineNumType line,
                     int         regNum,
                     struct dim *dimList) 
  {long addr; 
   int staticDist,i;
   struct env *dumEnv;
   
   if (doCodeGen == False) return; 
   if (sym == lambdaSym) return; 
   staticDist = env->level - sym->level;
   while (dimList != lambdaDim) 
      {fprintf(stderr,
               "missing index for line %d\'s reference to array %s.\n",
               line,sym->id
              ); 
       dimList = dimList->next;
       handleSemError();   
      }
   fprintf(tempOut,"\tfp[%d].up = fp;\n",RA(regNum)); 
   dumEnv = env; 
   for (i=0; i<staticDist; i++) 
       {dumEnv = dumEnv->outerEnv;
        fprintf(tempOut,"\tfp[%d] = *(fp[%d].up);\n",RA(regNum),RA(regNum)); 
       }
   addr = (sym->Class == VARIABLEC)?(VA(sym->addr)):(PA(sym->addr));
   fprintf(tempOut,"\tfp[%d].up += %ld;\n",RA(regNum),addr); 
   if (sym->Class == REFPARAMC)
      fprintf(tempOut,"\tfp[%d] = *(fp[%d].up);\n",
                      RA(regNum),RA(regNum) 
             );
  }



void cgIndexSeq(struct sym *sym, int regNum, struct dim *dimList) 
  {if (doCodeGen == False) return; 
   if (sym == lambdaSym) return; 
   fprintf(tempOut,"\tif (fp[%d].i < 0) goto IOBtrap;\n",
           RA(regNum + 1)
          ); 
   fprintf(tempOut,"\tif (fp[%d].i > %ld) goto IOBtrap;\n",
           RA(regNum + 1),
           dimList->size - 1
          ); 
   fprintf(tempOut,"\tfp[%d].i *= %ld;\n",RA(regNum + 1),dimList->scale); 
   fprintf(tempOut,"\tfp[%d].up -= fp[%d].i;\n",
           RA(regNum),RA(regNum + 1)
          ); 
  }



void cgConstSimpleExpr(int regNum, enum symType type, union num value) 
  {if (doCodeGen == False) return; 
   switch(type)
     {case INTT:    
        fprintf(tempOut,"\tfp[%d].i = %d;\n",RA(regNum),value.i);  
        break;
      case FLOATT:  
        fprintf(tempOut,"\tfp[%d].f = %f;\n",RA(regNum),value.f);  
        break;
      case BOOLT:   
        fprintf(tempOut,"\tfp[%d].b = %d;\n",RA(regNum),value.b);  
        break;
      default:
        break;
     }
  }



void cgEOF(int regNum) 
  {if (doCodeGen == False) return; 
   fprintf(tempOut,"\tfp[%d].i = EOF;\n",RA(regNum));  
  }



void cgAssign(int regNum, enum symType type) 
  {const char *typeExt;
   
   if (doCodeGen == False) return; 
   switch(type) 
     {case INTT:    typeExt = "i";  break;
      case FLOATT:  typeExt = "f";  break;
      case BOOLT:   typeExt = "b";  break;
      default:                      break;
     }
   fprintf(tempOut,"\t*(fp[%d].%sp) = fp[%d].%s;\n",
           RA(regNum),typeExt,RA(regNum + 1),typeExt
          ); 
   fprintf(tempOut,"\tfp[%d].%s = fp[%d].%s;\n",
           RA(regNum),typeExt,RA(regNum + 1),typeExt
          ); 
  }



void cgPush(int regNum) 
  {if (doCodeGen == False) return; 
   fprintf(tempOut,"\t*--sp = fp[%d];\n",RA(regNum));  
  }



void cgCondDeref(struct sym *formal, int regNum, enum symType type) 
  {const char *typeExt;

   if (doCodeGen == False) return; 
   if ((formal != lambdaSym) && (formal->Class == REFPARAMC)) return;
   switch(type) 
     {case INTT:    typeExt = "i";  break;
      case FLOATT:  typeExt = "f";  break;
      case BOOLT:   typeExt = "b";  break;
      default:                      break;
     }
   fprintf(tempOut,"\tfp[%d].%s = *(fp[%d].%sp);\n",
           RA(regNum),typeExt,
           RA(regNum),typeExt
          ); 
  }



void cgCast(int regNum, enum symType newType, enum symType oldType) 
  {const char *ext1,*ext2,*cCast;
  
   if (doCodeGen == False) return; 
   switch(newType) 
     {case INTT:    ext1 = "i"; cCast = "(int)";   break; 
      case FLOATT:  ext1 = "f"; cCast = "(float)"; break; 
      case BOOLT:   ext1 = "b"; cCast = "(int)";   break; 
      default:                                     break;
     }
   switch(oldType) 
     {case INTT:    ext2 = "i";  break;
      case FLOATT:  ext2 = "f";  break;
      case BOOLT:   ext2 = "b";  break;
      default:                   break;
     }
   fprintf(tempOut,"\tfp[%d].%s = %sfp[%d].%s;\n",
           RA(regNum),ext1,cCast,RA(regNum),ext2
          );
  }



void cgOutputNoArg(char *str, regNumType regNum) 
  {if (doCodeGen == False) return; 
   fprintf(tempOut,"\tfp[%d].i = printf(%s);\n",RA(regNum),str);
    /*  fprintf(tempOut,"\tfflush(stdout);\n"); */
  }



void cgOutputOneArg(char *str, int regNum, enum symType type) 
  {if (doCodeGen == False) return; 
   switch(type) 
     {case INTT: 
        fprintf(tempOut,"\tfp[%d].i = printf(%s,fp[%d].i);\n",
                RA(regNum),str,RA(regNum));
        break;
      case FLOATT:    
        fprintf(tempOut,
                "\tfp[%d].i = printf(%s,fp[%d].f);\n",
                RA(regNum),str,RA(regNum)
               );
        break;
      case BOOLT:   
        fprintf(tempOut,
          "\tfp[%d].i = printf(%s,(*(fp[%d].b)?\"true\":\"false\"));\n",
                RA(regNum),str,RA(regNum)
               );
        break;
      default:
        break;
     }
   /*  fprintf(tempOut,"\tfflush(stdout);\n"); */
  }



void cgInput(char *str, int regNum, enum symType type, lineNumType line) 
  {if (doCodeGen == False) return; 
   fprintf(tempOut,"\tfp[%d].i = ",RA(regNum));
   switch(type) 
     {case INTT: 
        fprintf(tempOut,"scanf(%s,fp[%d].ip);\n",
                str,RA(regNum));
        break;
      case FLOATT:    
        fprintf(tempOut,"scanf(%s,fp[%d].fp);\n",
                str,RA(regNum));
        break;
      case BOOLT:   
        fprintf(stderr,"attempted input of boolean near line %d.\n",line); 
        handleSemError(); 
        break;
      default:
        break;
     }
  }



static void condCheck(enum symType type, lineNumType line)
  {if (type != BOOLT) 
      {fprintf(stderr,
   "non-boolean expression as condition in control statement near line %d.\n",
               line
              );
       handleSemError();  
      } 
  } 



void cgIfThen(enum symType type, lineNumType line, labelType label) 
  {condCheck(type,line); 
   if (doCodeGen == False) return; 
   fprintf(tempOut,"e%d:\n",label); 
  }



void cgIfThenElse(enum symType type, lineNumType line, labelType label) 
  {condCheck(type,line); 
   if (doCodeGen == False) return; 
   fprintf(tempOut,"i%d:\n",label); 
  }



void cgThen(int regNum, labelType label) 
  {if (doCodeGen == False) return; 
   fprintf(tempOut,"\tif (!(fp[%d].i)) goto e%d;\n",RA(regNum),label); 
  }



void cgElse(labelType label) 
  {if (doCodeGen == False) return; 
        fprintf(tempOut,"kl%d:\tgoto i%d;\n",label,label); 
        fprintf(tempOut,"e%d:\n",label); 
   /* labels prefixed by kl are kludges to silence the C compiler's
    *   "statement not reached" warnings when the then clause ends
    *   with a return statement.
    */
  }



void cgWdecision(int regNum, labelType label) 
  {if (doCodeGen == False) return; 
   fprintf(tempOut,"\tif (!(fp[%d].i)) goto W%d;\n",RA(regNum),label); 
  }



void cgWhileStatement(enum symType type, lineNumType line, labelType label) 
  {condCheck(type,line); 
   if (doCodeGen == False) return; 
   fprintf(tempOut,"\tgoto w%d;\n",label); 
   fprintf(tempOut,"W%d:\n",label); 
  }



void cgWhile(labelType label) 
  {if (doCodeGen == False) return; 
   fprintf(tempOut,"w%d:\n",label); 
  }



void cgReturnStmt(struct env  *env,
                   regNumType   regNum,
                   enum symType type,
                   lineNumType  line) 
  {if (doCodeGen == False) goto err; 
   if (type != VOIDT) 
      fprintf(tempOut,"\tfp[%d] = fp[%d];\n",RETREGADDR1,RA(regNum)); 
   while (env->funcSym == lambdaSym) 
     {fprintf(tempOut,"\tsp = fp;\n"); 
      fprintf(tempOut,"\tfp = (*fp).up;\n"); 
      if (type != VOIDT) 
         fprintf(tempOut,"\tfp[%d] = sp[%d];\n",RETREGADDR1,RETREGADDR1); 
      if ((env = env->outerEnv) == lambdaEnv)
         {fprintf(stderr,"line %d: return from no function.\n",line); 
          goto err; 
         }
     }
   if (env->funcSym->type != VOIDT)  
      {if (type == VOIDT)
          {fprintf(stderr,
           "line %d: missing object of return in non-void function %s.\n",
                   line,env->funcSym->id 
                  ); 
           goto err; 
          }
       if (env->funcSym->type != type) 
          {fprintf(stderr,
           "line %d: mismatched type of object returned from function %s.\n",
                  line,env->funcSym->id 
                 ); 
           goto err; 
          }
      }
      else
      {if (type != VOIDT)
          {fprintf(stderr,"line %d: object returned by void function %s.\n",
                   line,env->funcSym->id 
                  ); 
           goto err; 
          }
      }
   fprintf(tempOut,"\tgoto BE%d;\n",env->exitLabel); 
   return;
err:
   handleSemError(); 
   return;
  }



void cgExitStmt(regNumType regNum, enum symType type, lineNumType line) 
  {if (doCodeGen == False) return; 
   if (type != INTT) 
      {fprintf(stderr,"line %d: non-integer object of exit statement.\n",
               line
              );
       handleSemError();
       return;
      }
   fprintf(tempOut,"\texit(fp[%d].i);\n",RA(regNum)); 
  }



