/* $Id: scan.c 2098 2021-02-16 02:47:30Z teshields $ */ 

#if defined (__cplusplus)
#include <cstdio>
#include <climits> 
#else
#include <stdio.h> 
#include <limits.h> 
#endif

#include "gppl_glob.h"
#include "util.h"

extern int yylineno; 
#define DFAACCEPT 0 

 

void eatComment(void) 
  {char c; 
   int state = 1; 
   while (state != DFAACCEPT)  
     {c = yyyinput();  
      switch (state)   
        {case 1: switch(c) 
                   {case '*' : state = 2; break;  
                    default : ;
                   } 
                 break;  
         case 2: switch(c) 
                   {case '*' : break;  
                    case '/' : state = DFAACCEPT; 
			       break;  
                    default  : state = 1;  
                   }
                 break; 
         default : ;   
        } 
     }  /* while */ 
  } 



#define strBufSize 512

char *copyLexemeOfSTRING(void) 
  {char c; 
   char *cp,*afterBuf,*temp;
   char buf[strBufSize]; 
   int state = 1; 

   cp = buf; 
   *cp++ = '\"';
   afterBuf = buf + strBufSize; 
   while (state != DFAACCEPT)  
     {*cp++ = c = yyyinput();  
      if (cp == afterBuf) internal("string too long\n");  
      switch (state)   
        {case 1: switch(c) 
                   {case '\\' : state = 2; break;  
                    case '\"' : state = DFAACCEPT; break;  
                    default : ;  
                   }
                 break;  
         case 2: state = 1; 
                 break; 
         default : ;   
        } 
     }  /* while */
   if ((temp = (char *)malloc((size_t)(1 + cp - buf))) 
       == 
       ((char *)NULL)
      )
      internal("malloc error in string allocation\n");
   *cp = '\0';
   strcpy(temp,buf); 
   return temp; 
  } 



/* The size of the array of pointers to permanent null-terminated copies 
 * of lexemes of identifiers.  
 */  
int maxNstrings = 509;  


/* stringTable[i] is either (char *)NULL or a 
 * permanent null-terminated copy of the lexeme of an identifier. 
 */ 
static char **stringTable; 



/*  generates and returns a pointer to a permanent null-terminated copy 
 *  of the first len characters of the string cp.
 */  
static char *allocString(char *cp, int len) 
  {char *ncp, *str_save; 
   if ((ncp = ((char *) malloc((size_t)(len+1)))) 
       == 
       ((char *) NULL) 
      )  
      internal("malloc error in allocString\n"); 
   str_save = ncp; 
   while (len--) 
         *ncp++ = *cp++; 
   *ncp = '\0';  
   return(str_save); 
  } 



/*  compute a hash value for a string */  
static long stringHash(char *cp, int len) 
  {int hashval; 
   switch (hashval = len) 
     {default: hashval += cp[len - 2] * 512; 
      case 3:  hashval += cp[len - 1] * 64; 
      case 2:  hashval += cp[1] * 8; 
      case 1:  hashval += cp[0]; break;  
      case 0:  internal("can't seek the empty string\n"); 
     }
   return(hashval); 
  } 



/* Search the string table for a string matching the first len characters of
 * cp.  If it is absent and insertion is requested, try to insert it, returning
 * a pointer to a permanent null-terminated copy if successful, null pointer 
 * if unsuccessful (due to a full table).  If absent and insertion not 
 * requested, return null pointer.  If present, return pointer to the copy. 
 */  
char *stringLookup(char *cp, int len, int insertIfAbsent) 
  {int i,itry;  
   char **dum; 
   itry = stringHash(cp,len) % maxNstrings; 
   for (i=0; i<maxNstrings; i++) 
       {dum = stringTable + itry;  
        itry = (itry + 1) % maxNstrings;     
        if (*dum == (char *)NULL) 
           {if (insertIfAbsent) *dum = allocString(cp,len); 
            return(*dum); 
           } 
        if (strlen(*dum) != len) continue; /* don't match proper prefixes */ 
        if (!strncmp(*dum,cp,(size_t)len)) return(*dum); 
       } 
   handleOverflow("out of room in string table");  
   /*NOTREACHED*/
#ifdef __GNUC__
   __builtin_unreachable();
#endif
  } 



void scanSetup(void) 
  {int i;

   if ((stringTable = (char **) malloc((size_t)(maxNstrings * sizeof(char *))))
       != 
       ((char **) NULL) 
      )   
      for (i=0; i<maxNstrings; i++) stringTable[i] = (char *)NULL; 
      else 
      internal("malloc error in string table allocation\n");  
  }


