/* $Id: sem.c 1713 2020-06-20 22:23:30Z teshields $ */

#include "gppl_glob.h" 
#include "parser.h"
#include "util.h"
#include "code.h"

static int symSpaceSize = 1000; 
       struct sym *symSpace; 
static struct sym *symStock,*afterSymSpace; 


static int envSpaceSize = 1000; 
       struct env *envSpace; 
static struct env *envStock,*afterEnvSpace; 

static int dimListSpaceSize = 1000; 
       struct dim *dimListSpace; 
static struct dim *dimListStock,*afterDimListSpace; 



void handleSemError(void) 
  {doCodeGen = False; 
  }



void semSetup(void) 
  {spaceSetup("symbol",struct sym,symSpace,
              symSpaceSize,3,afterSymSpace,symStock
             ) 

   spaceSetup("environment",struct env,envSpace,
              envSpaceSize,2,afterEnvSpace,envStock
             ) 

   spaceSetup("dimension list",struct dim,dimListSpace,
              dimListSpaceSize,2,afterDimListSpace,dimListStock
             ) 

   lambdaSym->next = lambdaSym;
   lambdaSym->dimList = lambdaDim;
   lambdaSym->paramLoc = lambdaSym;

   lambdaEnv->level = -1; 
   lambdaEnv->varLoc = lambdaSym;
   lambdaEnv->funcLoc = lambdaSym;
   lambdaEnv->outerEnv = lambdaEnv;  

   lambdaDim->next = lambdaDim;
  }
struct dim *consDimList(long size, struct dim *dimList, lineNumType line)
  {struct dim *dum;
   if (size < 1) 
      {fprintf(stderr,
         "nonpositive dimension in array declaration near line %d.\n",line
              ); 
       handleSemError();
      }
   if ((dum = dimListStock++) == afterDimListSpace)
      handleOverflow("out of space for dimension lists\n"); 
   dum->size = size;
   dum->next = dimList;
   return dum; 
  }



static enum pred nameConflict(struct sym *sym1, struct sym *sym2)
  {if (sym1->id == sym2->id) 
      {fprintf(stderr,
               "symbol %s multiply defined: lines %d and %d.\n",
               sym1->id,sym1->line,sym2->line
              );
       handleSemError();  
       return True; 
      }
      else 
      return False; 
  } 
      


static void symDupCheck(struct sym *loc1, struct sym *loc2) 
  {struct sym *dum;  
   while (loc1 != lambdaSym) 
     {dum = loc2;  
      while (dum != lambdaSym)
        {if (nameConflict(loc1,dum)) break; 
         dum = dum->next; 
        }
      loc1 = loc1->next;  
     }
  } 



/* returns the concatenation of two sequences of symbols, 
 * issuing an error message when there is a symbol E1 in
 * one sequence and a symbol E2 in the other sequence 
 * such that E1 and E2 have the same identifier. 
 * Tacks rightLoc onto the end of leftLoc and returns leftLoc. 
 */  
struct sym *concatLocs(struct sym *leftLoc, struct sym *rightLoc) 
  {struct sym **prevLeftP;
   
   symDupCheck(rightLoc,leftLoc); 
   prevLeftP = &leftLoc;
   while (*prevLeftP != lambdaSym) prevLeftP = &((*prevLeftP)->next); 
   *prevLeftP = rightLoc; 
   return leftLoc; 
  } 



struct dim *cdrDimList(struct dim *dimList, struct sym *sym, lineNumType line)
  {if (dimList == lambdaDim)
      {if (sym != lambdaSym) 
          fprintf(stderr,"extra index in line %d\'s reference to %s.\n",
                  line,sym->id
                 );
       handleSemError(); 
       return lambdaDim; 
      }
      else
      return dimList->next; 
  }



struct dim *adjustDimList(struct sym *formalSym,
                           struct sym *actualSym,
                           lineNumType line) 
  {struct dim *formDL,*actDL;
   const char *msg;

   if (formalSym == lambdaSym) return actualSym->dimList; 
   if (actualSym == lambdaSym) return lambdaDim;
   if (formalSym->type != actualSym->type) 
      {msg = "type mismatch.\n"; goto errD;}
   formDL = formalSym->dimList;
   actDL = actualSym->dimList;
   while (formDL != lambdaDim) 
     {if ((actDL == lambdaDim) || 
          (formDL->scale != actDL->scale) || 
          (formDL->size > actDL->size)
         )
         {msg = "incompatible dimensions.\n"; goto errD;}
      formDL = formDL->next; 
      actDL = actDL->next; 
     }
   return actDL;
errD:
   fprintf(stderr,
      "actual parameter %s matched to formal parameter %s\n  at line %d: %s", 
           actualSym->id,formalSym->id,line,msg
          ); 
   handleSemError(); 
   while (formDL != lambdaDim) {formDL = formDL->next; actDL = actDL->next;}
   return actDL;
  }



struct sym *cdrParamList(struct sym *loc, struct sym *func, lineNumType line)
  {if ((loc == lambdaSym) || (loc->next == lambdaSym))
      {if (func != lambdaSym) 
          fprintf(stderr,"extra actual parameter in call to %s near line %d.\n",
                  func->id,line
                 );
       handleSemError(); 
      }
  if (loc == lambdaSym)
     return lambdaSym; 
     else
     return loc->next; 
  }



void paramMatchCheck(struct sym *formParamList,
                      enum symType exprType,
                      enum pred isStgRef,
                      enum pred isFirstExpr,
                      struct sym *func,
                      lineNumType line)
  {enum pred mismatch = False;
   struct sym *symDum;

   if ((func == lambdaSym) || (formParamList == lambdaSym)) return;
   if (isFirstExpr)
      {symDum = formParamList;
       symDum = symDum->next;
       while (symDum != lambdaSym) 
         {fprintf(stderr,"missing actual parameter in line %d's call to %s.\n",
                         line,func->id
                 ); 
          handleSemError();
          mismatch = True;
          symDum = symDum->next;
         }
      }
   if (mismatch) return;
   if ((formParamList->Class == REFPARAMC) && !isStgRef)
      {fprintf(stderr,
       "actual parameter in call to %s near line %d is not a reference,\n\
   but matched to formal reference parameter %s.\n",
       func->id,line,formParamList->id 
              );
       handleSemError(); 
      }
   if (formParamList->type != exprType)
      {fprintf(stderr,
       "parameter type mismatch in call to %s near line %d: \n\
    formal parameter %s.\n",
       func->id,line,formParamList->id 
              );
       handleSemError(); 
      }
  }

 

static long scaleCalc(struct dim *dimList, enum symType type)
  {long prevScale,prevSize,sizeOfEach;

   sizeOfEach = stgSize(type,1L); 
   prevScale = 1; 
   prevSize = 1;
   while (dimList != lambdaDim) 
     {dimList->scale = prevScale * prevSize;   
      prevScale = dimList->scale; 
      prevSize = dimList->size; 
      dimList->scale *= sizeOfEach;
      dimList = dimList->next; 
     } 
   return (prevScale * prevSize * sizeOfEach); 
  }



struct sym *newStgSym(char *lexeme,
                      lineNumType line,
                      enum symType type,
                      enum stgClass Class,
                      struct dim *dimList)
  {struct sym *dum; 
   
   if (type == VOIDT) 
      {fprintf(stderr,"line %d: declared storage of type void for %s.\n",
               line,lexeme
              ); 
       handleSemError();
       type = INTT;
      }

   if ((dum = symStock++) == afterSymSpace)
      handleOverflow("out of space in symbol table\n"); 
   dum->id = lexeme; 
   dum->line = line; 
   dum->type = type; 
   dum->Class = Class; 
   dum->size = scaleCalc(dimList,dum->type); 
   if ((Class == VALPARAMC) && (dimList != lambdaDim))
      {fprintf(stderr,"array %s declared as value parameter at line %d.\n",
               lexeme,line
              ); 
       handleSemError();
      }
   if (Class == REFPARAMC) dum->size = stgSize(type,1L); 
   dum->dimList = dimList; 
   dum->next = lambdaSym; 
   return dum; 
  }
static long assignAddresses(struct sym *loc)
  {long size = 0;

   while (loc != lambdaSym) 
     {loc->addr = size; 
      size += loc->size,
      loc = loc->next; 
     } 
   return size;
  }



static void assignLevel(struct sym *loc, levelType level)
  {while (loc != lambdaSym) 
     {loc->level = level; 
      loc = loc->next; 
     } 
  }



struct sym *newFuncSym(char *lexeme,
                       lineNumType line, 
                       enum symType type,
                       labelType label,
                       struct sym *paramLoc)
  {struct sym *dum; 

   if ((dum = symStock++) == afterSymSpace)
      handleOverflow("out of space in symbol table\n"); 
   dum->id = lexeme; 
   dum->line = line; 
   dum->type = type; 
   dum->paramsSize = assignAddresses(paramLoc); 
   dum->label = label;
   dum->paramLoc = paramLoc; 
   dum->next = lambdaSym; 
   return dum; 
  }



struct env *newEnv(struct env *outerEnv, 
                    struct sym *funcSym,
                    struct sym *varLoc,
                    struct sym *funcLoc,
                    regNumType maxRegNum) 
  {struct env *innerEnv; 

   if ((innerEnv = envStock++) == afterEnvSpace)
      handleOverflow("out of space in environment table\n"); 
   symDupCheck(funcSym->paramLoc,varLoc); 
   symDupCheck(varLoc,funcLoc); 
   symDupCheck(funcSym->paramLoc,funcLoc); 
   innerEnv->funcSym = funcSym;
   innerEnv->level = outerEnv->level + 1; 
   assignLevel(funcSym->paramLoc,innerEnv->level);  
   assignLevel(varLoc,innerEnv->level);    
   assignLevel(funcLoc,innerEnv->level);   
   innerEnv->funcSym->paramLoc = funcSym->paramLoc; 
   innerEnv->varLoc =   varLoc; 
   innerEnv->funcLoc =  funcLoc; 
   innerEnv->nRegs = maxRegNum + 1;
   innerEnv->exitLabel = newLabelNum();
   innerEnv->varsSize = assignAddresses(varLoc);  
   innerEnv->outerEnv = outerEnv; 
   return innerEnv; 
  } 



struct sym *stgRefLookup(struct env *env, char *str, lineNumType line)
  {struct sym *dumSym; 

   while (env != lambdaEnv) 
     {dumSym = env->funcSym->paramLoc; 
      while ((dumSym != lambdaSym) && (dumSym->id != str))
        dumSym = dumSym->next; 
      if (dumSym != lambdaSym) return dumSym; 

      dumSym = env->varLoc; 
      while ((dumSym != lambdaSym) && (dumSym->id != str))
        dumSym = dumSym->next; 
      if (dumSym != lambdaSym) return dumSym; 

      env = env->outerEnv; 
     } 
   fprintf(stderr,"undeclared storage identifier %s at line %d.\n",
                  str,line
          );
   handleSemError(); 
   return lambdaSym; 
  }



struct sym *funcInvocLookup(struct env *env, char *str, lineNumType line)
  {struct sym *dumSym; 

   while (env != lambdaEnv) 
     {dumSym = env->funcLoc; 
      while ((dumSym != lambdaSym) && (dumSym->id != str))
        dumSym = dumSym->next; 
      if (dumSym != lambdaSym) return dumSym; 

      env = env->outerEnv; 
     } 
   fprintf(stderr,"undeclared function identifier %s at line %d.\n",
                  str,line
          );
   handleSemError(); 
   return lambdaSym; 
  }



static enum mapType inTypeOf(int op) 
  {switch (op) 
     {
      case '*':  
      case '+':  
      case '-':  
      case '/': 
      case LT:  
      case GT:  
      case EQ:  
      case LEQ:  
      case GEQ:  
      case NEQ: return ARITHM; 
      case NOT:
      case AND:  
      case OR: return LOGM; 
      default: internal("unknown operator 1\n"); 
     } 
   /*NOTREACHED*/
#ifdef __GNUC__
   __builtin_unreachable();
#endif
  }



static enum mapType outTypeOf(int op) 
  {switch (op) 
     {
      case '*':  
      case '+':  
      case '-':  
      case '/': return ARITHM; 
      case LT:  
      case GT:  
      case EQ:  
      case LEQ:  
      case GEQ:  
      case NEQ:  
      case AND:  
      case OR:  
      case NOT: return LOGM; 
      default: internal("unknown operator 2\n"); 
     } 
   /*NOTREACHED*/
#ifdef __GNUC__
   __builtin_unreachable();
#endif
  }



enum symType binOpTypeCheck(enum symType type1,
                            enum symType type2,
                            int op,
                            lineNumType line) 
  {if ((type1 == ERRT) || (type2 == ERRT)) return ERRT; 
   if (
       (type1 != type2) 
       || 
       (type1 == VOIDT)  
       || 
       ((inTypeOf(op) == ARITHM) != ((type1 == INTT) || (type1 == FLOATT)))
      ) 
      {fprintf(stderr,"binary operator type mismatch at line %d.\n",line);  
       handleSemError();  
       return ERRT;
      }
   return (outTypeOf(op) == ARITHM)?type1:BOOLT;
  } 



enum symType unrOpTypeCheck(enum symType type, int op, lineNumType line) 
  {if (type == ERRT) return ERRT; 
   if (
       (type == VOIDT)  
       || 
       ((inTypeOf(op) == ARITHM) != ((type == INTT) || (type == FLOATT)))
      ) 
      {fprintf(stderr,"unary operator type mismatch at line %d.\n",line);  
       handleSemError();  
       return ERRT;
      }
   return (outTypeOf(op) == ARITHM)?type:BOOLT;
  } 



enum symType assignmentTypeCheck(enum symType type1,
                                 enum symType type2,
                                 lineNumType line) 
  {if ((type1 == ERRT) || (type2 == ERRT)) return ERRT; 
   if (type1 != type2) 
      {fprintf(stderr,
               "type mismatch in assignment at line %d.\n",line
              );
       handleSemError();  
       return ERRT;
      } 
   return type1; 
  }



