#if defined (__cplusplus)
#include <cstring>
#include <cstdio>
#include <cstdlib>
#else
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#endif

#include "types.h"
#include "lit.h"
#include "util.h"
#include "symbol.h"

static S_table lit_set;
static bool lit_set_initialized = false;
static void lit_set_init ( void )
{ lit_set = S_empty (); /* Make a new table */
  lit_set_initialized = true; }

lit_defn in_lit_set ( string str )
{ if ( !lit_set_initialized ) lit_set_init();

  /* return binding of "str", or NULL if "str" is unbound. */
  return (lit_defn) S_look ( lit_set, S_Symbol ( toString ( str ) ) );
}

lit_defn add_lit ( char *str )
{ lit_defn lit;

  /* get binding of "str", or NULL if "str" is unbound. */
  if ( ( lit = in_lit_set ( str ) ) ) return lit;

  /* Enter a binding "sym->value" into "t", shadowing but not deleting
   *    any previous binding of "sym". */
  lit = (lit_defn) checked_malloc ( sizeof (struct _lit_defn) );
  S_enter ( lit_set, S_Symbol ( toString ( str ) ), (void *) lit );

  return lit;
}

string g_lit_name ( lit_defn lit )
{
  return lit->C_literal;
}

void s_int_lit ( lit_defn lit, char *lit_value )
{
  lit->kind = int_lit;
/*
  lit->value.integral_value = strtoll ( spelling, NULL, 10 ); */
  lit->C_literal = toString ( lit_value );
}

void s_real_lit ( lit_defn lit, char *lit_value )
{
  lit->kind = real_lit;
  char *s = strdup ( lit_value );
  char *r = strchr ( s, '\\' ); if ( r ) *r = 'e';
/*
  lit->value.real_value = strtold ( spelling, NULL ); */
  lit->C_literal = toString ( lit_value );
  free (s);
}

void s_bool_lit ( lit_defn lit, char *lit_value )
{
  lit->kind = bool_lit;
/*
  lit->value.boolean_value = ( !strcmp ( lit_value, "TRUE" ) ? 1 : 0 ); */
  lit->C_literal = toString ( lit_value );
}

void s_bits_lit ( lit_defn lit, char *lit_value )
{
  lit->kind = bits_lit;
/*
  char *bits_ptr; int radix = strtol ( spelling, &bits_ptr, 10 );
  lit->value.bits_value = strtoull ( bits_ptr+1, NULL, radix ); */
  lit->C_literal = toString ( lit_value );
}

void s_string_lit ( lit_defn lit, char *lit_value )
{
  lit->kind = string_lit;
/*
  lit->value.string_value = toString ( lit_value ); */
  lit->C_literal = toString ( lit_value );
}

static void show_lit ( S_symbol sym, lit_defn binding )
{
  fprintf ( stderr, "Key: [%s] ", S_name( sym ) );
  switch ( binding -> kind ) {
    case int_lit:
   /* fprintf ( stderr, "Int: [%-lld]\n", binding->value.integral_value ); */
      fprintf ( stderr, "Int: [%s]\n", binding->C_literal );
      break;
    case real_lit:
   /* fprintf ( stderr, "Real: [%-.20Lg]\n", binding->value.real_value ); */
      fprintf ( stderr, "Real: [%s]\n", binding->C_literal );
      break;
    case bool_lit:
   /* fprintf ( stderr, "Bits: [%-llu]\n", binding->value.bits_value ); */
      fprintf ( stderr, "Bool: [%s]\n", binding->C_literal );
      break;
    case bits_lit:
   /* fprintf ( stderr, "Bits: [%-llu]\n", binding->value.bits_value ); */
      fprintf ( stderr, "Bits: [%s]\n", binding->C_literal );
      break;
    case string_lit:
   /* fprintf ( stderr, "String: [%s]\n", binding->value.string_value ); */
      fprintf ( stderr, "String: [%s]\n", binding->C_literal );
      break;
    default:
      fprintf ( stderr, "Undefined:\n" );
      break;
  }
}

void dump_lit ( void )
{ if ( !lit_set_initialized ) lit_set_init();

  fprintf ( stderr, "\n\nLiteral Table:\n" );
  S_dump ( lit_set, (void (*)(S_symbol, void *)) show_lit );
}
