/* $Id: mDeps.h 35 2004-09-24 05:04:01Z Tom Shields $ */ 

/* The information in this file is the intellectual property of 
 * Kurt M. Bischoff.  It may not be used, copied, or transferred 
 * without permission from the author. 
 *
 * Copyright (c) 1992, Kurt M. Bischoff.  All rights reserved.
 *
 * Beta Test Release--not for general distribution.
 */ 

/*  "target machine" as used in this comment means the machine on which the 
 *   generated evaluator is to run. 
 *
 *
 *  ALIGNsIZE should be assigned the smallest integer such that: 
 *
 *      if (n is an address on the target machine and 
 *          (n % alignSize == 0)
 *         ) 
 *      then any data structure declared in a program that runs on the 
 *           target machine will satisfy the machines's alignment 
 *           requirements if it resides at n.     
 */ 
#define ALIGNsIZE 4 



/* true if the character set of the compiler development machine is the 
 * same as the character set of the machine on which the developed 
 * compiler will run 
 */
#define cscdmEQcscm 0 



/* The following constants control the sizes of data structures
   in the generated evaluator.  Each is listed with its size in the
   Ox distribution package.  These sizes assume that there is not a
   lot of memory (about 1 megabyte) available on the machine(s) 
   on which the generated evaluators are to run.  They can be 
   changed at evaluator-generation time by using Ox's -Y command 
   line options, or now (at Ox-generation time). It is unlikely that 
   the sizes other than the first three listed need to be changed
   at Ox-generation time.  Anyway, the generated evaluators will let
   you know if they run out of memory.

   * MAXnBYTESnODEsTG  400000L 
   The number of bytes allocated for parse-tree node and attribute-
   instance storage.

   * MAXnREFcOUNTS     100000L 
   The number of dependee counts (one byte each) allocated.
   
   * MAXnCHILDREN       12000L 
   The number of pointers allocated for child lists.  This is the maximum
   number of non-root parse-tree nodes.  Each one consumes sizeof(void *)
   bytes.

   * MAXsTACKsIZE        1000L 
   The maximum number of subtrees on the parse-tree stack.  This should
   be increased if the attribute grammars are expected to use much right 
   recursion and the generated evaluators are expected to process large 
   inputs.

   * ssalSPACEsIZE      10000L 
   The maximum number of solved synthesized attributes of the roots of 
   stacked subtrees.

   * rsMAXsIZE            500L 
   The maximum size of the ready set.   

   * TRAVsTACKmAXsIZE    1000L
   The maximum size of the postdecoration-traversal stack.  The size of
   stack needed for a depth-first traversal of the parse tree. 
 */



/* Below are three models for memory allocation in the generated 
   evaluators.  Choose one or design one of your own.
 */


/* small (distribution) model 
 define MAXnBYTESnODEsTG  400000L 
 define MAXnREFcOUNTS     100000L 
 define MAXnCHILDREN       12000L 
 define MAXsTACKsIZE        1000L 
 define ssalSPACEsIZE      10000L 
 define rsMAXsIZE            500L 
 define TRAVsTACKmAXsIZE    1000L
*/



/* medium model 
 define MAXnBYTESnODEsTG  2000000L 
 define MAXnREFcOUNTS      500000L 
 define MAXnCHILDREN        60000L 
 define MAXsTACKsIZE        2000L 
 define ssalSPACEsIZE      20000L 
 define rsMAXsIZE           1000L 
 define TRAVsTACKmAXsIZE    2000L
 */



/* large model 
 define MAXnBYTESnODEsTG  4000000L 
 define MAXnREFcOUNTS     1000000L 
 define MAXnCHILDREN       120000L 
 define MAXsTACKsIZE        2000L 
 define ssalSPACEsIZE      20000L 
 define rsMAXsIZE           1000L 
 define TRAVsTACKmAXsIZE    2000L
 */



/* small (distribution) model */
#define MAXnBYTESnODEsTG  400000L 
#define MAXnREFcOUNTS     100000L 
#define MAXnCHILDREN       12000L 
#define MAXsTACKsIZE        1000L 
#define ssalSPACEsIZE      10000L 
#define rsMAXsIZE            500L 
#define TRAVsTACKmAXsIZE    1000L

