/* $Id: opts.c 2098 2021-02-16 02:47:30Z teshields $ */ 

/* The information in this file is the intellectual property of 
 * Kurt M. Bischoff.  It may not be used, copied, or transferred 
 * without permission from the author. 
 *
 * Copyright (c) 1992, Kurt M. Bischoff.  All rights reserved.
 *
 * Beta Test Release--not for general distribution.
 */ 

#if defined (__cplusplus)
#include <cstdlib>
#include <cstdio>
#include <cstring>
#else
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#endif

#include "mDeps.h"
#include "opts.h"
#include "scan.h"
#include "trav.h"
#include "patchlevel.h"

extern long refSegBufferSize; 

extern void scanInit1(void); 
extern void rmTempFiles(void);
extern void erasf(const char *);
extern void printStringArray(const char **,int,FILE *); 

FILE *yoxInG; 
char *yoxInFileStrG; 
FILE *yoxOutG; 
const char *yoxOutFileStrG = "ox.out.y"; 

int nLfiles = 0; 
#define maxNLfiles 32  
struct LfileRec Lfile[maxNLfiles]; 
int currentLfile = 0; 

/*   extension must be ".l",  because an integer
     modifier will be shoved in between it and 
     the rest of the name if
     there are multiple L-files.  
*/ 
const char *yoxOutFileStrL = "ox.out.l"; 

FILE *yoxOutH; 
const char *yoxOutFileStrH = "ox.out.h"; 


int WAmax = 0; 
int RCmax = 0; 
int WSmax = 0; 


/* See mdeps.h */  
int alignSize = ALIGNsIZE;  

/* See mdeps.h */  
int CSCDMeqCSCM = cscdmEQcscm;    

/* See mdeps.h */ 
long maxNbytesNodeStg = MAXnBYTESnODEsTG;  
long maxNrefCounts    = MAXnREFcOUNTS; 
long maxNchildren     = MAXnCHILDREN; 
long maxStackSize     = MAXsTACKsIZE;  
long SSALspaceSize    = ssalSPACEsIZE;  
long RSmaxSize        = rsMAXsIZE; 
long travStackMaxSize = TRAVsTACKmAXsIZE;

/* when showStackOpts is set, ox produces an evaluator that displays 
   the sequence of operations on the stack of subtrees. 
 
   showPTsizeStats causes generation of an evaluator that reports 
   maximum sizes of some parse-tree-related data structures upon 
   termination. 
 */ 
int buildHeaderFile = 0; 
int suppressPruning = 0; 
int genEvUserCLoptsCode = 0;  
int neutralizeSemActions = 0; 
int stripOxConstructs = 0;
int showPTsizeStats = 0; 
int showStackOps = 0; 
int suppressCycleReports = 0; 

const char *ctSA[] =  
{
"/* read the command line for changes in sizes of \n", 
"                  the evaluator's data structures */\n",
"void yyyCheckForResizes(argc,argv) \n", 
"  int argc; \n", 
"  char *argv[]; \n", 
"  {int i; \n", 
"   long dum; \n", 
" \n", 
"   if (!yyyPermitUserAlloc) return; \n", 
"   for (i=1;i<argc;i++) \n", 
"     { \n", 
"      if ((argv[i][0] != \'-\') || (argv[i][1] != \'Y\')) continue; \n", 
"      if (strlen(argv[i]) < 4) goto yyyErrO1; \n",
"      if (sscanf(argv[i]+3,\"%%d\",&dum) != 1) goto yyyErrO1;\n",
"      if (dum < 2) dum = 2;\n",
"      switch (argv[i][2]) \n",
"        {case yyyNSof:   yyyMaxNbytesNodeStg = dum; break; \n",
"         case yyyRCof:   yyyMaxNrefCounts    = dum; break; \n",
"         case yyyCLof:   yyyMaxNchildren     = dum; break; \n",
"         case yyySof:    yyyMaxStackSize     = dum; break; \n",
"         case yyySSALof: yyySSALspaceSize    = dum; break; \n",
"         case yyyRSof:   yyyRSmaxSize        = dum; break; \n",
"         case yyyTSof:   yyyTravStackMaxSize = dum; break; \n",
"         default : goto yyyErrO1; \n",
"        }\n",
"      continue;  \n",
"   yyyErrO1 : fprintf(stderr,\"invalid command line option: %%s\\n\",\n",
"                             argv[i] \n", 
"                     ); \n",
"     } \n",
"  } \n", 
"   \n", 
"   \n", 
"   \n", 
};



void printPatchlevelComment(FILE *fp)
  {fprintf(fp,"/* output from Ox beta version %5.2f */\n",
           PATCHLEVEL
	  );
  }



void printPatchlevel(FILE *fp)
  {fprintf(fp,"Ox beta version %5.2f\n",PATCHLEVEL);
  }



const char *usageText[] = 
{
"*******************************************************************\n",
"This is Ox.\n",
"Copyright (c) 1992, Kurt M. Bischoff.  All rights reserved.\n",
"Unauthorized use or distributiion is prohibited.\n",
"\n",
"THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR\n",
"IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED\n",
"WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.\n",
"*******************************************************************\n",
};



void printUsageStatement(FILE *fp)
  {printStringArray(usageText,sizeof(usageText),fp);
  }



void outputUserOptsCode(void) 
  {int i; 
   for (i=0;i < (sizeof(ctSA) / sizeof(char *)); i++) 
       fprintf(yoxOutG,ctSA[i]); 
  }   



/* Set the size of one of ox's internal data structures */ 
void setAnEGsize(int len, char *str) 
  {long dum;
   if (len < 4) goto errO3; 
   if (sscanf(str+3,"%ld",&dum) != 1) goto errO3;
   if (dum < 2) dum = 2; 
   switch (str[2]) 
     {case NPDS:  maxNentsPerDepSpec    = dum; break; 
      case EPDS:  maxNeesPerDepSpec     = dum; break; 
      case AOPP:  maxNattrbOccrsPerProd = dum; break; 
      case LGNS:  ldgNodeStgSize        = dum; break; 
      case PRDL:  maxProdLength         = dum; break; 
      case NAPS:  maxNattrbsPerSym      = dum; break; 
      case RSBS:  refSegBufferSize      = dum; break; 
      case NTRM:  maxNterms             = dum; break; 
      case NSTR:  maxNstrings           = dum; break; 
      case MNGS:  maxNgSyms             = dum; break; 
      case MNST:  maxNstructTags        = dum; break; 
      case NPRD:  maxNprods             = dum; break; 
      case NMCD:  macTabSz              = dum; break; 
      case NMFP:  macFrmPrmSpSz         = dum; break; 
      case MBSS:  macBodyStgSz          = dum; break; 
      case MSSZ:  macStackSz            = dum; break; 
      case NMAP:  macActPrmSpSz         = dum; break; 
      case MAPS:  macActPrmStgSz        = dum; break; 
      case NTRV:  maxNtravs             = dum; break; 
      case NTSB:  maxNtravSegBufs       = dum; break; 
      default : goto errO3; 
     }
   return;  
errO3 : fprintf(stderr,"ox: invalid command line option: %s\n",str); 
  } 



/* set (at evaluator generation time) the size of one of the evaluator's
                 data structures */ 
void setaGEdefaultSize(int len, char *str) 
  {long dum;
   if (str[2] == 'Y') 
      {if (len > 3) goto errO1; 
       genEvUserCLoptsCode = 1;  
       return; 
      } 
   if (len < 4) goto errO1; 
   if (sscanf(str+3,"%ld",&dum) != 1) goto errO1;
   if (dum < 2) dum = 2; 
   switch (str[2]) 
     {case 'n': maxNbytesNodeStg = dum; break; 
      case 'r': maxNrefCounts    = dum; break; 
      case 'c': maxNchildren     = dum; break; 
      case 's': maxStackSize     = dum; break; 
      case 'S': SSALspaceSize    = dum; break; 
      case 'q': RSmaxSize        = dum; break; 
      case 't': travStackMaxSize = dum; break;
      default : goto errO1; 
     }
   return;  
errO1 : fprintf(stderr,"ox: invalid command line option: %s\n",str); 
  } 



FILE *sfopen(const char *name, const char *mode) 
  {FILE *dum; 
   if ((dum = fopen(name,mode)) 
       == 
       ((FILE *)NULL) 
      ) 
      {fprintf(stderr,"fatal error opening %s.\n",name); 
       exit(-1); 
      } 
      else 
      return dum; 
  } 



void openIOfilesAndSetOptions(int argc, char *argv[]) 
  {int i; 
 
   if ((argc == 2) && (argv[1][0] == '-'))
      {int j,len;

       len = strlen(argv[1]); 
       for (j=1;j<len;j++) 
           {switch (argv[1][j])
              {case 'U' : printUsageStatement(stdout); break;
               case 'V' : printPatchlevel(stdout);     break;
               default  : ;
              }
           }
       exit(0);
      }

   i=0; 
  
   while ((++i < argc) && (argv[i][0] == '-'))     ; 
   if (i == argc) 
      {fprintf(stderr,"ox: command line lacks name for Y-file\n"); 
       exit(-1); 
      } 
      else   
      yoxInFileStrG = argv[i]; 

   yoxInG = sfopen(yoxInFileStrG,"r");  
   yoxOutG = sfopen(yoxOutFileStrG,"w");  
   printPatchlevelComment(yoxOutG);
 
   while (++i < argc)  
     {if (argv[i][0] != '-') 
        {if (nLfiles == maxNLfiles) fatal("too many L-files\n"); 
         Lfile[nLfiles].inName = argv[i]; 
         if ((Lfile[nLfiles].outName = 
              (char *)malloc(strlen(yoxOutFileStrL) + 5)
             )
             == 
             ((char *)NULL) 
            ) 
            fatal("error mallocing outName\n"); 
         strcpy(Lfile[nLfiles].outName,yoxOutFileStrL); 
         nLfiles++; 
        }
     }

   if (!nLfiles) 
      {fprintf(stderr,"ox: command line lacks name for L-file\n");  
       exit(-1); 
      } 

   for (i=0;i<nLfiles;i++)  
     {Lfile[i].inPtr = sfopen(Lfile[i].inName,"r"); 
      if (nLfiles>1) 
         {sprintf(Lfile[i].outName +
                  strlen(yoxOutFileStrL) -
                  strlen(".l"),
                  ".%d",i
                 ); 
          strcat(Lfile[i].outName,".l"); 
         }
      Lfile[i].outPtr = sfopen(Lfile[i].outName,"w"); 
      printPatchlevelComment(Lfile[i].outPtr);
     } 

   yoxOutH = sfopen(yoxOutFileStrH,"w");  
   if (buildHeaderFile) printPatchlevelComment(yoxOutH);

   for (i=1;i<argc;i++) 
     {int j,len;
      long dum; 
      if (argv[i][0] != '-') continue;  
      len = strlen(argv[i]); 
      for (j=1;j<len;j++) 
        { 
         switch (argv[i][j]) 
           {case 'd' : suppressPruning = 1; break;  
            case 'h' : buildHeaderFile = 1; break;  
            case 'A' : CSCDMeqCSCM = 1; break;  
            case 'n' : neutralizeSemActions = 1; break;
            case 'S' : stripOxConstructs = 1; break;
            case 's' : showStackOps = 1; break;
            case 'u' : showPTsizeStats = 1; break;
            case 'c' : suppressCycleReports = 1; break;
	    case 'U' : printUsageStatement(stdout); break;
            case 'V' : printPatchlevel(stdout); break;
            case 'a' : 
                       if (j != 1) goto errO2; 
                       if (strlen(argv[i]) < 3) goto errO2; 
                       if (sscanf(argv[i]+2,"%ld",&dum) != 1) goto errO2;
                       switch (argv[i][j]) 
                         {case 'a' : if (dum < 1)  goto errO2; 
                                     alignSize = dum; 
                                     break; 
                          default:; 
                         } 
                       j = len; 
                       break; 
            case 'X' : if (j != 1) goto errO2; 
                       setAnEGsize(len,argv[i]); 
                       j = len; 
                       break; 
            case 'Y' : if (j != 1) goto errO2; 
                       setaGEdefaultSize(len,argv[i]); 
                       j = len; 
                       break; 
            default: ; 
           }  
        }
      continue; 
errO2 : fprintf(stderr,"ox: error in command line option: %s\n",argv[i]); 
     } 
   scanInit1();  
  } 



void closeIOfiles(void) 
  {int i;  
   fclose(yoxInG); yoxInG=0;   
   fclose(yoxOutG); yoxOutG=0;   
   for (i=0;i<nLfiles;i++) 
     {fclose(Lfile[i].inPtr); 
      fclose(Lfile[i].outPtr);Lfile[i].outPtr=0; 
     } 
   fclose(yoxOutH);yoxOutH=0;   
  } 



void rmOutputFiles(void) 
  {int i; 
   if (yoxOutG) fclose(yoxOutG); 
   erasf(yoxOutFileStrG); 
   for (i=0;i<nLfiles;i++) 
       {if(Lfile[i].outPtr)fclose(Lfile[i].outPtr); 
        erasf(Lfile[i].outName); 
       }
   if (buildHeaderFile) 
      {if(yoxOutH)fclose(yoxOutH); 
       erasf(yoxOutFileStrH); 
      }
   fprintf(stderr,"ox did not generate output.\n"); 
  } 



/* report overflow of one of the data structures of the evaluator generator.
   suggest a new size.  */
void handleOverflow(char which) 
  {const char *msg1,*msg2; 
   int  oldSize,newSize; 
   switch(which) 
     {
      case NPDS: 
           msg1 = "out of space for dependents: ";
           oldSize = maxNentsPerDepSpec; 
           break; 
      case EPDS: 
           msg1 = "out of space for dependees: ";
           oldSize = maxNeesPerDepSpec; 
           break; 
      case AOPP: 
           msg1 = "out of space for attribute occurrences: ";
           oldSize = maxNattrbOccrsPerProd; 
           break; 
      case LGNS: 
           msg1 = "out of space for dependency graph: ";
           oldSize = ldgNodeStgSize; 
           break; 
      case PRDL: 
           msg1 = "out of space for grammar rule: ";
           oldSize = maxProdLength; 
           break; 
      case NAPS: 
           msg1 = "too many attributes for one symbol: ";
           oldSize = maxNattrbsPerSym; 
           break; 
      case NTRM: 
           msg1 = "out of space for terminals: ";
           oldSize = maxNterms; 
           break; 
      case NSTR: 
           msg1 = "out of space for strings: ";
           oldSize = maxNstrings; 
           break; 
      case MNGS: 
           msg1 = "out of space for grammar symbols: ";
           oldSize = maxNgSyms; 
           break; 
      case MNST: 
           msg1 = "out of space for structure tags: ";
           oldSize = maxNstructTags; 
           break; 
      case NPRD: 
           msg1 = "out of space for productions: ";
           oldSize = maxNprods; 
           break; 
      case NMCD: 
           msg1 = "out of space for macro definitions: "; 
           oldSize = macTabSz; 
           break; 
      case NMFP: 
           msg1 = "out of space for macro formal parameters: "; 
           oldSize = macFrmPrmSpSz; 
           break; 
      case MBSS: 
           msg1 = "out of space for macro text: "; 
           oldSize = macBodyStgSz; 
           break; 
      case MSSZ: 
           msg1 = "out of space for macro use stack: "; 
           oldSize = macStackSz; 
           break; 
      case NMAP: 
           msg1 = "too many macro actual parameters: "; 
           oldSize = macActPrmSpSz; 
           break; 
      case MAPS: 
           msg1 = "macro actual parameters too long: "; 
           oldSize = macActPrmStgSz; 
           break; 
      case NTRV: 
           msg1 = "too many traversals: "; 
           oldSize = maxNtravs; 
           break; 
      case NTSB: 
           msg1 = "too many traversal actions: "; 
           oldSize = maxNtravSegBufs; 
           break; 
      default        :;  
     }
   newSize = (5*oldSize)/4; 
   if (newSize < 16) newSize = 16; 
   fprintf(stderr,msg1); 
   fprintf(stderr,"size was %d. \n",oldSize); 
   msg2 = "     Try -X%c%d option.\n"; 
   fprintf(stderr,msg2,which,newSize); 
   rmTempFiles();
   exit(-1); 
  }

