/* $Id: trav.c 2098 2021-02-16 02:47:30Z teshields $ */

/* The information in this file is the intellectual property of 
 * Kurt M. Bischoff.  It may not be used, copied, or transferred 
 * without permission from the author. 
 *
 * Copyright (c) 1992, Kurt M. Bischoff.  All rights reserved.
 *
 * Beta Test Release--not for general distribution.
 */ 

#if defined (__cplusplus)
#include <cstdio>
#else
#include <stdio.h>
#endif

#include "glob.h"
#include "opts.h"
#include "trav.h"
extern long refSegBufferPointer;
extern int prodNumSeq;
extern void printStringArray(const char **,int,FILE *); 
extern void appendToOutFile(FILE *,const char *,FILE *); 
extern void translateAndOutputRefSeg(enum instRefType,long *,long,int,int,FILE *); 

int travNumSeq = -1;
int maxNtravs = 64;
struct trav *travArr; 
struct trav travBuf;

int maxNtravSegBufs = 64;
struct travSegBuf *travSegBufArr,*lambdaTSB,travSegBufBuf;
int travSegBufAvail;
int maxNactions = 0;

int currentTrav;



void newTrav(const char *id)
  {if ((travBuf.RL && travBuf.LR) 
       ||
       (travBuf.PRE && travBuf.POST)
      )
      yerror("contradictory traversal declaration\n");
   if (!travBuf.RL  && !travBuf.LR)   travBuf.LR   = 1;
   if (!travBuf.PRE && !travBuf.POST) travBuf.POST = 1;
   if (travNumSeq == maxNtravs) handleOverflow(NTRV);
   travArr[travNumSeq] = travBuf; 
   travArr[travNumSeq].id = id; 
   travArr[travNumSeq].dirRevCondStartPtr = 
              travArr[travNumSeq].dirRevCondStopPtr = 0; 
   travArr[travNumSeq].nActions = 0;
   travArr[travNumSeq].TSBlist = lambdaTSB;
   travNumSeq++; 
  }



void clearTravBuf(void)
  {travBuf.RL = 0; 
   travBuf.LR = 0; 
   travBuf.PRE = 0; 
   travBuf.POST = 0; 
   travBuf.DA = 0;
  }



int travLookup(char *id)
  {int i;
   for (i=0; i<travNumSeq; i++) 
     if (id == travArr[i].id) return i;
   yerror("unknown traversal:\n");
   fprintf(stderr,"  %s\n",id);
   return -1;
  } 



void clearErrorTrav(void)
  {travArr[-1].dirRevCondStartPtr = travArr[-1].dirRevCondStopPtr = 0; 
   travArr[-1].nActions = 0; 
   travArr[-1].TSBlist = lambdaTSB;
  }



void outputTravCodePerProd(void)
  {int i,j;
   struct travSegBuf *TSBdum;

   if (!travNumSeq) return;
   fprintf(tempTravOut,"case %d:\n",prodNumSeq-1);
   if (travNumSeq > 1) fprintf(tempTravOut,"\tswitch(yyyi)\t{ \n"); 
   for (i=0; i<travNumSeq; i++) 
     {if (travNumSeq > 1) fprintf(tempTravOut,"\t\tcase %d:\n",i);
      fprintf(tempTravOut,"\t\t\tswitch(yyyPass)\t{\n");
      fprintf(tempTravOut,"\t\t\t\tcase 0:\n");
      fprintf(tempTravOut,"yyyRL = "); 
      if (travArr[i].dirRevCondStartPtr != travArr[i].dirRevCondStopPtr)
         {if (!travArr[i].LR) fprintf(tempTravOut,"!");
          translateAndOutputRefSeg(trav,
                                   &(travArr[i].dirRevCondStartPtr),
                                   travArr[i].dirRevCondStopPtr,
                                   DONTCARE,DONTCARE,tempTravOut
                                  ); 
         }
         else
         fprintf(tempTravOut,"%d",!travArr[i].LR);
      fprintf(tempTravOut,";");
      TSBdum = travArr[i].TSBlist;
      j = 0; 
      while (TSBdum != lambdaTSB)
        {if (TSBdum->orderRevCondStartPtr != TSBdum->orderRevCondStopPtr)
            {fprintf(tempTravOut,"\nif (");
             if (!travArr[i].POST) fprintf(tempTravOut,"!");
             translateAndOutputRefSeg(trav,
                                      &(TSBdum->orderRevCondStartPtr),
                                      TSBdum->orderRevCondStopPtr,
                                      DONTCARE,DONTCARE,tempTravOut
                                     );
             fprintf(tempTravOut,") yyySetCond(%d)\n",j);
            }
            else
            {if (!travArr[i].POST) fprintf(tempTravOut,"yyySetCond(%d)\n",j);
            }
         TSBdum = TSBdum->next;
         j++;
        }
      fprintf(tempTravOut,"\n\t\t\t\tcase 1:\n");  
      TSBdum = travArr[i].TSBlist;
      j = 0; 
      while (TSBdum != lambdaTSB)
        {fprintf(tempTravOut,"\nif (yyyCond(%d) != yyyPass) {",j); 
         translateAndOutputRefSeg(trav,
                                  &(TSBdum->actionStartPtr),
                                  TSBdum->actionStopPtr,
                                  DONTCARE,DONTCARE,tempTravOut
                                 );
         fprintf(tempTravOut,"}"); 
         TSBdum = TSBdum->next;
         j++;
        }
      fprintf(tempTravOut,"\n\t\t\t\tbreak;\n");
      fprintf(tempTravOut,"\t\t\t\t\t}\n"); 
      if (travNumSeq > 1) fprintf(tempTravOut,"\t\tbreak;\n"); 
      if (maxNactions < travArr[i].nActions) 
         maxNactions = travArr[i].nActions;
      travArr[i].nActions = 0; 
      travArr[i].TSBlist = lambdaTSB; 
     } /* for */
   if (travNumSeq > 1) fprintf(tempTravOut,"\t\t\t}\n"); 
   fprintf(tempTravOut,"\nbreak;\n");
   travSegBufAvail = 0;  
   clearErrorTrav();
  }



void printTravDefines(FILE *fp)
  {fprintf(fp,"\n\n\n"); 
   if      (maxNactions <= 16) 
      {fprintf(fp,"#define condStg unsigned int conds;\n");
       fprintf(fp,"#define yyyClearConds {yyyTST->conds = 0;}\n");
       fprintf(fp,"#define yyySetCond(n) {yyyTST->conds += (1<<(n));}\n");
       fprintf(fp,"#define yyyCond(n) ((yyyTST->conds & (1<<(n)))?1:0)\n");
      }
   else if (maxNactions <= 32)
      {fprintf(fp,"#define condStg unsigned long conds;\n");
       fprintf(fp,"#define yyyClearConds {yyyTST->conds = 0;}\n");
       fprintf(fp,"#define yyySetCond(n) {yyyTST->conds += (1<<(n));}\n");
       fprintf(fp,"#define yyyCond(n) ((yyyTST->conds & (1<<(n)))?1:0)\n");
      }
   else 
      {fprintf(fp,
           "#define yyyNCW ceiling(%d,sizeof(unsigned int))];\n",maxNactions
              );
       fprintf(fp,"#define condStg unsigned int conds[yyyNCW];\n");
       fprintf(fp,
   "#define yyyClearConds {int i; for (i=0;i<yyyNCW;i++)yyyTST->conds[i]=0;}\n"
              );
       fprintf(fp, 
         "#define yyySetCond(n) {yyyTST->conds[(n)/sizeof(unsigned int)] "
              );
       fprintf(fp,"+= (1<<(n%%sizeof(unsigned int)));}\n");
       fprintf(fp,
        "#define yyyCond(n) ((yyyTST->conds[(n)/sizeof(unsigned int)] "
              );
       fprintf(fp,"& (1<<(n%%sizeof(unsigned int))))?1:0)\n");
      }
  }



void startTravDirCond(void)
  {if (travArr[currentTrav].dirRevCondStartPtr 
       !=
       travArr[currentTrav].dirRevCondStopPtr 
      )
      {yerror("multiple @revdirection\n");
       fprintf(stderr,"  traversal name: %s\n",travArr[currentTrav].id);
      }
   travArr[currentTrav].dirRevCondStartPtr = travSegBufBuf.actionStartPtr;
  }



void startTravOrderCond(void)
  {if (travSegBufBuf.orderRevCondStartPtr 
       != 
       travSegBufBuf.orderRevCondStopPtr
      )
      {yerror("multiple @revorder\n");
       fprintf(stderr,"  traversal name: %s\n",travArr[currentTrav].id);
      }
   travSegBufBuf.orderRevCondStartPtr = travSegBufBuf.actionStartPtr;
  }



void startTravAction(void)
  {travSegBufBuf.actionStartPtr = refSegBufferPointer;}



void wrapUpTravDirCond(void)
  {travArr[currentTrav].dirRevCondStopPtr = refSegBufferPointer - 1;}



void wrapUpTravOrderCond(void)
  {travSegBufBuf.orderRevCondStopPtr = refSegBufferPointer - 1;}



void clearTSBbuf(void)
  {travSegBufBuf.orderRevCondStartPtr = travSegBufBuf.orderRevCondStopPtr = 0;
   travSegBufBuf.actionStartPtr = travSegBufBuf.actionStopPtr;
  }



void wrapUpTravAction(void)
  {struct travSegBuf **pTSBdum;

   travSegBufBuf.actionStopPtr = refSegBufferPointer - 1;
   travSegBufBuf.next = lambdaTSB;
   travSegBufArr[travSegBufAvail] = travSegBufBuf;
   clearTSBbuf();
   travArr[currentTrav].nActions++; 
   pTSBdum = &(travArr[currentTrav].TSBlist);
   while (*pTSBdum != lambdaTSB) pTSBdum = &((*pTSBdum)->next);
   *pTSBdum =  travSegBufArr + travSegBufAvail;
   travSegBufAvail++; 
  }



const char *ct45[] = 
{
"\n",
"\n",
"\n",
"struct yyyTravStackItem {yyyGNT *node; \n",
"                         char isReady;\n",
"                         condStg\n",
"                        };\n",
"\n",
"\n",
"\n",
"yyyDoTraversals()\n",
"{struct yyyTravStackItem *yyyTravStack,*yyyTST,*yyyAfterTravStack;\n",
" register yyyGNT *yyyTSTn,**yyyCLptr1,**yyyCLptr2; \n",
" register int yyyi,yyyRL,yyyPass;\n",
"\n",
" if (!yyyYok) return;\n",
" if ((yyyTravStack = \n",
"                 ((struct yyyTravStackItem *) \n",
"                  malloc(yyyTravStackMaxSize * sizeof(struct yyyTravStackItem))\n",
"                 )\n",
"     )\n",
"     == \n",
"     (struct yyyTravStackItem *)NULL\n",
"    ) \n",
"    {fprintf(stderr,\"malloc error in traversal stack allocation\\n\"); \n",
"     exit(-1); \n",
"    } \n",
"\n",
"yyyAfterTravStack = yyyTravStack + yyyTravStackMaxSize; \n",
"yyyTravStack++; \n",
"\n",
"\n",
};



const char *ct45_5[] = 
{
"yyyTST = yyyTravStack; \n",
"yyyTST->node = yyyStack->node;\n",
"yyyTST->isReady = 0;\n",
"yyyClearConds\n",
"\n",
"while(yyyTST >= yyyTravStack)\n",
"  {yyyTSTn = yyyTST->node;\n",
"   if (yyyTST->isReady)  \n",
"      {yyyPass = 1;\n",
"       goto yyyTravSwitch;\n",
"yyyTpop:\n",
"       yyyTST--;\n",
"      } \n",
"      else \n",
"      {yyyPass = 0;\n",
"       goto yyyTravSwitch;\n",
"yyyTpush:\n",
"       yyyTST->isReady = 1;  \n",
"       if (yyyTSTn->prodNum)\n",
"          if (yyyRL)\n",
"             {yyyCLptr2 = yyyTSTn->cL; \n",
"              while \n",
"                ((yyyCLptr2 != yyyNextCLspace)\n",
"                 &&\n",
"                 ((*yyyCLptr2)->parent == yyyTSTn) \n",
"                )  \n",
"                {if (++yyyTST == yyyAfterTravStack)\n",
"                    yyyHandleOverflow(yyyTSof);\n",
"                    else\n",
"                    {yyyTST->node = *yyyCLptr2; \n",
"                     yyyTST->isReady = 0; \n",
"                     yyyClearConds\n",
"                    }\n",
"                 yyyCLptr2++; \n",
"                } \n",
"             } /* right to left */\n",
"             else  /* left to right */\n",
"             {yyyCLptr1 = yyyCLptr2 = yyyTSTn->cL; \n",
"              while \n",
"                ((yyyCLptr2 != yyyNextCLspace)\n",
"                 &&\n",
"                 ((*yyyCLptr2)->parent == yyyTSTn) \n",
"                )  \n",
"                yyyCLptr2++; \n",
"              while (yyyCLptr2-- > yyyCLptr1)\n",
"                if (++yyyTST == yyyAfterTravStack)\n",
"                   yyyHandleOverflow(yyyTSof);\n",
"                   else\n",
"                   {yyyTST->node = *yyyCLptr2; \n",
"                    yyyTST->isReady = 0; \n",
"                    yyyClearConds\n",
"                   }\n",
"             } /* left to right */\n",
"      } /* else */\n",
"   continue;\n",
"yyyTravSwitch:\n",
"\t\t\t\tswitch(yyyTSTn->prodNum)\t{\n",
};



const char *ct48[] = 
{
"\t\t\t\t\t\t\t\t} /* switch */ \n",
"   if (yyyPass) goto yyyTpop; else goto yyyTpush; \n",
"  } /* while */ \n",
" } /* for */ \n",
"} /* yyyDoTraversals */ \n",
"\n",
}; 



int someTravEnabled(void)
  {int i,dum;
   for (i=0;i<travNumSeq;i++) if (! travArr[i].DA) return 1;
   return 0;
  }



void printTravCode(void)
  {int i;
  
   if (!travNumSeq) 
      {fprintf(temp1Out,"yyyDoTraversals(){}\n");
       return;
      }
   printTravDefines(temp1Out);
   printStringArray(ct45,sizeof(ct45),temp1Out);
   fprintf(temp1Out,"for (yyyi=0; yyyi<%d; yyyi++) {\n",travNumSeq);
   for (i=0;i<travNumSeq;i++) 
       if (travArr[i].DA)  fprintf(temp1Out,"if (yyyi == %d) continue;\n",i);
   printStringArray(ct45_5,sizeof(ct45_5),temp1Out);
   appendToOutFile(tempTravOut,tempTravFileID,temp1Out); 
   printStringArray(ct48,sizeof(ct48),temp1Out);
  } 



