%{
/*
  Lexical Analyzer
*/
%}

letter			[a-zA-Z_]
digit			[0-9]
hex_digit		[a-fA-F0-9]
oct_digit		[0-7]
nonzero_digit		[1-9]
dec_constant		{nonzero_digit}{digit}*
hex_constant		(0x|0X){hex_digit}+
oct_constant		0{oct_digit}*

%%
[ \t]*			{ count(); /* ignore whitespace */ }
\n			{ linenumber++; columnnumber = 0; }
"/*"			{ comment(); }

int			{ count(); return TK_INT; }
extern			{ count(); return TK_EXTERN; }
struct			{ count(); return TK_STRUCT; }
if			{ count(); return TK_IF; }
else			{ count(); return TK_ELSE; }
for			{ count(); return TK_FOR; }
do			{ count(); return TK_DO; }
while			{ count(); return TK_WHILE; }
sizeof			{ count(); return TK_SIZEOF; }
break			{ count(); return TK_BREAK; }
return			{ count(); return TK_RETURN; }

({letter}|_)({letter}|{digit}|_)*	{ count(); id(); return TK_IDENTIFIER; }

{dec_constant}		{ count(); dec_constant(); return TK_ICONSTANT;
			  @{ @TK_ICONSTANT.val@ = yylval.i; @}
			}
{hex_constant}		{ count(); hex_constant(); return TK_ICONSTANT;
			  @{ @TK_ICONSTANT.val@ = yylval.i; @}
			}
{oct_constant}		{ count(); oct_constant(); return TK_ICONSTANT;
			  @{ @TK_ICONSTANT.val@ = yylval.i;
			     @TK_ICONSTANT.val@ = yylval.i; @}
			}


"++"			{ count(); return TK_INC; }
"--"			{ count(); return TK_DEC; }
"->"			{ count(); return TK_PTR; }
"||"			{ count(); return TK_OR; }
"<="			{ count(); return TK_LE; }
">="			{ count(); return TK_GE; }
"=="			{ count(); return TK_EQ; }
"!="			{ count(); return TK_NE; }
"&&"			{ count(); return TK_AND; }

"."			{ count(); return('.'); }
","			{ count(); return(','); }
";"			{ count(); return(';'); }
"{"			{ count(); return('{'); }
"}"			{ count(); return('}'); }
"("			{ count(); return('('); }
")"			{ count(); return(')'); }
"["			{ count(); return('['); }
"]"			{ count(); return(']'); }
"&"			{ count(); return('&'); }
"!"			{ count(); return('!'); }
"="			{ count(); return('='); }
"-"			{ count(); return('-'); }
"+"			{ count(); return('+'); }
"*"			{ count(); return('*'); }
"/"			{ count(); return('/'); }
"%"			{ count(); return('%'); }
"<"			{ count(); return('<'); }
">"			{ count(); return('>'); }

.			{ /* sop up any other characters */ error("unknown\
 character 0%o (%c)", yytext[0], yytext[0]); }

%%
